<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version
 * @author Federico Tarantini federicot@gmail.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

/**
 * Purge old db entries cronjob
 */

require_once("system/Cronjob.php");
require_once('core/ActivityLog.php');
require_once('core/FilesystemPeer.php');
require_once('core/UserSessionPeer.php');
require_once('core/UserPreferencePeer.php');
require_once('lib/html/UploadHandler.php');
require_once('plugins/pastebin/lib/Pastebin.php');
require_once('plugins/pastebin/lib/PastebinPeer.php');

class PurgeOldEntries extends Cronjob {
	public function __construct() {
		$this->jobname = basename(__FILE__);
		parent::__construct();
	}

	public function run() {
		$this->purge_activity_log();
		$this->purge_user_accounts();
		$this->purge_deleted_files();
		$this->purge_user_sessions();
		$this->purge_cron_history();
		$this->purge_uploaded_files();
		$this->purge_pastes();
	}

	//filesystem entries that are marked deleted by fules
	public function purge_deleted_files() {
		$crit=new Criteria();
		$crit->add(FilesystemPeer::DELETED,true);
		$files=FilesystemPeer::doSelect($crit);
		foreach ($files as $file) {
			$file->delete();
		}
	}

	//Purges older than 30 days activity_log table entries 
	public function purge_activity_log() {
		$now = time();
		$m30days = $now-2592000; //today minus 30 days (2592000 seconds)
		$date = date('Y-m-d h:i:s', $m30days);
		
		$crit = new Criteria();
		$crit->add(ActivityLogPeer::ACTIVITY_DATE, $date, Criteria::LESS_EQUAL);
		ActivityLogPeer::doDelete($crit);
	}

	//Purges useraccounts that have STAUS_PENDING and that are 30 days old
	public function purge_user_accounts() {
		$now = time();
		$m30days = $now-2592000; //today minus 30 days (2592000 seconds)

		$crit = new Criteria();
		$crit->add(UserPeer::STATUS,User::STATUS_PENDING);
		$crit->add(UserPeer::CREATE_DATE, date('Y-m-d h:i:s', $m30days), Criteria::LESS_EQUAL);
		$sel_user = UserPeer::doDelete($crit);
	}

	//purge login records 30 days after they're last touched.
	// We don't want to use login_time, because we allow
	// sessions to be used as long as users are active enough.
	public function purge_user_sessions() {
		$now = time();
		$m30days = $now-2592000; //today minus 30 days (2592000 seconds)

		$crit = new Criteria();
		$crit->add(UserSessionPeer::MODIFIED, date('Y-m-d h:i:s', $m30days), Criteria::LESS_EQUAL);
		$sel_user = UserSessionPeer::doDelete($crit);
	}

	//purge all cron activity after 30 days
	public function purge_cron_history() {
		$now = time();
		$m30days = $now-2592000; //today minus 30 days (2592000 seconds)

		$crit = new Criteria();
		$crit->add(CronHistoryPeer::RUNDATE, date('Y-m-d h:i:s', $m30days), Criteria::LESS_EQUAL);
		$sel_user = CronHistoryPeer::doDelete($crit);
	}

	// Purge the files uploaded using the upload widget that weren't properly handled
	public function purge_uploaded_files() {
		$uploadTmpDir = UploadHandler::getTmpUploadDir();
		$dh = opendir($uploadTmpDir);
		if (!$dh) return;

		$threeDaysAgo = time() - 60*60*24*3;
		while (($dir = readdir($dh)) !== false) {
			// $dir must have the format Y-m-d-H-i-s, as returned by UploadHandler::generateUploadHash
			if (!preg_match('/^([0-9]{4})-([0-9]{2})-([0-9]{2})-([0-9]{2})-([0-9]{2})-([0-9]{2})-.*$/', $dir, $matches)) {
				continue;
			}

			list($_, $y, $m, $d, $h, $i, $s) = $matches;
			$timestamp = mktime($h, $i, $s, $m, $d, $y);
			if ($timestamp < $threeDaysAgo) {
				// The files were uploaded more than three days ago, so purge the directory.
				$uploadDir = SystemUtils::buildFilePath($uploadTmpDir, $dir);
				$output = array();
				SystemUtils::execute("rm -rf $uploadDir", $output);
			}
		}
	}

	public function purge_pastes() {
		$sql = "SELECT * FROM pastebin WHERE ";

		//Check expiration
		$expiration_offsets = PastebinPeer::getExpirationOffsets();
		$arr_sql_blocks = array();
		foreach($expiration_offsets as $expiration_type => $expiration_offset) {
			$arr_sql_blocks[] = "(expiration = " . $expiration_type . " AND (post_date + interval '" . $expiration_offset ." seconds') < now())";
		}
		$sql_expiration = implode( $arr_sql_blocks , ' OR ');

		$sql .= $sql_expiration;

		$con = Propel::getConnection(UserPeer::DATABASE_NAME);
		$stmt = $con->createStatement();
		$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_NUM);

		try {
			$pastes = PastebinPeer::populateObjects($rs);
			foreach($pastes as $paste) {
				$paste->delete();
			}
		} catch(Exception $e) {
			return array();
		}
	}
}

$cronjob = new PurgeOldEntries();
$cronjob->run();
?>
