
CREATE TABLE calendar (
    calendar_id serial NOT NULL,
    ref_id integer NOT NULL,
    section character varying(20) NOT NULL,
    calendar_name character varying(255),
    is_public boolean,
    description text,
    timezone text,
    working_hour_start time without time zone,
    working_hour_end time without time zone,
    first_day_of_week integer DEFAULT 1
);

ALTER TABLE ONLY calendar
    ADD CONSTRAINT calendar_pkey PRIMARY KEY (calendar_id);

CREATE INDEX calendar_id_idx ON calendar USING btree (calendar_id);

CREATE INDEX ref_id_idx ON calendar USING btree (ref_id);

CREATE TABLE calendar_event (
    calendar_event_id serial NOT NULL,
    calendar_id integer NOT NULL,
    subject character varying(255) NOT NULL,
    description text,
    from_date timestamp with time zone NOT NULL,
    to_date timestamp with time zone NOT NULL,
    from_hour time without time zone,
    to_hour time without time zone,
    occurrence integer DEFAULT 0,
    occurrence_to timestamp with time zone,
    is_public boolean DEFAULT false
);


--

ALTER TABLE ONLY calendar_event
    ADD CONSTRAINT calendar_event_id_pkey PRIMARY KEY (calendar_event_id);

ALTER TABLE ONLY calendar_event
    ADD CONSTRAINT calendar_id_fkey FOREIGN KEY (calendar_id) REFERENCES calendar(calendar_id);



CREATE INDEX calendar_event_from_date_idx ON calendar_event USING btree (from_date);

CREATE INDEX calendar_event_to_date_idx ON calendar_event USING btree (to_date);


CREATE OR REPLACE FUNCTION create_serial_event() RETURNS trigger AS
$$
DECLARE
	start_time TIMESTAMP WITH TIME ZONE;
	end_time TIMESTAMP WITH TIME ZONE;
	max_time TIMESTAMP WITH TIME ZONE;
	interval_leap INTERVAL;
BEGIN

	IF NEW.occurrence <>0 AND NEW.occurrence_to IS NOT NULL THEN
		IF NEW.occurrence = 1 THEN
			interval_leap := CAST ('1 day' AS INTERVAL);
		ELSIF NEW.occurrence = 4 THEN
			interval_leap := CAST ('1 week' AS INTERVAL);
		ELSIF NEW.occurrence = 5 THEN
			interval_leap := CAST ('1 month' AS INTERVAL);
		ELSE 
			RETURN NULL;
		END IF;

		max_time := NEW.occurrence_to + interval_leap;
		start_time := NEW.from_date + interval_leap;
		end_time := NEW.to_date + interval_leap;

		WHILE start_time <= max_time LOOP
			INSERT INTO calendar_event (calendar_id,subject,description,from_date,to_date,from_hour,
			to_hour,is_public) VALUES (NEW.calendar_id, NEW.subject, NEW.description,
			start_time, end_time, NEW.from_hour, NEW.to_hour, NEW.is_public);

			start_time :=  start_time +  interval_leap;
			end_time := end_time +  interval_leap;
		END LOOP;
	END IF;

	RETURN NULL;
END;
$$ LANGUAGE 'plpgsql';


CREATE OR REPLACE FUNCTION update_reminders() RETURNS trigger AS
$$
DECLARE
	reminders RECORD;
BEGIN

	FOR reminders IN SELECT * FROM calendar_event_reminder cr WHERE cr.calendar_event_id = NEW.calendar_event_id LOOP
		IF reminders.time_unit = 1 THEN
			UPDATE calendar_event_reminder SET send_time =  NEW.from_date - INTERVAL '1 hour' WHERE calendar_event_id = NEW.calendar_event_id;
		ELSIF reminders.time_unit = 2 THEN
			UPDATE calendar_event_reminder  SET send_time =  NEW.from_date - INTERVAL '1 day' WHERE calendar_event_id = NEW.calendar_event_id;
		ELSIF reminders.time_unit = 3 THEN
			UPDATE calendar_event_reminder  SET send_time =  NEW.from_date - INTERVAL '1 week' WHERE calendar_event_id = NEW.calendar_event_id;
		END IF;
	END LOOP;

RETURN NEW;
END;
$$ LANGUAGE 'plpgsql';

CREATE TRIGGER create_serial_event_trigg
    AFTER INSERT ON calendar_event
    FOR EACH ROW
    EXECUTE PROCEDURE create_serial_event();


CREATE TRIGGER update_reminders_trigg
    AFTER UPDATE ON calendar_event
    FOR EACH ROW
    EXECUTE PROCEDURE update_reminders();
--



CREATE TABLE calendar_event_reminder (
    calendar_event_id integer NOT NULL,
    send_time timestamp with time zone NOT NULL,
    time_unit integer DEFAULT 0
);


ALTER TABLE ONLY calendar_event_reminder
    ADD CONSTRAINT calendar_event_reminder_pkey PRIMARY KEY (calendar_event_id, time_unit);

ALTER TABLE ONLY calendar_event_reminder
	ADD CONSTRAINT calendar_event_fkey FOREIGN KEY (calendar_event_id) REFERENCES calendar_event(calendar_event_id)
	ON DELETE CASCADE;

CREATE INDEX calendar_event_reminder_send_time ON calendar_event_reminder USING btree (send_time);


CREATE OR REPLACE function update_reminder_send_time() RETURNS trigger AS
$$
DECLARE
new_date  TIMESTAMP WITH TIME ZONE;
BEGIN
	SELECT calendar_event.from_date INTO new_date FROM calendar_event WHERE calendar_event_id = NEW.calendar_event_id;
		IF NEW.time_unit = 1 THEN
			NEW.send_time = new_date - INTERVAL '1 hour';
		ELSIF NEW.time_unit = 2 THEN
			NEW.send_time = new_date - INTERVAL '1 day';
		ELSIF NEW.time_unit = 3 THEN
			NEW.send_time = new_date - INTERVAL '1 week';
		END IF;

	RETURN NEW;
END;
$$ LANGUAGE 'plpgsql';


CREATE TRIGGER update_send_time_trigg
    BEFORE UPDATE ON calendar_event_reminder
    FOR EACH ROW
    EXECUTE PROCEDURE update_reminder_send_time();


INSERT INTO plugin (plugin_name,plugin_desc,plugin_type,plugin_order) VALUES ('usercalendar', 'User calendar', 'user', 3); 

--Activate user plugins
CREATE OR REPLACE function activate_all_users_plugins() RETURNS Integer AS
$$
DECLARE
	u RECORD;
	plg RECORD;
	
BEGIN

	FOR plg IN SELECT * FROM plugin WHERE plugin_type='user' LOOP
		FOR u IN SELECT * FROM "user" WHERE user_id NOT IN (SELECT user_id FROM user_plugin WHERE plugin_id=plg.plugin_id) LOOP
			INSERT INTO user_plugin (plugin_id, user_id) VALUES (plg.plugin_id, u.user_id);
			--Calendar
			IF plg.plugin_name='usercalendar' THEN
				INSERT INTO calendar(ref_id, section,calendar_name,is_public,description) SELECT u.user_id, 'user', 'My calendar', false, 'User calendar' WHERE u.user_id NOT IN (SELECT ref_id FROM calendar WHERE section='user');
			END IF;
		END LOOP;
	END LOOP;

	RETURN 0;
END;
$$ LANGUAGE 'plpgsql';
SELECT activate_all_users_plugins();
--DROP function activate_all_users_plugins();
