<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli	<marcelo@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('exceptions/InvalidFileException.php');
require_once('exceptions/FilePermissionsException.php');
require_once('SystemUtils.php');
require_once('plugin/ActionManager.php');
require_once('core/SettingPeer.php');

class GFConf {
	private static $data = array();
	private static $defaults = array();
	private static $initialized = false;

	const SECTION_CONFIG = 'config';
	
	/**
	 * Private constructor - this class can't be instantiated
	 */
	private function __construct() {
	}
	
	/**
	 * Reads the cache file
	 */
	private static function initialize() {
		if (self::getMainConfigFile()) {
			//Main configuration file still exists, load from cache
			$cacheFile = self::getCacheFile();
			if (!file_exists($cacheFile)) {
				throw new Exception("Configuration cache doesn't exist, run create_config_cache script first");
			} else if ( ($serializedCache = file_get_contents($cacheFile)) === false ) {
				throw new FilePermissionsException("cacheFile");
			}

			$unserializedCache = unserialize($serializedCache);
			self::$data = $unserializedCache[self::SECTION_CONFIG];
		}
		self::$defaults = self::getDefaults();
		self::$initialized = true;
	}
	
	/**
	 * Reload the configuration from the filesystem.
	 */
	public static function reload() {
		self::initialize();
	}

	public static function buildCacheContent($verbose = true) {
		$mainConfigFile = self::getMainConfigFile();
		
		// Build the array of configuration files to parse
		$configMap = array();
		$configMap['system'] = $mainConfigFile;
		
		// Now search for the plugins' configuration files
		include($mainConfigFile);
		if (!array_key_exists('pluginsPath', $config)) {
			throw new Exception('"pluginsPath" variable not set in main configuration file');
		}
		$pluginsPath = $config['pluginsPath'];
		if (!is_dir($pluginsPath)) {
			throw new InvalidFileException($pluginsPath);
		}
		$dh = opendir($pluginsPath);
		while ($pluginName = readdir($dh)) {
			if (preg_match('/^\\./', $pluginName)) {
				continue;
			}
			$gforge_etc_dir = getenv('GFORGE_ETC_DIR');
			if (empty($gforge_etc_dir)) {
				$gforge_etc_dir='/etc/gforge';
			}
			$confFile = $gforge_etc_dir.'/plugins/'.$pluginName.'/'.$pluginName.'.conf';
			$defaultConfigFile = $pluginsPath .DIRECTORY_SEPARATOR . $pluginName . DIRECTORY_SEPARATOR .$pluginName.'.conf';
			if (file_exists($confFile)) {
				$configMap[$pluginName] = $confFile;
			} else if (file_exists($defaultConfigFile)) {
				echo "Missing Config For: $pluginName\n";
			}
		}
		
		$cacheContents = array();
		$cacheContents[self::SECTION_CONFIG] = array();
		foreach ($configMap as $class => $file) {
			if ($verbose) {
				echo 'Parsing '.$file."...\n";
			}
			$arr = self::_parseCache($file);
			foreach ($arr as $key=>$val) {
				$cacheContents[self::SECTION_CONFIG][$class.".".$key]=$val;
			}
		}
		
		// See if some values need to be overridden (license check)
		$overrides = ActionManager::notify('VALIDATE_CONFIG_CACHE', $cacheContents[self::SECTION_CONFIG]);
		foreach ($overrides as $override) {
			$result = $override->getResult();
			foreach ($result as $var => $val) {
				if (isset($cacheContents[self::SECTION_CONFIG][$var]) && $cacheContents[self::SECTION_CONFIG][$var] != $val && $verbose) {
					echo 'Warning: the value of the variable "'.$var.'" was overridden to "'.$val.'"'."\n";
				}
				$cacheContents[self::SECTION_CONFIG][$var] = $val;
			}
		}
		
		return $cacheContents;
	}
	
	/**
	 * Creates the configuration cache
	 *
	 * @param	array $config	Associative class that contains config class => file to parse
	 * @param 	string $configFile	File to store the cache
	 */
	public static function generateCache($verbose=true) {
		$cacheFile = self::getCacheFile();

		if (!($fh = fopen($cacheFile, "w"))) {
			throw new FilePermissionsException($cacheFile);
		}

		$cacheContents = self::buildCacheContent($verbose);
		
		fputs($fh, serialize($cacheContents));
		fclose($fh);
	}

	private static function _parseCache($file) {
		$config = (isset($config)) ? $config : array();
		if (!file_exists($file)) {
			//throw new InvalidFileException($file);
			return array();
		}
		
		include($file);

		// $config now holds the configuration variables
		return $config;
	}
	
	/**
	 * Returns the full file name of GForge's main configuration file
	 * @return	mixed path of the main config file, false if it doesn't exists
	 */
	public static function getMainConfigFile() {
		$gforge_etc_dir = getenv('GFORGE_ETC_DIR');
		if (empty($gforge_etc_dir)) {
			$gforge_etc_dir='/etc/gforge';
		}
		$mainConfigPath = $gforge_etc_dir;
		$mainConfigFile = SystemUtils::buildFilePath($mainConfigPath, 'gforge.conf');
		if (!file_exists($mainConfigFile)) {
			return false;
		}
		return $mainConfigFile;
	}
	
	/**
	 * Return the file name of the configuration cache file
	 * @return	string
	 */
	public static function getCacheFile() {
		$mainConfigFile = self::getMainConfigFile();

		if (!file_exists($mainConfigFile)) {
			throw new InvalidFileException($mainConfigFile);
		} else if (!@include($mainConfigFile)) {
			throw new InvalidFileException($mainConfigFile);
		}
		
		// at this point, $config holds the configuration variables
		if (!array_key_exists('cacheDir', $config)) {
			throw new Exception('"cacheDir" not set in main configuration file');
		}
		
		$cacheFile = SystemUtils::buildFilePath($config['cacheDir'], 'config.cache');
		return $cacheFile;
	}

    protected static function unserializeIfNeeded($raw) {
    	return self::isSerialized($raw) ? @unserialize($raw) : $raw;
    }

	protected static function isSerialized($str) {
	    return ($str === serialize(false) || @unserialize($str) !== false);
	}

	/**
	 * Retrieve a configuration variable's current value
	 * @param  string $name Setting name to retrieve
	 * @param  mixed $default Value to return if the setting does not exist
	 * @param  boolean $dbOnly (deprecated) Only return values from the database setting table, not file-based cache files.
	 * @throws Exception
	 * @return mixed  The retrieved value (or default if none is found).  Existing values will
	 * be deserialized (e.g., arrays, etc.) and returned as PHP types.
	 */
	public static function get($name, $default = null, $dbOnly = false) {

		//This has been deprecated
		if ($dbOnly) return $default;
		if (!self::$initialized) {
			//See if the main config file still exists, if it doesn't, use the database
			self::initialize();
		}


		if (strpos($name, '.') === false) {
			//Pedantic check, not sure if we need this one anymore
			throw new Exception('Configuration variable "'.$name.'" is not in the form "(class).(variable)"');
		}


		if (isset(self::$data[$name])) {
			//If the setting is already in the cache, return it
			return self::$data[$name];
		} else {
			//No setting in internal cache, let's hit the database
			$ret = SettingPeer::FindSetting(self::SECTION_CONFIG, 0, $name);

			if ($ret != null) {
				//Found it, store it in the cache for future reference
				$ret = self::unserializeIfNeeded($ret->getSettingValue());
				self::$data[$name] = $ret;
				return $ret;
			} else if (isset(self::$defaults[$name])) {
				return self::$defaults[$name];
			} else {

				return $default;
			}
		}


	}

	public static function listLike($glob) {

		// Step 1: Gather from the DB
		$list = SettingPeer::FindSettingsLike(self::SECTION_CONFIG, 0, $glob);

		$ret = array();
		foreach ($list as $key => $value) {
			$ret[$key] = self::unserializeIfNeeded($value);
		}

		if (!self::$initialized) {
			self::initialize();
		}
		
		// Step 2: Gather from local config cache
		foreach (self::$data as $key => $value) {			
			if (fnmatch($glob, $key)) {
				// Only include a file config entry if it's not already there (prefer DB values)
				if (! array_key_exists($key, $ret)) {
					$ret[$key] = $value;
				}
			}
		}

		// Step 3: Gather from defaults
		foreach (self::$defaults as $key => $value) {
			if (fnmatch($glob, $key, FNM_CASEFOLD)) {
				// Only include a file config entry if it's not already there (prefer DB values)
				if (! array_key_exists($key, $ret)) {
					$ret[$key] = $value;
				}
			}
		}

		return $ret;
	}

	/**
	 * Create or update a database record for the specified setting.  All GFConf settings are stored with section = 'config', refId = 0.
	 * @param string $name The name to be used for the setting.
	 * @param mixed $value A scalar or complex value, which is serialized for storage.  The get function handles deserialization.
	 */
	public static function set($name, $value) {
		SettingPeer::SaveSetting(self::SECTION_CONFIG, 0, $name, serialize($value), 0);
		self::$data[$name] = $value;
	}

	public static function delete($name) {
		SettingPeer::DeleteSetting(self::SECTION_CONFIG, 0, $name);
		if (isset(self::$data[$name])) {
			unset(self::$data[$name]);
		}
	}
	
	/**
	 * Set a configuration variable for this request.
	 * DO NOT INVOKE THIS FUNCTION. Used only for testing purposes (i.e. testing different configurations without
	 * rebuilding the cache while performing unit tests)
	 */
	public static function _set($name, $value) {

		if (!self::$initialized) {
			self::initialize();
		}
		
		self::$data[$name] = $value;
	}
	
	/**
	 * Whether a configuration variable exists or not
	 */
	public static function varExists($name) {
		if (!self::$initialized) {
			self::initialize();
		}

		return (null !== self::get($name));
	}
	

	/**
	 * Returns the current plugins directory
	 */
	private static function getPluginsDir() {
		$mainConfigFile = self::getMainConfigFile();
		include($mainConfigFile);
		if (!array_key_exists('pluginsPath', $config)) {
			throw new Exception('"pluginsPath" variable not set in main configuration file');
		}
		$pluginsPath = $config['pluginsPath'];
		if (!is_dir($pluginsPath)) {
			throw new InvalidFileException($pluginsPath);
		}
		return $pluginsPath;
	}


	/**
	 * Return config files array
	 */
	public static function getConfigArray($file='') {
		if (!is_file($file)) {
			return array();
		}
		return self::_parseCache($file);
	}


	/**
	 * Return plugins config files array
	 */
	public static function getPluginsConfigArray() {		
		// Search for the plugins' configuration files
		$pluginsPath = self::getPluginsDir();
		$dh = opendir($pluginsPath);

		while ($pluginName = readdir($dh)) {
			if (preg_match('/^\\./', $pluginName)) {
				continue;
			}
			$gforge_etc_dir = getenv('GFORGE_ETC_DIR');
			if (empty($gforge_etc_dir)) {
				$gforge_etc_dir='/etc/gforge';
			}
			$confFile = $gforge_etc_dir.'/plugins/'.$pluginName.'/'.$pluginName.'.conf';
			if (file_exists($confFile)) {
				$configMap[$pluginName] = $confFile;
			}
		}
		
		$cacheContents = array();
		foreach ($configMap as $class => $file) {
			$cacheContents[$class] = self::_parseCache($file);
		}
		return $cacheContents;
	}
	
	/**
	 * Return a list of all the plugins in the code folder
	 */
	public static function getAllPluginsConfigs($folder = false) {

		if (! $folder) {
			// Why does this function require a different folder than
			// the one returned by getPluginsDir?  Oh, well.
			$folder = dirname(self::getPluginsDir());
		}

		$configMap = array();
		// Search for the plugins' configuration files
		$pluginsPath = $folder.'/plugins';
		$dh = opendir($pluginsPath);

		while ($pluginName = readdir($dh)) {
			if (preg_match('/^\\./', $pluginName)) {
				continue;
			}
			$confFile = $folder.'/plugins/'.$pluginName.'/conf/'.$pluginName.'.conf';
			if (file_exists($confFile)) {
				$configMap[$pluginName] = $confFile;
			}
		}
		
		return $configMap;
	}

	public static function getDefaults() {

		$defaults = array();

		//where gforge php code will be
		$defaults['libDir']='/opt/gforge';
		//where gforge config will be
		$defaults['etcDir']='/etc/gforge';
		//where gforge data will be
		$defaults['varDir']='/var/lib/gforge';

		//homedir prefix
		$defaults['homeDir']='/home';

		//availability of the users tab
		$defaults['usersTabAvailable'] = false;

		//where language and config caches will be
		$defaults['cacheDir'] = $defaults['varDir'].'/caches';
		//where the gforge filesystem will store files
		$defaults['filesPath'] = $defaults['varDir'].'/filesystem';
		//where the plugins will exist
		$defaults['pluginsPath']=$defaults['libDir'].'/plugins';

		//session defaults
		$defaults['defaultLanguage'] = 'en';
		$defaults['defaultLanguageId'] = '1';
		$defaults['sessionTimeout'] = 60*60*24*7;
		$defaults['sessionKey'] = 'This should be a long and random text';
		$defaults['defaultTheme'] = 'gforge5';
		$defaults['defaultThemeId'] = '1';
		//Default time zone for new registered users
		$defaults['defaultTimeZone'] = 'UTC';

		$defaults['apacheUser']='apache';
		$defaults['apacheGroup']='apache';
		$defaults['name']='MyGforge';
		$defaults['hostName'] = `hostname -f`;
		$defaults['use_ssl'] = false;
		$defaults['dbtype'] = 'postgres'; // postgres | mysql | oracle

		// Whether to use a full "from" address on emails sent by the server
		// (qmail may have problems with full addresses. If it doesn't like the addresses,
		// set this to true)
		$defaults['simpleFromEmail'] = false;

		//only site admin register projects?
		$defaults['projectRegRestricted'] = false;

		//trove configuration
		//enter the trove_cat_id of the root elements
		$defaults['troveDevStatus']=6;
		$defaults['troveLicense']=13;
		$defaults['troveOS']=199;
		$defaults['troveSpokenLanguage']=274;
		$defaults['troveProgrammingLanguage']=160;
		$defaults['troveTopic']=18;
		$defaults['troveIntendedAudience']=1;
		$defaults['extraTroveRootCategories']=array();

		//License enforcement
		$defaults['licenseType']='advanced';
		$defaults['licenseMaxUsers']=15;
		$defaults['licenseMaxMembers']=0;
		$defaults['licenseMaxProjects']=0;
		$defaults['licenseExpire']=0;
		$defaults['licenseKey']='ecc49e91d7abd33b4ceb9bbeb58a7ee34aa9b9b2';

		// ldap authentication
		$defaults['usesLdap'] = false;

		// Whether to use more than one SCM server
		$defaults['splitScmServers'] = false;

		// New users need to be approved?
		$defaults['approveUsers'] = false;

		//text indexing path to "pdftotext" command
		$defaults['pdftotextPath']='/usr/bin/';

		// Path to the doxygen code indexer
		$defaults['doxygenPath'] = '/usr/bin';

		//text indexing path to "wvText" command
		$defaults['wvTextPath']='/usr/bin/';

		//NIS - if the machine has accounts created with NIS
		//if so, local users and passwords will not be created in /etc, 
		//but /etc/groups will be created
		$defaults['usesNis'] = false;

		//redirect downloads - all downloads will be redirected
		//to a separate vhost which will free up the php process and save memory
		//the download vhost should have /var/lib/gforge/filesystem as the / directory
		$defaults['redirectDownloads'] = false;
		$defaults['redirectDownloadHost'] = 'http://download.'.$defaults['hostName'];

		//used to validate usernames - rules may vary by organization
		$defaults['usernameregex']='^[a-z0-9_.-]{3,15}$';

		//used to validate passwords - rules my vary by organization
		$defaults['passwordregex']='^.{6,30}$';

		//additional ftp directories

		$defaults['ftp_upload'] = false;
		$defaults['ftp_upload_dir'] = '/var/tmp/uploads/'; // must end with "/"

		$defaults['projectsVHostDir'] = '/home/groups/'; // must end with "/"

		//mandate that everyone must login except when viewing homepage
		$defaults['forceAllLogin'] = false;

		// Export/Import configuration
		$defaults['exportProjects'] = $defaults['varDir'] . '/exportProjects';

		//option to display a click-thru license on project registration page
		//the text for this license should be in /etc/gforge/PROJECT_LICENSE
		$defaults['projectRegClickLicense']=false;

		//option to display a click-thru license on user registration page
		//the text for this license should be in /etc/gforge/USER_LICENSE
		$defaults['userRegClickLicense']=false;

		//single sign-on options
		$defaults['useSSO'] = false;
		//the name of the variable in $_SERVER
		$defaults['SSOServerVar'] = 'REMOTE_USER';
		//in some cases, the server var is like an email address 
		//and you need only the first part
		$defaults['SSOServerVarSplitChar'] = '';
		//in some cases, SiteMinder does not present the HTTP_SM_USER var
		$defaults['SSOServerVarAlwaysPresent'] = true;
		//the SSOServerVar can be in either user.unix_name or user.external_id
		$defaults['SSOIDMatchesUnixName'] = true;
		//in many SSO configurations, no password is stored in the database
		$defaults['SSORequireStoredPassword'] = false;

		//this value prevents search engines from indexing SCM full structure
		$defaults['indexSCM'] = false;

		//Enable captcha on user registration form
		$defaults['useCaptcha'] = false;

		// Whether the system uses an external NSS (Name Service Switch) that interacts with GForge (e.g. nss-pgsql)
		$defaults['useNSS'] = false;

		//Enable site wide project join requests
		$defaults['allowProjectJoinRequests'] = true;

		//Show developers on each project summary page, default true
		$defaults['showDevelopersOnProjectSummary'] = true;

		//User Gforge daemon instead of cronjobs for scm and mailman
		$defaults['useDaemon'] = false; 

		//Default whitelist for the html editor, basic and full
		$defaults['basicHTMLWhiteList'] = 'p,a[href],em,u,strike,strong,br,div';
		$defaults['fullHTMLWhiteList'] = '*[style],p,span,a[href|name|id],em,u,strike,strong,sub,sup,ul,li,ol,blockquote,img[alt|src|width|height],table[align|center|border|cellpadding|cellspacing|width|summary],caption,colgroup,col[width],thead,tbody,tr,td[colspan|rowspan],th[scope|colspan],hr,div,h1,h2,h3,h4,h5,h6,pre,address,br';

		//Sets the password aging time, in days
		$defaults['user_password_aging'] = 0;

		//Show users avatars?
		$defaults['displayAvatar']=true;

		//API Key for Google Safe Browsing
		$defaults['safeBrowsingKey']='';

		//Allow outbound connections
		$defaults['outboundConnections']=true;

		//Set who can see user profile (0: visible to everyone, including anonymous users, 1: visible to any logged user, 2: visible to site admins and project admins only
		$defaults['profile_info_level']=1;

		//Sets the default maximum tries a user has to login before they are locked out. 0 means lockouts are disabled.
		$defaults['maxLoginAttempts'] = 5;

		//Enforce cookies security. If set to true, cookies will only be allowed under HTTPS. Warning: if you enable this feature, your system will be more secure, but login won't work on HTTP (only on HTTPS).
		$defaults['secureCookies'] = false;

		//Sets the configuration of the trove UI. (0: disabled, 1: enabled, but optional, 99: enabled). Default is 1
		$defaults['troveEnabled'] = 1;

		//Sendmail path
		$defaults['sendmailPath'] = '/usr/bin/sendmail';

		$defaults['autoApproveProjects']=true;

		$defaults['expireOnLoginTime'] = false;

		$ret = array();
		foreach ($defaults as $key => $value) {
			$ret["system.{$key}"] = $value;
		}

		return $ret;
	}
}
