<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Tim Perdue tim@gforge.org
 * @date 2006-02-27
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('exceptions/InvalidFileException.php');
require_once('StringUtils.php');

/**
 * This class parses an email with its headers
 */
class MailParser {
	const CONTENT_TYPE_TEXT = 'text/plain';
	const CONTENT_TYPE_HTML = 'text/html';
	const DEFAULT_CHARSET = 'ASCII';
	const DEFAULT_ENCODING = '7bit';
	
	private $lines = array();
	private $body = null;
	private $bodyContentType = null;
	private $bodyCharset = null;
	private $bodyEncoding = null;
	private $mainHeaders = array();
	private $parts = array();
	private $from;
	private $content_type;
	private $is_valid = true;
	private $parseAttachments;
	private $attachments = array();
	private $registeredHeaders = array();
	public  $invalid_message ='';

	/**
	 * @param mixed $input If it's a string, it is the ile to read the email from. If it's an
	 *               array, it assumes it is the lines of the email.   If it's not defined,
	 *               email is read from stdin.
	 *
	 * @throws PermissionDeniedException
	 */
	function __construct($input = 'php://stdin') {
		if (is_array($input)) {
			$this->lines = $input;
		} else {
			$this->lines = file($input);
			if ($this->lines === false) {
				throw new PermissionDeniedException($input);
			}
		}

		$this->registeredHeaders = array('Date','From','Sender','Reply-To','To', 'Cc', 'Bcc', 'Message-ID','Message-Id','In-Reply-To', 'References','Subject','MIME-Version',
		'Content-ID', 'Content-Description','Content-Transfer-Encoding','Content-Type','Content-Base','Content-Location','Content-features', 'Content-Disposition',
		'Content-Language','Content-Alternative','Content-MD5','Content-Duration', 'Mime-Version', 'Auto-Submitted','X-AutoReply','X-AutoRespond', 'Precedence', 'X-Precedence', 'X-Auto-Respond-Suppress');
			$this->parse();
		}

	private function isRegisteredHeader($string) {
		$headers = $this->registeredHeaders;

		foreach ($headers as $header) {
			if(preg_match('/^'.strtolower($header).':/', strtolower($string))) {
				return $header;
			}
		}

		return null;
	}
	
	private function parse() {
		// First, parse the main headers
		$i = 0;
		$count = count($this->lines);
		$skip = false;
		while ($i < $count) {
			$line = $this->lines[$i];

			if ($skip) {
				$skip = false;
				continue;
			}

			if (!is_null($matched_header = $this->isRegisteredHeader($line)) ) {
				if (preg_match('/;$/',$line)) {
					//A ; at the end of the line means the line is splitted across several lines (I hope)
					$tmp = trim($line) . $this->lines[$i+1];
					$this->mainHeaders[strtolower($matched_header)] = imap_mime_header_decode($tmp);
					$skip = true;
				} else if ($matched_header == 'From' && preg_match('/^ <.*>$/', $this->lines[$i+1])) {
					//Thunderbird breaks the the From header, sigh
					$tmp = trim($line). $this->lines[$i+1];
					$this->mainHeaders[strtolower($matched_header)] = imap_mime_header_decode($tmp);
					$skip = true;
				} else {
					$this->mainHeaders[strtolower($matched_header)] = imap_mime_header_decode($line);
				}

			}

			if (preg_match('/^$/', $line)) {
				//End of main headers
				//We got all main headers, stop parsing, headers between boundaries will be added to the parts array
				break;
			}

			$i++;
		}

		if (!$this->isMime()) {
			$this->parseNonMime($i);

		} else if (!$this->isMimeMultipart()) {
			$this->parseSinglePartMime($i);
		} else {
			 $this->parseMultipartMime($i);
		}

	}

	private function isMime() {
		$header = $this->getMainHeader('mime-version');
		if ($header) {
			return true;
		} else {
			return false;
		}
	}

	private function parseNonMime($start) {

		$c = count($this->lines);
		for ($i = $start ;$i < $c; $i++) {
			$this->body .= $this->lines[$i];
		}
		$this->bodyContentType = self::CONTENT_TYPE_TEXT;
		$this->bodyCharset = self::DEFAULT_CHARSET;

	}

	public function getMainHeader($string) {
		if (array_key_exists($string, $this->mainHeaders)) {
			return $this->mainHeaders[$string];
		}
		return false;
	}

	public function isMimeMultipart() {
		$header = $this->getMainHeader('content-type');
		$value = $header[0]->text;

		if (strstr($value, 'multipart/mixed')
			|| strstr($value, 'multipart/alternative')
			|| strstr($value, 'multipart/signed')
			|| strstr($value, 'multipart/related')
			|| strstr($value, 'multipart/plain')) {

			return true;
		}

		return false;
	}

	private function parseSinglePartMime($start_line) {
		$c = count($this->lines);
		// Regular text message... just read the body
		for ($i=$start_line; $i < $c; $i++) {
			$this->body .= $this->lines[$i];
		}

		$header = $this->getMainHeader('content-type');
		if (!$header) {
			syslog(LOG_WARNING, 'GForge: '.__FUNCTION__.' content-type header not set in single part mime email,assuming defaults');
			$this->bodyContentType = self::CONTENT_TYPE_TEXT;
			$this->bodyCharset = self::DEFAULT_CHARSET;
			$this->bodyEncoding = self::DEFAULT_ENCODING;

			return;
		}
		$values = explode(';', $header[0]->text);

		$this->bodyContentType = trim(preg_replace('/Content-Type:/','',$values[0]));
		//Search for the charset, if defined
		$count = count($values);
		for ($i = 1; $i < $count; $i++) {
			if (strstr($values[$i], 'charset')) {
				$charset = explode("=", $values[$i]);
				$this->bodyCharset = trim($charset[1]);
				break;
			}
		}
		if (is_null($this->bodyCharset)) {
			$this->bodyCharset = self::DEFAULT_CHARSET;
		}

		$header = $this->getMainHeader('content-transfer-encoding');
		if ($header) {
			$this->bodyEncoding = array_shift($this->getHeaderValues($header));
		} else {
			$this->bodyEncoding = self::DEFAULT_ENCODING;
		}

		return;
	}

	private function parseMultipartMime($start_line) {
		$boundaries = array();

		$contentType = $this->getMainHeader('content-type');
		if (preg_match('/boundary="?([^"]*)"?/i',$contentType[0]->text, $matches)) {
			$boundaries[$matches[1]] = $start_line;
		}

		$count = count($this->lines);

		//Parse other boundaries
		for ($i = $start_line + 1; $i < $count; $i++) {
			if (preg_match('/boundary="?([^"]*)"?/i',$this->lines[$i], $matches)) {
				$boundaries[$matches[1]] = $i;
			}
		}

        $failedMail = empty($boundaries);
        if (!$failedMail) { // mail has boundaries, fetch parts
	        try {
		        foreach ($boundaries as $boundary => $line) {
                    $this->parseParts($boundary);
		        }
            } catch (Exception $e) {// mail boundaries corrupted
                $failedMail = true;
            }
        }
        if ($failedMail) {
            $this->body = '[MailParser: couldn\'t parse MIME email (1)';
            $this->invalid_message = $this->body;
            $this->is_valid = false;
            $this->parts = array();
        }

		//Parse the body
		foreach ($this->parts as $part) {
			$values = $this->getHeaderValues($part['headers']['content-type']);
			if (array_key_exists(self::CONTENT_TYPE_TEXT, $values)) {
				if (array_key_exists('charset', $values)) {
					$this->bodyCharset = trim($values['charset']);
				} else {
					$this->bodyCharset = 'ASCII';
				}
				if (isset($part['headers']['content-transfer-encoding'])) {
					$encodingValue = $this->getHeaderValues($part['headers']['content-transfer-encoding']);
					$this->bodyEncoding = trim(array_shift($encodingValue));
				} else {
					$this->bodyEncoding = '7bit';
				}
				$this->body = $part['content'];
				break;
			}
		}


		if (is_null($this->body)) {
			//No text part, crappy email client
			foreach ($this->parts as $part) {
				$values = $this->getHeaderValues($part['headers']['content-type']);
				if (array_key_exists(self::CONTENT_TYPE_HTML, $values)) {
					if (array_key_exists('charset', $values))
						$this->bodyCharset = trim($values['charset']);
				} else {
					$this->bodyCharset = 'ASCII';
				}
				if (isset($part['headers']['content-transfer-encoding'])) {
					$encodingValue = $this->getHeaderValues($part['headers']['content-transfer-encoding']);
					$this->bodyEncoding = trim(array_shift($encodingValue));
				} else {
					$this->bodyEncoding = '7bit';
				}
				$this->body = $part['content'];
				break;
			}
		}

		if (is_null($this->body)) {
			$this->invalid_message = true;
			$this->body = 'Error parsing MIME: no body found!';
		}

	}


	private function getHeaderValues($header) {
		$values = array();
		$ret = array();
		$subVal = explode(':', $header[0]->text);
		$values = explode(';', trim($subVal[1]));


		foreach ($values as $value) {
			$subVal = explode('=', trim($value));
			if (isset($subVal[1])) {
				$ret[$subVal[0]] = $subVal[1];
			} else {
				$ret[$value] = $value;
			}
		}

		return $ret;

	}


	private function parseParts($separator) {
		$chunks = array();
		$ret = array();
		$first_marker = 0;
		$last_marker = 0;

		//Find start and ending delimiters
		$count = count($this->lines);
		for ($i = 0; $i < $count; $i++) {
			if(preg_match('/^--'.preg_quote(trim($separator),'/')."$/",$this->lines[$i])) {
				$first_marker = $i;
				break;
			}
		}
		for ($i = $first_marker; $i < $count; $i++) {
			if (preg_match('/^--'.preg_quote(trim($separator),'/').'--$/',trim($this->lines[$i]))) {
				$last_marker = $i;
				break;
			}
		}

		if (!$first_marker || !$last_marker) {
			throw new Exception("Boundaries were not found in email!");
		}

		//Now, check if we have nested parts
		$parts = array();

		for ($i = $first_marker+1; $i < $last_marker-1; $i++) {
			if(preg_match('/^--'.preg_quote($separator,'/').'$/', $this->lines[$i])) {
				$parts[] = $i;
			}
		}


		//Got all the delimiters, parse the text between
		if (count($parts)) {
			//Add the start and end markers
			array_unshift($parts,$first_marker);
			array_push($parts, $last_marker);

			$begin = array_shift($parts);
			while ($end = array_shift($parts)) {
				$chunks[] = trim(implode(array_slice($this->lines, $begin+1, $end-$begin-1, true)));
				$begin = $end;
			}
		} else {
			$chunks[] = trim(implode(array_slice($this->lines, $first_marker+1, $last_marker-$first_marker-1, true)));
		}
		foreach ($chunks as $chunk) {
			$chunk = explode(PHP_EOL, $chunk);
			$skip = false;
			for ($i=0; $i < count($chunk); $i++) {
				if ($skip) {
					$skip = false;
					continue;
				}
				if (!is_null($matched_header = $this->isRegisteredHeader($chunk[$i])) ) {
					if (preg_match('/^\s.*=/',$chunk[$i+1])) {
						//Special case, some email clients will continue the header in the next line
						$tmp = trim($chunk[$i]) . $chunk[$i+1];
						$part['headers'][strtolower($matched_header)] = imap_mime_header_decode($tmp);
						$skip = true;
					} else {
						$part['headers'][strtolower($matched_header)] = imap_mime_header_decode($chunk[$i]);

					}
				} else if (preg_match('/^$/', $chunk[$i])) {
					//End of headers
					$content_index = $i;
					break;
				};


			}
			$part['content'] = implode(PHP_EOL,array_slice($chunk,$content_index,count($chunk) - $content_index, true));
			$this->parts[] = $part;

		}

	}
	
	/**
	 * Returns false if the email had parsing errors
	 */
	public function isValid() {
		if (!$this->getMainHeader('date') || !$this->getMainHeader('from') ||
			empty($this->body)) {
			$this->is_valid = false;
			$this->invalid_message = 'Empty date, sender, or body';
		}
		return $this->is_valid;
	}
	
	/**
	 * Returns the e-mail address of the sender
	 */
	public function getFrom() {
		$from = $this->getMainHeader('from');
		//Several lines, not sure why imap_decoder_header does this
		$from = trim($from[count($from)-1]->text);

		if (preg_match('/[<]([^@]+@.+)[>]/', $from, $matches)) {
			$from = $matches[1];
		} else if (preg_match('/[(<]([^@]+@.+)[)>]/', $from, $matches)) {
			// $from is something like "John Doe" <john@doe.com>
			$from = $matches[1];
		} else if (preg_match('/([^@[:blank:]]+@[^[:blank:]]+)/', $from, $matches)) {
			// $from is something like:
			// * john@doe.com
			// * john@doe.com (johndoe)
			$from = $matches[1];
		} else {
			// couldn't parse the From header -- shouldn't happen
			$from = '';
		}
		
		return $from;
	}

    public function getTo() {
        $full_to = explode(':',$this->getFullTo());
        $start = strpos($full_to[1], '<');
        if ($start === false) {
            return trim($full_to[1]);
        }
        $address = substr($full_to[1], $start, -1);

        return trim(preg_replace('/[<>]/','',$address));
    }

	public function getFullTo() {
        return $this->mainHeaders['to'][0]->text;
	}
	/**
	 * Returns the full "From:" string, not just the e-mail address
	 */
	public function getFullFrom() {
		$tmp = $this->getMainHeader('from');
		$from = '';
		if (count($tmp) > 1) {
			$i = count($tmp);
			for ($x = 0; $x < $i; $x++) {
				$from .= $tmp[$x]->text;
			}
		} else {
			$from = $tmp[0]->text;
		}

		return trim($from);

	}

	public function getBody() {
		$body = $this->body;
		if (strtolower($this->bodyEncoding) == 'quoted-printable') {
			$body = quoted_printable_decode($body);
		}
		if ($this->bodyEncoding == 'base64') {
			$body = base64_decode($this->body);
		}

		if(strtolower($this->bodyCharset) != 'utf-8') {
	 	   $body = iconv($this->bodyCharset, 'UTF-8', $body);
		}

		//Gmail introduces inline markers, let's remove it
		$body = preg_replace('/\[image:.*\]/','',$body);
		return $body;

	}

	public function getSubject() {
		$tmp = (isset($this->mainHeaders['subject'])) ? $this->mainHeaders['subject'] : '(no subject)';
		$subject = '';
		if (is_array($tmp)) {
			foreach($tmp as $part) {
				$subject .= $part->text;
			}
		} else {
			$subject = $tmp;
		}
		
		return trim(str_replace('Subject:','',$subject))   ;
	}
	
	public function getMessageId() {
		if (isset($this->mainHeaders['message-id'])) {
			return $this->mainHeaders['message-id'][0]->text;
		} else {
			return 0;
		}
	}

	public function getMessageIDFromAddress() {
		list($generic_addr, $tmp) = explode('+', $this->getTo());
		list($message_id, $hostname) = explode('@', $tmp);

		return $message_id;
	}

	
	/**
	 * Returns the UNIX timestamp of the date of the email
	 * @return	int
	 */
	public function getDate() {
		$date_str = preg_replace('/^Date\:\s*/i', '', $this->mainHeaders['date'][0]->text);
		$timestamp = strtotime($date_str);
		return $timestamp;
	}
	
	//public function getContentType() {
	//	return $this->content_type;
	//}


	function senderIsGmail() {
		if (strpos($this->getMessageId(), 'mail.gmail.com') ) {
			return true;
		}
		return false;
	}

	public function getBodyQuoteRemoved($quoteChar = '>', $body=null) {
		if ( is_null($body) ) {
			$body = $this->getBody();
		}


		//Is this outlook?
		if(preg_match('/^-{5,5}Original Message/m', $body)) {
			return trim(StringUtils::getLinesBeforeMarker($body, '-----Original Message-----'));
		}

		$result = explode(PHP_EOL, $body);

		$lines = count($result);

		//Search for the first quoted line
		$firstQuote = 0;
		for ($i = 0; $i < $lines; $i++){
			if (preg_match('/^'.preg_quote($quoteChar).'+.*$/s', $result[$i])) {

				$firstQuote = $i;
				break;
			}
		}

		//Delete the user wrote line (if exists)
      if ( isset($result[$firstQuote - 1]) && preg_match('/^.*<?.*@.*>?.*:/s', $result[$firstQuote -1])) {
	      unset($result[$firstQuote-1]);
	      //Some clients will use two lines
         if ( isset($result[$firstQuote - 2]) && !empty($result[$firstQuote-2]) ) {
	         unset($result[$firstQuote-2]);
          }
       }

		//Some others will put an empty line between firstQuote and user wrote message
		if ( isset($result[$firstQuote - 2]) && preg_match('/^.*<?.*@.*>?.*:/s', $result[$firstQuote -2])) {

            unset($result[$firstQuote-2]);
         //Some clients will use two lines
         if ( isset($result[$firstQuote - 3]) && !empty($result[$firstQuote-3]) ) {
				unset($result[$firstQuote-3]);
          }
      }

		//Finally, remove all the quoted lines

		for ($i = 0; $i < $lines; $i++) {
			if (isset($result[$i]) && preg_match('/^'.preg_quote($quoteChar,'/').'.*$/s', $result[$i]) ) {
				unset($result[$i]);
			}
            //Check if this is a reply to a notification and remove the line
            if (isset($result[$i]) && preg_match('/<noreply/', $result[$i])) {
                unset($result[$i]);
            }
		}



		$result = implode(PHP_EOL, $result);

		$result = $this->removeSignature($result);
		$result = preg_replace('/=$/',null,$result);
		$result = trim($result);
		//Hotmail adds a = to the end of the quotation
		return $result;

	}

	public function removeSignature($text) {
		return preg_replace ('/[\n\r]-- .*$/s', null, $text);
	}

	public function getAttachments() {
		$ret = array();
		$i = 0;
		foreach($this->parts as $part) {
			$i++;
			if (array_key_exists('content-disposition',$part['headers'])) {
				$values = $this->getHeaderValues($part['headers']['content-disposition']);
				if (array_key_exists('attachment', $values)) {
					$ret[] = $part;
				}
			} else {
				//inline attachment?
				$values = $this->getHeaderValues($part['headers']['content-type']);
				foreach ($values as $value) {
					if (preg_match('/image\//',$value)) {
						if (!array_key_exists('name',$values)) {
							//Use a generic name
							$part['headers']['name'] = 'attachment_'.$i;
						} else {
							//Grab the name
							$part['headers']['name'] = $values['name'];
						}
						$ret[] = $part;
					}
				}
			}
		}

		return $ret;

	}

    public function getRawMail() {
        return $this->lines;
    }

	public static function getAttachmentTmpPath() {
		return '/tmp/gforge-tmp-attachment';
	}

    public function dumpRawEmail() {
        $path = GFConf::get('system.varDir');
        $path .= DIRECTORY_SEPARATOR.'logs';
        $filename = 'email-debug-'.date('Ymd-His-u').'.txt';
        $dest = $path.DIRECTORY_SEPARATOR.$filename;
        SystemUtils::writeFile($dest,$this->lines);
    }

    public function getBodyCharset() {
        return $this->bodyCharset;
    }

    public function getBodyContentType() {
        return $this->bodyContentType;
    }

	/**
	 * Check headers to determine if the email is an auto-response
	 * @return bool
	 */
	public function isAutoResponse() {
		if (isset($this->mainHeaders['auto-submitted']) && $this->mainHeaders['auto-submitted'][0]->text != 'no') {
			return true;
		}

		if (isset($this->mainHeaders['precedence'])  && ($this->mainHeaders['precedence'][0]->text == 'auto_reply'
		 || $this->mainHeaders['precedence'][0]->text == 'bulk' || $this->mainHeaders['precedence'][0]->text == 'junk')) {
			return true;
		}

		if (isset($this->mainHeaders['x-precedence'])  && ($this->mainHeaders['x-precedence'][0]->text == 'auto_reply'
				|| $this->mainHeaders['x-precedence'][0]->text == 'bulk' || $this->mainHeaders['x-precedence'][0]->text == 'junk')) {
			return true;
		}

		if (isset($this->mainHeaders['x-autoreply']) || isset($this->mainHeaders['x-autorespond'])) {
			return true;
		}

		if (isset($this->mainHeaders['x-auto-response-suppress'])) {
			return true;
		}

		return false;
	}
}
