<?php
/**
 * GForge 5.0 RSS abstract class
 *
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Daniel Perez	daniel@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * http://blogs.law.harvard.edu/tech/rss
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

class RssGenerator {
	// variables and some defaults
	var $channels = array(); // default properties (title, link, description), items(array) collection of items, and optional properties
	var $version = '2.0';
	var $encoding = 'ISO-8859-1';

	/**
	 * change the rss version
	 *
	 * @param unknown_type $version
	 */
	public function setRssVersion($version) {
		$this->version = $version;
	}

	/**
	 * change the encoding
	 *
	 * @param unknown_type $encoding
	 */
	public function setEncoding($encoding) {
		$this->encoding = $encoding;
	}

	/**
	 * get the version tag
	 *
	 * @return unknown
	 */
	public function getRssVersionTag() {
		return '<rss version="'.$this->version.'">';
	}

	/**
	 * Get the xml header of the file
	 *
	 * @return unknown
	 */
	private function getHeader() {
		return '<?xml version="1.0" encoding="'.$this->encoding.'"?>';
	}

	/**
	 * returns the full rss feed from the generator
	 *
	 * @return string
	 */
	public function getRss() {
		$result = $this->getHeader() . "\n";
		$result .= $this->getRssVersionTag() . "\n";
		foreach ($this->channels as $channel) {
			$result .= $this->getChannelTag($channel['title']) . "\n";
		}
		$result .= '</rss>' . "\n";
		return $result;
	}


	public function getChannelTag($channel) {
		$return = '';
		if (array_key_exists($channel,$this->channels)) {
			$ch = $this->channels[$channel];
			$return .= "<channel>\n";
			$return .= '<title>'.$ch['title']."</title>\n";
			$return .= '<link>'.$ch['link']."</link>\n";
			$return .= '<description>'.$ch['description']."</description>\n";
			if (array_key_exists('optionals',$ch) && count($ch['optionals']>0)) {
				// add the optional tags
				$optionals = $ch['optionals'];
				reset($optionals);
				while ($val = current($optionals)) {
					switch (key($optionals)) {
						case 'language' :
						case 'pubDate' :
						case 'lastBuildDate' :
						case 'copyright' :
						case 'managingEditor' :
						case 'webMaster' :
						case 'category' :
						case 'generator' :
						case 'docs' :
						case 'ttl' : {
							$return .= '<'.key($optionals).'>'.$val.'</'.key($optionals)."\n";
							break;
						}
						case 'image' : {
							$return .= '<image><url>'.$val['url'].'</url><title>'.$val['title'].'</title><link>'.$val['link'].'</link>'."</image>\n";
							break;
						}
						case 'textInput' : {
							$return .= '<textInput><title>'.$val['title'].'</title><description>'.$val['description'].'</description><name>'.$val['name'].'</name><link>'.$val['link'].'</link>'."</textInput>\n";
							break;
						}
					}
					next($optionals);
				}
			}
			// now the items
			if (isset($ch['items']) && count($ch['items'])>0) {
				foreach ($ch['items'] as $item) {
					$return .= $this->getItemTag($item) . "\n";
				}
			}
			$return .= "</channel>\n";
		}
		return $return;
	}

	public function getItemTag($item) {
		reset($item);
		$return = '<item>';
		while ($val = current($item)) {
			$tag = key($item);
			$return .= "<$tag>$val</$tag>";
			next($item);
		}
		$return .= '</item>';
		return $return;
	}

	/**
	 * Add a new channel to the rss feed (usually it`s just one channel)
	 *
	 * @param The name of the channel
	 * @param The URL to the HTML website corresponding to the channel
	 * @param Phrase or sentence describing the channel
	 * @param $optional (attribute->value) array. attributes : language,pubDate,lastBuildDate,image(url,title,link),textInput(title,description,name,link)
	 * @param array containing items.
	 */
	public function addChannel($title,$link,$description,$optional=array(),$items=array()) {
		$data = array();
		$data['title'] = $title;
		$data['link'] = $link;
		$data['description'] = $description;

		if (count($optional)>0) {
			reset($optional);
			while ($val = current($optional)) {
				switch (key($optional)) { // let`s see which attributes from rss channel are here
					case 'language' :
					case	'pubDate' :
					case	'lastBuildDate' :
					case 'copyright' :
					case 'managingEditor' :
					case 'webMaster' :
					case 'category' :
					case 'generator' :
					case 'docs' :
					case 'ttl' : {
						$data['optionals'][key($optional)] = $val;
						break;
					}
					case 'image' : {
						$imgarr = array();
						$imgarr['url'] = $val['url'];
						$imgarr['title'] = $val['title'];
						$imgarr['link'] = $val['link'];
						$data['optionals']['image'] = $imgarr;
						break;
					}
					case 'textInput' : {
						$textarr = array();
						$textarr['title'] = $val['title'];
						$textarr['description'] = $val['description'];
						$textarr['name'] = $val['name'];
						$textarr['link'] = $val['link'];
						$data['optionals']['textInput'] = $textarr;
						break;
					}
				}
				next($optional);
			}
		}
		$this->channels[$title] = $data; // we identify the channel by its name
		if (count($items)>0) {
			// add the items
			foreach ($items as $item) {
				$this->addItem($title,$item);
			}
		}
	}

	/**
	 * adds an item to the channel specified by the name
	 *
	 * @param string $channel : the channel name
	 * @param array $values . attributes : title,link,description,author,comments,guid,pubDate
	 */
	public function addItem($channel,$values) {
		if (array_key_exists($channel,$this->channels)) {
			//  All elements of an item are optional, however at least one of title or description must be present
			if ( (!array_key_exists('title',$values)) && (!array_key_exists('description',$values)) ) {
				return;
			}
			$item = array();
			reset($values);
			while ($val = current($values)) {
				$keyname = key($values);
				if ($keyname == 'link') {
					$val = str_replace('&amp;', '&', $val);
				}
				$item[$keyname] = $val;

				next($values);
			}
			$this->channels[$channel]['items'][] = $item;
		}
	}
}


?>
