<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author	Marcelo Mottalli	<mottalli@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
require_once('core/om/BaseLanguage.php');
require_once('plugin/ActionManager.php');

require_once('GFConf.php');
require_once('exceptions/InvalidDirectoryException.php');
require_once('exceptions/FilePermissionsException.php');
require_once('SystemUtils.php');
require_once('session/Session.php');
require_once('html/GFWidget.php');

class Language extends BaseLanguage implements GFWidget {
	/** 
	 * ID of the language which has all the strings 
	 * (do not confuse with "default language" given by the configuration)
	 */
	const BASE_LANGUAGE_ID = 1;	
	
	private $textArray = array();
	private $languageLoaded = false;
	/** Pointer to the base language (the language that has all the strings) */
	private $baseLanguage = null;

	/**
	 * Removes this object from datastore and sets delete attribute. Informs the plugin manager.
	 * This method was added by propel_customizer.php.
	 * 
	 * @param Connection $con
	 * @return void
	 * @throws PropelException
	 * @see Language::setDeleted()
	 * @see Language::isDeleted()
	 */
	function delete($con = null) {
		ActionManager::notify("LANGUAGE_DELETE", $this);
		parent::delete($con);
	}

	/**
	 * Stores the object in the database using the Propel methods. Informs the plugin manager.
	 * This method was added by propel_customizer.php.
	 *
	 * @param Connection $con
	 * @return void
	 * @throws PropelException
	 */
	function save($con = null) {
		parent::save($con);
		ActionManager::notify("LANGUAGE_SAVE", $this);
	}

	/**
	 * Constructor
	 */
	public function __construct() {
	}
	
	
	/**
	 * Retrieve a localized text string for this language.
	 *
	 * @param	string $path	ID of the string to retrieve, in the form "page.category"
	 * @param	array $args		Array of parameters to the string
	 * @return string
	 */
	public function getText($tag, $args=array()) {
		if (!$this->languageLoaded) {
			$this->loadLanguage();
		}
		
		if (array_key_exists($tag, $this->textArray)) {
			$text = $this->textArray[$tag];
			for ($i=0; $i < sizeof($args); $i++) {
				// Replace $i but not \$i
				$text = preg_replace('/\\$'.($i+1).'/', $args[$i], $text);
			}
			// Replace \n
			$text = preg_replace('/\\\n/', "\n", $text);

			return $text;
		} else if (!is_null($this->baseLanguage)) {
			// string wasn't found in this language... try to see if it is in the default language
			return $this->baseLanguage->getText($tag, $args);
		} else {
			return "";
		}
	}

	/**
	 * Retrieves the language object for the current session
	 *
	 * @return Language
	 */
	public static function getSessLang() {
		return Session::getSession()->getLanguage();
	}
	
	/**
	 * Retrieves a translated text using the current session's language
	 *
	 * @return string
	 */
	public static function getSessText($path, $args=array()) {
		return Session::getSession()->getLanguage()->getText($path, $args);
	}
	
	/**
	 * Retrieves a string in the system's default language
	 */
	public static function getDefaultText($tag, $args=array()) {
		static $default_lang = null;
		if (is_null($default_lang)) {
			$default_lang = LanguagePeer::retrieveByPk(GFConf::get('system.defaultLanguageId'));
		}
		
		return $default_lang->getText($tag, $args);
	}
	
	/**
	 * Loads the language strings for this language
	 *
	 * @return void
	 */
	private function loadLanguage() {
		$this->readFromCache();		
		
		// If this is not the base language, get a copy of the base language.
		// We'll use this for overriding strings
		if ($this->getLanguageId() != Language::BASE_LANGUAGE_ID) {
			$this->baseLanguage = LanguagePeer::retrieveByPk(Language::BASE_LANGUAGE_ID);
		}
		
		$this->languageLoaded = true;
		
	}
	
	private function _readLanguageFiles() {
		$files = $this->searchLanguageFiles();
		foreach ($files as $file) {
			$this->importFile($file);
		}
	}
	
	/**
	 * Build the language cache
	 */
	public static function generateCache() {
		// retrieve all the languages
		$languages = LanguagePeer::doSelect(new Criteria());
		foreach ($languages as $language) {
			$language->buildCache();
		}
	}
	
	/**
	 * @todo	Add the 'encoding' field to the 'language' table and remove this method
	 * @return string
	 */
	public function getEncoding() {
		return 'UTF-8';
	}

	/**
	 * @todo	Add the 'font' field to the 'language' table and remove this method
	 * @return string
	 */
	public function getFont() {
		return 'verdana,arial,helvetica,sans-serif';
	}	
	
	
	
	/**
	 * Generates the cache for this language
	 */
	private function buildCache() {
		$files = $this->searchLanguageFiles();
		foreach ($files as $file) {
			$this->importFile($file);
		}
		
		$serializedData = serialize($this->textArray);
		
		$cacheFile = $this->getCacheFilename();
		$cacheDir = dirname($cacheFile);
		if (!is_dir($cacheDir) && !mkdir($cacheDir, 0755, true)) {
			throw new InvalidDirectoryException($cacheDir);
		}		
		
		$fh = @fopen($cacheFile, "w");
		if (!$fh) {
			throw new FilePermissionsException($cacheFile);
		}
		fwrite($fh, $serializedData);
		fclose($fh);
	}
	
	/**
	 * Searches the language files for this language
	 *
	 * @return array
	 */
	private function searchLanguageFiles() {
		$libDir = GFConf::get("system.libDir");
		$pluginsPath = GFConf::get("system.pluginsPath");
		
		if (!is_dir($libDir)) {
			throw new InvalidDirectoryException($libDir);
		} else if (!is_dir($pluginsPath)) {
			throw new InvalidDirectoryException($pluginsPath);
		}
		
		$dirs = array();
		$dirs[] = SystemUtils::buildFilePath($libDir, "/languages");
		
		// search for the "languages" dir in each plugin
		$dh = opendir($pluginsPath);
		while ($pluginDir = readdir($dh)) {
			if (preg_match("/^\\./", $pluginDir)) {		//starts with a .
				continue;
			}
			$fullDir = SystemUtils::buildFilePath($pluginsPath, $pluginDir);
			if (is_dir($fullDir)) {
				$dirs[] = SystemUtils::buildFilePath($fullDir, "/languages");
			}
		}
		
		// Now search for the existing language files
		$files = array();
		$fileName = $this->getFilename();
		
		foreach ($dirs as $dirName) {
			$fullPath = SystemUtils::buildFilePath($dirName, $fileName);
			if (file_exists($fullPath)) {
				$files[] = $fullPath;
			}
		}
		
		return $files;
	}
	
	/**
	 * Reads the language file from the cache
	 */
	private function readFromCache() {
		$cacheFile = $this->getCacheFilename();
		if (!file_exists($cacheFile)) {
			throw new Exception("Language cache not generated, run create_lang_cache script");
		}
		
		// unserialize the cache
		$fh = @fopen($cacheFile, "r");
		if (!$fh) {
			throw new FilePermissionsException($cacheFile);
		}
		$cacheData = unserialize(fread($fh, filesize($cacheFile)));
		fclose($fh);
		
		// Couldn't read the data
		if (!is_array($cacheData)) {
			throw new Exception("Language cache file invalid: ".$cacheFile);
		}
		
		$this->textArray = $cacheData;
	}
	
	/**
	 * Gets the name of the file where the localization cache for this language resides
	 *
	 * @return string
	 */	
	private function getCacheFilename() {
		$cacheDir = SystemUtils::buildFilePath(GFConf::get("system.cacheDir"), "localization-cache");
		if (!$cacheDir) {
			throw new Exception("Localization cache dir isn't defined");
		}
		$cacheFile = $this->getFilename().".cache";
		$cacheFilePath = SystemUtils::buildFilePath($cacheDir, $cacheFile);
		return $cacheFilePath;
	}
	
	/**
	 * Parses a language file
	 */
	private function importFile($fileName) {
		// Parse the whole file
		echo "Parsing: $fileName\n";
		$lines = file($fileName);
		$n = sizeof($lines);
		for ($i=0; $i < $n; $i++) {
			$line = trim($lines[$i]);
			
			// empty line?
			if (strlen($line) == 0) {
				continue;
			}

			// comment?
			if (preg_match("/^[[:blank:]]*#/", $line)) {
				continue;
			}
			
			$fields = explode("\t", $line);
			if (!isset($fields[0])) {		// no tag defined
				$errorline = $i+1;
				throw new Exception("Invalid language file: $fileName, error in line #$errorline");
			}
			$tag = $fields[0];
			
			// Now discard all the tabs and whitespaces after $tag
			// NOTE: could cause problems if $tag has some preg-unfriendly characters
			preg_match("/".$tag."[[:blank:]]*(.*)/", $line, $matches);
			if (count($matches) != 2) {
				echo "\nInvalid line #".($i+1).", text: \"$line\"\n";
				continue;
			}
			
			$string = $matches[1];
			if (array_key_exists($tag, $this->textArray)) {
				// concatenate the text if it already exists
				$this->textArray[$tag] .= "\n".$string;
			} else {
				$this->textArray[$tag] = $string;
			}
		}
	}
	
	// These functions come from GFWidget
	public function getID() {
		return $this->getLanguageId();
	}
	
	// getName is generated by propel


	/**
	 * Gets proper font type and style
	 *
	 * @return array
	 */	
	public function getFontProperties($font = '', $style = '', $size = 10) {
		require_once('lib/external/jpgraph/jpgraph.php');
		if ($font == '') {
			$font = FF_FONT1;
		}
		if ($style == '') {
			$style = FS_NORMAL;
		}

		$codes = explode('-', $this->getLanguageCode());
		foreach ($codes as $code) {
			$code = trim(strtolower($code));
			if (in_array($code, array('tw'))) {
				// Traditional Chinese
				if (file_exists(TTF_DIR.'bsmi00lp.ttf')) {
					return array(FF_CHINESE2, FS_NORMAL, $size + 1);
				} elseif (file_exists(TTF_DIR.'bkai00mp.ttf')) {
					return array(FF_CHINESE, FS_NORMAL, $size + 1);
				} elseif (file_exists(TTF_DIR.'simsun.ttf ')) {
					return array(FF_SIMSUN, FS_NORMAL, $size + 1);
				}
			} elseif (in_array($code, array('zh','cn'))) {
				// Simplified Chinese
				if (file_exists(TTF_DIR.'gbsn00lp.ttf')) {
					return array(FF_SCHINESE, FS_NORMAL, $size + 1);
				} elseif (file_exists(TTF_DIR.'gkai00mp.ttf')) {
					return array(FF_SCHINESE2, FS_NORMAL, $size + 1);
				}
			} elseif (in_array($code, array('jp', 'ja'))) {
				// Japanese
				if (file_exists(TTF_DIR.'ipamp.ttf')) {
					return array(FF_MINCHO, FS_NORMAL, $size + 1);
				} elseif (file_exists(TTF_DIR.'kochi-mincho.ttf')) {
					return array(FF_KOCHI_MINCHO, FS_NORMAL, $size + 1);
				} elseif (file_exists(TTF_DIR.'kochi-mincho-subst.ttf')) {
					return array(FF_KOCHI_MINCHO_SUBST, FS_NORMAL, $size + 1);
				} elseif (file_exists(TTF_DIR.'sazanami-mincho.ttf')) {
					return array(FF_SAZANAMI_MINCHO, FS_NORMAL, $size + 1);
				} elseif (file_exists(TTF_DIR.'ipag.ttf')) {
					return array(FF_GOTHIC, FS_NORMAL, $size + 1);
				}
			} elseif (in_array($code, array('ko'))) {
				// Korean
				if (file_exists(TTF_DIR.'gulim.ttf')) {
					return array(FF_GULIM, FS_NORMAL, $size + 1);
				} elseif (file_exists(TTF_DIR.'dotum.ttf')) {
					return array(FF_DOTUM, FS_NORMAL, $size + 1);
				} elseif (file_exists(TTF_DIR.'batang.ttf')) {
					return array(FF_BATANG, FS_NORMAL, $size + 1);
				}
			}
		}
		if (file_exists(TTF_DIR.'verdana.ttf')) {
			// Verdana
			return array(FF_VERDANA, FS_NORMAL, $size);
		} elseif (file_exists(TTF_DIR.'arial.ttf')) {
			// Arial
			return array(FF_ARIAL, FS_NORMAL, $size);
		} else {
			return array($font, $style, $size);
		}
	}

}
?>
