<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id: Layout.php 719 2006-04-04 18:45:41Z danper $
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('core/TroveLink.php');
require_once('core/TroveCategory.php');

/**
 * This class manages the Trove forms
 */
class TroveRenderer {
	private $section = null;
	private $ref_id = null;
	private $troves;
	private $trove_name_trans_keys;
	private static $_fixedCategoriesWithEnableSettings = array();
	
	public function __construct() {
		// Translation keys for fixed categories' names
		$this->trove_name_trans_keys = array(
			GFConf::get('system.troveDevStatus') => 'Trove.DevStatus',
			GFConf::get('system.troveLicense') => 'Trove.License',
			GFConf::get('system.troveOS') => 'Trove.OS',
			GFConf::get('system.troveSpokenLanguage') => 'Trove.SpokenLanguage',
			GFConf::get('system.troveProgrammingLanguage') => 'Trove.ProgrammingLanguage',
			GFConf::get('system.troveTopic') => 'Trove.Topic',
			GFConf::get('system.troveIntendedAudience') => 'Trove.IntendedAudience',
			GFConf::get('system.troveGfProjTemplates') => 'Trove.ProjectTemplates',
		);
		// Root troves
		$this->troves = TroveCategoryPeer::getRootCategoriesAsArray();

	}
	
	public function setSection($section) {
		$this->section = $section;
	}
	
	public function setRefId($ref_id) {
		$this->ref_id = $ref_id;
	}
	
	public function getSection() {
		return $this->section;
	}
	
	public function getRefId() {
		return $this->ref_id;
	}
    
	/**
	 * Returns the translated name for the trove identified by $trove_id or NULL if no translation key exists
	 * 
	 * @param int $trove_id trove identifier
	 * @return string | NULL
	 */
	private function _getTranslatedTroveName($trove_id) {
		$lang = Session::getSession()->getLanguage();
		if (array_key_exists($trove_id,  $this->trove_name_trans_keys)) {
			$name = $lang->getText($this->trove_name_trans_keys[$trove_id]);
		} else {
			$name = NULL;
		}
		return $name;
	}
	
	/**
	 * Transform a flat array of TroveCategory objects into a hierarchical array
	 * that represents the trove subtree
	 * 
	 * @param array $troves
	 * @return array
	 */
	private function _buildHierarchicalTrovesArray($troves) {
		$hierarchical_troves = array();
		foreach ($troves as $trove) {
			$parent_id = $trove->getParentTroveCategoryId();
			if(!isset($hierarchical_troves[$parent_id])) {
				$hierarchical_troves[$parent_id] = array();				
			}
			$hierarchical_troves[$parent_id][] = $trove;
		}
		return $hierarchical_troves;
	}
	
	/**
	 * Builds the name for the HTML select box for category identified by $trove_id
	 * 
	 * @param int $trove_id trove identifier
	 * @return string
	 */
	private function _buildSelectName($trove_id) {
		return 'trove-'.$trove_id;
	}
	
	/**
	 * Extracts the trove identifier out of the name of an HTML select box
	 * 
	 * @param string $select_name
	 * @return string
	 */
	private function _getTroveIdFromSelectName($select_name) {
		return str_replace('trove-', '', $select_name);
	}
	
	public function renderTroveControl($trove_category) {
		$trove_selector = '';
		$trove_id = $trove_category->getTroveCategoryId();
		$project = GFUrl::getUrl()->getProject();
		
		// Not rendering control for associating project to template trove container when the project is new (default association applies)
		// or when the project is not a template (no association applies)
		if (!$trove_category || ($trove_id == GFConf::get('system.troveGfProjTemplates') && (!$project || !$project->getIsTemplate()))) {
			return '';
		}
        
		// Dont' render control for disabled fixed categories
		$enabled = $trove_category->isEnabled();
        if ($enabled !== NULL && !$enabled) {
			return '';
        }

		// Descendant troves
		$descendants = TroveCategoryPeer::getDescendantsAsArray($trove_id);
		
		$trove_selector .= '<div><p>';
		$trove_name = $this->_getTranslatedTroveName($trove_id);
		if ($trove_name === NULL) {$trove_name = $trove_category->getName();}
		$trove_selector .= '<strong>'.$trove_name.'</strong>'."\n";
		$trove_selector .= '<br />';

		// Selected option values
		$selected_categories = array();
		if ($project) {
			$rs = TroveLinkPeer::getProjectLinksToTroveSubtreeRs($project->getProjectId(), $trove_id);
			while($rs->next()) {
				$selected_categories[] = $rs->getInt('trove_category_id');
			}			
		}

		// Disabled option values
		$disabled_categories = array();
		foreach ($descendants as $element) {
			if ($element->getAccessLevel() == 99 && !(Session::getSession()->getUser()->isSiteAdmin())) {
				$disabled_categories[] = $element->getId();
			}
		}

		// Render select box
		$res = HTMLWidget::renderArranged($this->_buildHierarchicalTrovesArray($descendants), $trove_id, 0, array(), false, 'array');
		switch ($trove_id) {
			case GFConf::get('system.troveGfProjTemplates'):
				$res = array($trove_id => array('name'=>'', 'attr'=>'')) + $res;
				$trove_selector .= HTMLWidget::selectBoxFromAssoc(
					$res, 
					$this->_buildSelectName($trove_id), 
					count($selected_categories) ? $selected_categories[0] : null,
					false,
					false
				);
				break;
			default:
				$trove_selector .= HTMLWidget::multiSelectFromAssoc($this->_buildSelectName($trove_id).'[]',$res,$selected_categories,'',false,'',false,'',20,null,false, $disabled_categories);
		}

		$trove_selector .= '</div></p>';
		return $trove_selector;
	}

    public function showTroveForm($troves = null) {
		if (is_null($troves)) {
			$troves = $this->troves;
		}

        
		foreach ($troves as $trove) {            
			echo $this->renderTroveControl($trove);
		}
	}
    
    /**
     * Ouputs the client side code required for the trove selectors to use jquery.chosen plugin
     */
    public function renderChosenJavascript() {
        $language = Language::getSessLang();
        $single_select_message = $language->getText('TroveRenderer.SingleSelectPlaceholder');
        $multi_select_message = $language->getText('TroveRenderer.MultiSelectPlaceholder');
		$gf_template_trove_id = GFConf::get('system.troveGfProjTemplates');		
		$default_trove_template_name = $this->_getTranslatedTroveName($gf_template_trove_id);		
        $js = <<<EOT
            <script>
            $(function(){
                function enhanceTroveSelect(select) {
                    var options = {};
                    if (select.length > 0) {
                        if (select.attr('multiple')) {
                            select.css('width', '50%');
                            select.attr('data-placeholder', '$multi_select_message');
                        } else {
							if (select.attr('name') == 'trove-$gf_template_trove_id') {
								select.css('width', '50%');
								options.allow_single_deselect = true;
								select.attr('data-placeholder', '$default_trove_template_name');
							} else {
								select.attr('data-placeholder', '$single_select_message');
							}							
                        }
                        select.chosen(options); 
                    }
                }

                $('select[name^="trove-"]').each(function(){enhanceTroveSelect($(this))});
EOT;
        $js .= <<<EOT
            })
            </script>
EOT;
        echo $js;
    }

	
	public function savePOSTData() {
		// make sure section and ref_id are set
		if (is_null($this->section) || is_null($this->ref_id)) {
			throw new Exception('Trove.DataNotSet');
		}
		$isSiteAdmin = Session::getSession()->getUser()->isSiteAdmin();
		foreach ($_POST as $index => $value) {
			$root_trove_id = Validator::sanitizeInt($this->_getTroveIdFromSelectName($index));
			$troves = $value;
			if ($root_trove_id > 0) {
				TroveLinkPeer::removeProjectLinksToTroveSubtree($this->ref_id, $root_trove_id);
				if (!is_array($troves)) {
					$troves = array($troves);
				}
				$c = new Criteria();
				$c->add(TroveCategoryPeer::TROVE_CATEGORY_ID, $troves, Criteria::IN);
				$troveCategories = TroveCategoryPeer::doSelect($c);
				foreach ($troveCategories as $tc) {
					if (!$tc->getAccessLevel() || ($tc->getAccessLevel() == 99 && $isSiteAdmin)) {
						$troveLink = new TroveLink();
						$troveLink->setSection($this->section);
						$troveLink->setRefId($this->ref_id);
						$troveLink->setTroveCategoryId( Validator::sanitizeInt($tc->getTroveCategoryId()) );
						try {
							$troveLink->save();
						} catch (Exception $e) {
							error_log($e->getMessage());
						}
					}
				}
			}
		}
	}

	/**
	 * Traverses trove category tree and build a list of hierarchical options
	 *
	 * @param int $start
     * @param bool $add_root_cat_option wheter to add an option for Root Cat or not
	 */
	public static function buildTroveOptions($start=0, $add_root_cat_option=true) {

        if ($add_root_cat_option) {
            echo "<option value=\"0\" style=\"padding-left: 6px\">Root Cat</option>";
        }
		self::traverseTreeApplyCallback(TroveCategoryPeer::getTroveTree($start),'self::_renderSelectOption', array(
            'level_1_padding' => $add_root_cat_option ? 29 : 6
        ));

	}
	
	/**
	 * Traverses trove categories tree building a list of hierarchical options 
	 * with options for each project linked to each trove category
	 * 
	 * @param array $projects linking from trove_cat to project info
	 */
	public static function buildTroveWithProjectsOptions($projects, $default_user_template = null) {
        self::traverseTreeApplyCallback(TroveCategoryPeer::getTroveTree(GFConf::get('system.troveGfProjTemplates')),'self::_renderSelectOptionWithProjects', array(
            'level_1_padding' => 6,
			'projects' => $projects,
                        'default_user_template' => $default_user_template
        ));
	}
	

	/**
	 * @param int $start
	 * @param Language $lang
	 * @param SiteAdminTemplate $template
	 */
	public static function buildUlEntries($start=0, $template) {

		self::getFixedCategoriesWithEnableSettings();
		$lang = Session::getSession()->getLanguage();

		$image_url = $template->image('ic/cfolder15.png',15,13);

		//We set the language strings since the class methods are not accessible in the callback function
		$options = array(
			'edit'=>$lang->getText('TroveEdit.Edit'),
			'add'=>$lang->getText('TroveAdd.Add'),
			'delete'=>$lang->getText('TroveDelete.Delete'),
			'enabled'=>$lang->getText('TroveToggle.Enable'),
			'disabled'=>$lang->getText('TroveToggle.Disable'),
			'admin'=>$lang->getText('TroveEdit.AccessAdmin'),
            'show_in_search'=>$lang->getText('TroveEdit.ShowInSearch')
		);

		if ($start == 0) {
			echo '<li><span> Root ';
            echo '<a href="' . GFUrl::getUrl()->currentBaseURL('TroveAdd', array('parent_trove_cat_id' => 0)) . '">[' . $lang->getText('TroveAdd.Add') . ']</a> ';
            echo '</span>';
            echo '<ul>';
		}


		self::traverseTreeApplyCallback(TroveCategoryPeer::getTroveTree($start), 
            'self::_renderUlEntry', 
            $options, 
            function() {echo '</li>';}, 
            null,
            function() {echo '<ul>';}, 
            null,
            function() {echo '</ul>';}, 
            null
        );
        
        if ($start == 0) {
            echo '</ul>';
            echo '</li>';
        }
	}

	/**
	 * Get category settings and store it as property
	 */
	private static function getFixedCategoriesWithEnableSettings() {
		self::$_fixedCategoriesWithEnableSettings = TroveCategoryPeer::getFixedCategoriesIdsWithEnableSettings();
	}

	/**
	 * Traverse a tree and apply a callback
	 *
	 * @param array $categories
	 * @param callable $callback visit node callback
	 * @param array $options
     * @param callable $after_callback after traversing descendants nodes callback
     * @param array $after_options
     * @param callable $before_child_callback before visiting children callback
     * @param array $before_child_options
     * @param callable $after_child_callback after visiting children callback
     * @param array $after_child_options
	 */
	private static function traverseTreeApplyCallback(
        $categories, 
        $callback, 
        $options,  
        $after_callback = null,
        $after_options = null, 
        $before_child_callback = null, 
        $before_child_options = null, 
        $after_child_callback = null, 
        $after_child_options = null
    ) {

		foreach ($categories as $category) {
			call_user_func_array($callback, array($category, $options));
			if (array_key_exists('children', $category)) {
                if (!empty($before_child_callback)) {
                    call_user_func_array($before_child_callback, array($category, $before_child_options));
                }
				if(count($category['children'])) {
					self::traverseTreeApplyCallback(
                        $category['children'], 
                        $callback, $options, 
                        $after_callback, $after_options, 
                        $before_child_callback, $before_child_options, 
                        $after_child_callback, $after_child_options
                    );
				}
                if (!empty($after_child_callback)) {
                    call_user_func_array($after_child_callback, array($category, $after_child_options));
                }
			}
            if (!empty($after_callback)) {
                call_user_func_array($after_callback, array($category, $after_options));
            }
		}
	}

	/**
	 * Renders an HTML option element for $trove
	 *
	 * @param array $category
	 * @internal int $category['trove_category_id']
	 * @internal string $category['level'] Level of depth in the tree
	 * @internal string $category['category_name']
	 */
	private static function _renderSelectOption($category, $options) {
        echo "<option value=\"".$category['trove_category_id']."\" style=\"padding-left: ".($options['level_1_padding'] + 23 * (int)$category['level'])."px\">".$category['category_name']."</option>";
	}
	
	/**
	 * Renders an HTML option element for $category (disabled, used as optgroup)
	 * and one HTML option element for each project in $options['project']
	 * 
	 * @param array $category trove
	 * @param array $options
	 */
	private static function _renderSelectOptionWithProjects($category, $options) {
		echo "<option disabled=\"disabled\" style=\"padding-left: ".($options['level_1_padding'] + 23 * (int)$category['level'])."px\">{$category['category_name']}</option>";
		$category_id = $category['trove_category_id'];
		if (isset($options['projects'][$category_id])) {
			foreach ($options['projects'][$category_id] as $project_id => $project_name) {
                            if ($options['default_user_template'] && $project_id == $options['default_user_template']) {
                                echo "<option value=\"".$project_id."\" style=\"padding-left: ".($options['level_1_padding'] + 23 * (int)$category['level'])."px\" selected>".$project_name."</option>";
                            } else {
                                echo "<option value=\"".$project_id."\" style=\"padding-left: ".($options['level_1_padding'] + 23 * (int)$category['level'])."px\">".$project_name."</option>";
                            }
			}
		}
	}

	/**
	 * Callback function to render an ul list (deprecated)
	 *
	 * @param array $category
	 * @param array $options
	 * @internal int $category['trove_category_id']
	 * @internal string $category['category_name']
	 * @internal int $category['access_level']
	 * @inernall int $category['parent_trove_category_id']
	 * @internal Language $options['language']
	 */
	private static function _renderUlEntry($category,$options) {
        $trove_id = $category['trove_category_id'];
		$trove_name = $category['category_name'];
		$access_level = $category['access_level'];
        $show_in_search = $category['show_in_search'];
		$parent_trove_category_id = $category['parent_trove_category_id'];		

		$editStr = $options['edit'];
		$deleteStr = $options['delete'];
		$addStr = $options['add'];
		$enabledStr = $options['enabled'];
		$disabledStr = $options['disabled'];
		$adminStr = $options['admin'];
        $showInSearchStr = $options['show_in_search'];


		$is_fixed_category = ($parent_trove_category_id == 0) && array_key_exists($trove_id, self::$_fixedCategoriesWithEnableSettings);
		$is_enabled = $is_fixed_category && self::$_fixedCategoriesWithEnableSettings[$trove_id];
		echo '<li>';
        echo '<div>';
        if ($trove_id != 0) {
			echo "<input type=\"checkbox\" id=\"trove_$trove_id\" name=\"trove_ids[]\" value=\"$trove_id\" style=\"margin-left: 10px;\" data-category-name=\"$trove_name\"></input>";
		}
        echo '<span> '.$trove_name.' ';
 		if ($trove_id != 0) {
 			echo '<a href="'.GFUrl::getUrl()->currentBaseURL('TroveEdit', array('trove_cat_id' => $trove_id)).'">['.$editStr.']</a> ';
 		}
		echo '<a href="'.GFUrl::getUrl()->currentBaseURL('TroveAdd', array('parent_trove_cat_id' => $trove_id)).'">['.$addStr.']</a> ';
		if ($trove_id != 0) {
			if (!$is_fixed_category) {
				echo '<a href="'.GFUrl::getUrl()->currentBaseURL('TroveDelete', array('trove_cat_id' => $trove_id)).'">['.$deleteStr.']</a> ';
			} else {
				echo '<a href="#" class="trove-enabled-toggle" id="toggle-trove-'.$trove_id.'" data-category-name="'.$trove_name.'"'
						. (($trove_id == GFConf::get('system.troveGfProjTemplates')) ? ' style="display: none" ' : '')
						.'>['
						.(!$is_enabled ? $enabledStr : $disabledStr)
						.']</a> ';
			}
		}
		if($trove_id != 0 && $access_level == 99) {
			echo '<strong>['.$adminStr.']</strong>';
		}
        if($show_in_search) {
			echo ' <strong>['.$showInSearchStr.']</strong>';
		}
		echo '</span>';
        echo '</div>';        
    }

}
