<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version $Id$
 * @author Marcelo Mottalli <marcelo@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

class DAVConfigManager
{
	private $project;
	private $accessFile, $authFile;
	
	private $paths = array();
	private $userGroups = array();
	private $users = array();
	
	public function __construct($project, $accessFile, $authFile) {
		$this->project = $project;
		$this->accessFile = $accessFile;
		$this->authFile = $authFile;
	}
	
	public function addUser($user, $passwordHash = '') {
		if ($user instanceof User) {
			if (!$passwordHash) {
				$passwordHash = $user->getPasswordCrypt();
			}

			$user = $user->getUnixName();
		}
		$this->users[$user] = $passwordHash;
	}
	
	public function addUserGroup($groupName) {
		$this->userGroups[$groupName] = array();
	}
	
	public function addUserToGroup($user, $groupName) {
		if ($user instanceof User) {
			$user = $user->getUnixName();
		}

		if (!array_key_exists($groupName, $this->userGroups)) {
			throw new Exception("Tried to add an user to non-existing group");
		} elseif (!array_key_exists($user, $this->users)) {
			throw new Exception("Tried to add non-existent user to group");
		}

		$this->userGroups[$groupName][] = $user;
	}
	
	/**
	 * Sets the *minimum* access for the user. Successive calls to this function will
	 * give the user the maximum access specified. E.g.:
	 * 
	 * <code>
	 * $this->setMinUserAcess('user', '/path', ACCESS_READ);
	 *  ...
	 * $this->setMinUserAcess('user', '/path', ACCESS_READ_WRITE);
	 * </code>
	 * 
	 * Will give the user read/write access to /path.
	 * The idea behind this function is that some users also belong to user groups, and you can
	 * give write access to the user and read-only access to the group (and viceversa)
	 */
	public function setMinUserAccess($user, $path, $access) {
		if (!array_key_exists($path, $this->paths)) {
			$this->paths[$path] = array();
		}
		
		if ($user instanceof User) {
			$user = $user->getUnixName();
		}

		if ($user != '*' && !array_key_exists($user, $this->users)) {
			throw new Exception("Tried to set access for non-existent user");
		}
		
		if (array_key_exists($user, $this->paths[$path])) {
			// The permission for the user was previously set. Keep the maximum of both.
			$access = max($access, $this->paths[$path][$user]);
		}
		
		$this->paths[$path][$user] = $access;
	}
	
	public function setGroupAccess($groupName, $path, $access) {
		if (!array_key_exists($path, $this->paths)) {
			$this->paths[$path] = array();
		}
		
		$this->paths[$path]['@'.$this->getFullGroupName($groupName)] = $access;
	}
	
	protected function getFullGroupName($groupName) {
		return $this->project->getUnixName().'-'.$groupName;
	}
	
	public function writeChanges() {
		$this->writeAuthFile();
		$this->writeAccessFile();
	}
	
	private function writeAuthFile() {
		// New data to be written
		$contents = array();
		$locked = false;
		
		// Parse the old auth file, if it exists
		$oldContents = array();
		if (file_exists($this->authFile)) {
			//$oldContents = file($this->authFile, FILE_IGNORE_NEW_LINES);
			$fp = fopen($this->authFile, 'r');
			if ($fp) {
				$locked = flock($fp, LOCK_EX);
				while ($str = fgets($fp)) {
					$st = rtrim($str, "\r\n") ;
					if ( $st === '' ) {
						continue;
					} else {
						$oldContents[] = $st;
					}
				}
				fclose($fp);
			} else {
				syslog(LOG_ERR, "Couldn't open file {$this->authFile} \n");
			}
		}
		
		// Keeps the old users
		foreach ($oldContents as $line) {
			if (preg_match('/^(.*):/', $line, $matches)) {
				$user = $matches[1];
				if (array_key_exists($user, $this->users)) {
					// Duplicate user, do nothing
				} else {
					// Keep the line as is
					$contents[] = $line;
				}
			}
		}
		
		// Now, write the new users
		foreach ($this->users as $user => $password) {
			$contents[] = "$user:$password";
		}
		
		SystemUtils::writeFile($this->authFile, $contents, $locked);
	}
	
	private function writeAccessFile() {
		// New data to be written
		$contents = array('[groups]');			// First line must be "[groups]", where we define the user groups
		$locked = false;
		$oldContents = array();
		
		// Parse the old access file, if it exists
		if (file_exists($this->accessFile)) {
			//$oldContents = file($this->accessFile, FILE_IGNORE_NEW_LINES);
			//File() doesn't support locks. We need to lock while reading and until the file
			// is written to avoid concurrent access problems
			$fp = fopen($this->accessFile, 'r');

			if ($fp) {
				$locked = flock($fp, LOCK_EX);
				while ($str = fgets($fp)) {
					$st = rtrim($str, "\r\n") ;
					if ( $st === '' ) {
						continue;
					} else {
						$oldContents[] = $st;
					}
				}
				fclose($fp);
				assert($oldContents[0] == "[groups]");
			} else {
				syslog(LOG_ERR, "Couldn't open file {$this->accessFile} \n");
			}
		}
		
		// Generate the [groups] section keeping the old groups and writing the new ones
		$prefix = $this->project->getUnixName().'-';
		for ($i = 1; $i < count($oldContents); $i++) {
			$line = $oldContents[$i];
			
			// End of the [groups] section
			if (strpos($line, '[') === 0) {
				break;
			}
			
			// If the group corresponds to this project, delete it (they will be written later)
			if (empty($line) || strpos($line, $prefix) === 0) {
				continue;
			}
			
			$contents[] = $line;
		}
		
		// Now append this project's groups
		foreach ($this->userGroups as $groupName => $users) {
			$users = implode(',', $users);
			$contents[] = $this->getFullGroupName($groupName).' = '.$users;
		}

		// Now, skip all the old sections that corresponded to this project
		$skip = false;
		for (; $i < count($oldContents); $i++) {
			$line = $oldContents[$i];
			if (empty($line)) {
				continue;
			}
			
			if (preg_match('/^\\[([^:]*):/', $line, $matches)) {		// Begin of section
				$sectionProject = $matches[1];
				if ($sectionProject == $this->project->getUnixName()) {
					// This section is related to this project -- skip it (they will be written later)
					$skip = true;
				} else {
					$contents[] = '';			// Insert a blank line for beautification purposes
					$skip = false;
				}
			}
			
			if (!$skip) {
				$contents[] = $line;
			}
		}
		
		// Now, write this project's sections
		foreach ($this->paths as $path => $userPermissions) {
			$contents[] = '';
			$contents[] = '['.$this->project->getUnixName().':'.$path.']';
			foreach ($userPermissions as $user => $permission) {
				if ($permission == ScmPlugin::NOACCESS) $permission = '';
				elseif ($permission == ScmPlugin::READ) $permission = 'r';
				elseif ($permission == ScmPlugin::WRITE) $permission = 'rw';
				$contents[] = "$user = $permission";
			}
		}

		//By setting $locked=true we instruct writeFile to not try to lock again, only release the lock once
		//the write is done
		SystemUtils::writeFile($this->accessFile, $contents, $locked);
	}

}
?>