<?php

/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Ruben Gutierrez	<ruben@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
require_once('plugin/GFPlugin.php');
require_once('plugin/ActionManager.php');
require_once('plugin/PluginResult.php');
require_once('core/Language.php');
require_once('lib/GFSearch.php');
require_once('lib/GFUrl.php');
require_once('lib/core/ProjectPluginPeer.php');
require_once('lib/core/PluginPeer.php');

class BuildPlugin implements GFPlugin {

	const NOACCESS = -1;
	const ADMIN = 3;
	const SECTION = 'build';
	const PERMLEVEL_NONE = -1;
	const PERMLEVEL_READ = 0;
	const PERMLEVEL_ADMIN = 3; //can read/edit all
	const PLUGIN_NAME = 'build';

	function __construct() {

	}

	/**
	 * 	Different types of plugins are possible -
	 * 	GFPlugin::SITE_PLUGIN = 'site'
	 * 	GFPlugin::PROJECT_PLUGIN = 'project'
	 * 	GFPlugin::USER_PLUGIN = 'user'
	 */
	public function getType() {
		return GFPlugin::PROJECT_PLUGIN;
	}

	/**
	 * 	get the name of this plugin. example: 'cruisecontrol'
	 * 	@return	string
	 */
	public function getName() {
		return self::PLUGIN_NAME;
	}

	/**
	 * 	this is called when the first Action is called. Action manager
	 * 	gets the listeners for each plugin.
	 */
	public function initialize() {
		ActionManager::registerListeners($this, $this->getListeners());
	}

	/**
	 * 	get the array of action names that this plugin answers to.
	 * 	@return	array
	 */
	public function getListeners() {
		return array(
			'BUILD_PROJECT_TABS',
			'GET_PROJECT_ADMIN_NAV_ELEMENTS',
			'GET_ROLE_CHOICES',
			'UPDATE_ROLE_CHOICES',
			'GET_PROJECT_REPORTING_NAV_ELEMENTS',
			'CRUISECONTROL_SAVE',
			'CRUISECONTROL_BUILD_SAVE'
		);
	}

	/**
	 * 	take a search object and handle the search internally.
	 * 	@param	GFSearch
	 * 	@return PluginResult
	 */
	function search(GFSearch $search) {

	}

	/**
	 * 	handle action has a case statement to redirect the action to the appropriate function.
	 * 	@param string
	 * 	@param mixed
	 * 	@return PluginResult
	 */
	public function handleAction($listenerName, $params) {
		switch ($listenerName) {
			case 'BUILD_PROJECT_TABS': {
					return $this->_buildMainTabsListener($params);
					break;
				}
			case 'GET_PROJECT_ADMIN_NAV_ELEMENTS': {
					return $this->getProjectAdminNavElements($params);
					break;
				}
			case 'GET_ROLE_CHOICES': {
					return $this->getRoleChoices($params);
					break;
				}
			case 'UPDATE_ROLE_CHOICES': {
					return $this->updateRoleChoices($params);
					break;
				}
			case 'GET_PROJECT_REPORTING_NAV_ELEMENTS': {
					return $this->getProjectReportingNavElements($params);
					break;
				}
			case 'CRUISECONTROL_SAVE': {
					return $this->doBuildSave($params['is_new'], $params['item']);
					break;
				}
			case 'CRUISECONTROL_BUILD_SAVE': {
					return $this->doBuildResultSave($params['item']);
					break;
				}
		}
	}

	private function _buildMainTabsListener($params) {
		if (!isset($params['project']) || !is_object($params['project']) || !$params['project']->usesPlugin($this->getName())) {
			return;
		}
		$project = $params['project'];

		$result = new PluginResult($this);
		$value = array(
			'text' => Language::getSessText('Build.TabString'),
			'action' => GFUrl::getUrl()->projectURL($project, '', 'build'),
			'id' => 'cruisecontrol'
		);
		$result->setResult($value);
		return $result;
	}

	/**
	 *      getProjectAdminNavElements - get the local nav elements for cruisecontrol
	 *      @return PluginResult
	 */
	private function getProjectAdminNavElements($params) {
		if (!isset($params['project']) || !is_object($params['project']) || !$params['project']->usesPlugin($this->getName())) {
			return;
		}
		$result = new PluginResult($this);
		$arr = array(
			'title' => 'CruiseControl',
			'elements' => array(
				array('link' => '/gf/project/' . $params['project']->getUnixName() . '/build/admin/',
					'description' => Language::getSessText('Build.Admin'))
			)
		);
		$result->setResult($arr);
		return $result;
	}

	/**
	 * 	get project uri to be used as root for constructing uris
	 * 	@return string
	 */
	function getProjectPluginUri() {
		return GFUrl::getUrl()->projectURL($this->project, '', 'build');
	}

	function enableInProject(Project $project) {
		return;
	}

	function disableFromProject() {
		// unregister the actions
		ActionManager::unregisterListeners($this);
	}

	/**
	 * returns the possible role settings available for this plugin
	 *
	 * @return array	the roles
	 */
	function getAvailableRoleSettings() {
		return parent::getAvailableRoleSettings();
	}

	/**
	 * returns the possible site-wide admin role settings for this plugin
	 *
	 * @return array	the roles
	 */
	function getAdminRoleSettings() {
		$return = array();
		$return[0][] = Language::getSessText('Build.None'); //descriptions
		$return[1][] = self::NOACCESS;  // values
		$return[0][] = Language::getSessText('Build.Read');
		$return[1][] = self::PERMLEVEL_READ;
		$return[0][] = Language::getSessText('Build.Admin');
		$return[1][] = self::ADMIN;
		return $return;
	}

	/**
	 * 	The getRoleChoices listener - this hook will fire
	 * 	when a new role is added or one is going to be updated.
	 * 	It returns an array of role settings
	 * 	@return	array
	 */
	function getRoleChoices($params) {
		if (!is_object($params['project']) || !$params['project']->usesPlugin($this->getName())) {
			return;
		}

		$roles = array();
		$roles[0]['ref_id'] = 0;
		$roles[0]['section'] = Language::getSessText('Build.SiteWideAdmin');
		$roles[0]['description'] = Language::getSessText('Build.SiteWideAdmin.Description');
		$roles[0]['subsection'] = '-';
		$roles[0]['settings'][0] = self::SECTION;
		$roles[0]['settings'][] = $this->getAdminRoleSettings();

		$result = new PluginResult($this);
		$result->setResult($roles);
		return $result;
	}

	/**
	 * returns the settings for the cruisecontrol status
	 *
	 * @return array
	 */
	function getStatusObserverSettings() {
		return parent::getStatusObserverSettings();
	}

	/**
	 * 	The getObserverRoleChoices listener - this hook will fire
	 * 	when the observer role is modified.
	 * 	It returns an array of role settings or null
	 * 	@return	array or null
	 */
	function getObserverRoleChoices($params) {
		//return parent::getObserverRoleChoices($params);
		return null;
	}

	/**
	 * Retrieves the concerning data and updates its observer role settings
	 *
	 * @param array	$data
	 */
	function updateObserverSettings($data) {
		return parent::updateObserverSettings($data);
	}

	private function updateRoleChoices($params) {
		if (!is_object($params['project']) || !$params['project']->usesPlugin($this->getName())) {
			return;
		}

		$section = self::SECTION;

		$project = $params['project'];
		$data = $params['data'];
		$role = $params['role'];

		if (!isset($data[$section])) {
			return;
		}

		foreach ($data[$section] as $ref_id => $value) {
			$c = new Criteria();
			$c->add(RoleSettingPeer::ROLE_ID, $role->getPrimaryKey());
			$c->add(RoleSettingPeer::SECTION, $section);
			$c->add(RoleSettingPeer::REF_ID, $ref_id);
			$rolesetting = RoleSettingPeer::doSelectOne($c);
			if (!is_object($rolesetting)) {
				// Enter here if we have a role setting which we didn't have previously (e.g.: activate an inactive plugin)
				$rolesetting = new RoleSetting();
				$rolesetting->setSection($section);
				$rolesetting->setRefId($ref_id);
				$rolesetting->setRoleId($role->getPrimaryKey());
			}

			// Being a project admin overrides any other value
			if (isset($data['projectadmin']) && $data['projectadmin'][0] >= Project::PROJECT_ROLE_ADMIN) {
				$value = self::ADMIN;
			}

			$rolesetting->setValue($value);
			$rolesetting->save();
		}

		return;
	}

	private function getProjectReportingNavElements($params) {
		if (!isset($params['project']) || !is_object($params['project']) || !$params['project']->usesPlugin($this->getName())) {
			return;
		}
		$result = new PluginResult($this);
		$arr = array(
			'title' => 'Cruise Control',
			'elements' => array(
				array('link' => $this->getBuildReportUrl($params['project']),
					'description' => Language::getSessText('CruiseControlBuildReport.BuildsReport'))
			)
		);
		$result->setResult($arr);
		return $result;
	}

	private function getBuildReportUrl($project) {
		if (!is_object($project) || !$project->usesPlugin($this->getName())) {
			return '';
		}
		return GFUrl::getUrl()->projectURL($project, 'CruisecontrolBuildReport', $this->getName());
	}

	private function doBuildSave($new, $build) {
		$source = "Build {$build->getBuildName()}";
		$user_id = 100;
		$user_name = 'Anonymous User';
		if (Session::getSession()) {
			$user = Session::getSession()->getUser();
			if ($user != null) {
				$user_id = $user->getUserId();
				$user_name = $user->getUnixName();
			}
		}

		$verb = $new ? 'create' : 'update';

		$url = $this->getBuildReportUrl($build->getProject());

		$description = "{$user_name} {$verb}d {$source}";

		if ($item != null) {
			$extended_data = PropelJson::ShallowJson($item);
		} else {
			$extended_data = PropelJson::ShallowJson($build);
		}

		$event = array(
			'project_id' => $build->getProject()->getProjectId(),
			'section' => BuildPlugin::SECTION,
			'ref_id' => $build->getPrimaryKey(),
			'user_id' => $user_id,
			'activity_date' => time(),
			'description' => $description,
			'url' => $url,
			'activity_type' => $verb,
			'extended_data' => $extended_data
		);

		$result = ActionManager::notify('ACTIVITY_LOG_EVENT', $event);

		$ret = new PluginResult($this);
		$ret->setResult(true);

		return $ret;
	}

	private function doBuildResultSave($event) {
		$build = $event->getCruisecontrol();
		$source = "Build {$build->getBuildName()}";

		$user_id = 100;
		$user_name = 'Anonymous User';
		if (Session::getSession()) {
			$user = Session::getSession()->getUser();
			if ($user != null) {
				$user_id = $user->getUserId();
				$user_name = $user->getUnixName();
			}
		}

		if ($event->getWasFix()) {
			$verb = 'FIXED';
		} elseif ($event->getWasSuccessful()) {
			$verb = 'PASSED';
		} else {
			$verb = 'FAILED';
		}

		$event_type = CruiseControlBuildPeer::getEventTypeDescription($event->getEventType());

		$dt = $event->getBuildDate();

		$url = $this->getBuildReportUrl($build->getProject());

		if ($event->getWasSuccessful()) {
			$description = "Automated build {$event_type} for {$source} at {$dt}";
		} else {
			$description = "Automated build failed for {$source} at {$dt}";
		}
		$extended_data = PropelJson::ShallowJson($event);

		$event = array(
			'project_id' => $build->getProject()->getProjectId(),
			'section' => BuildPlugin::SECTION,
			'ref_id' => $build->getPrimaryKey(),
			'user_id' => $user_id,
			'activity_date' => time(),
			'description' => $description,
			'url' => $url,
			'activity_type' => $verb,
			'extended_data' => $extended_data
		);

		$result = ActionManager::notify('ACTIVITY_LOG_EVENT', $event);

		$ret = new PluginResult($this);
		$ret->setResult(true);

		return $ret;
	}

}

