<?php

require_once('plugin/GFPlugin.php');
require_once('plugin/ActionManager.php');
require_once('plugin/PluginResult.php');
require_once('core/Language.php');
require_once('GFUrl.php');
require_once('lib/core/SettingPeer.php');

class DiskquotaPlugin implements GFPlugin {

    CONST ALLOW = 1;
    CONST DISABLE = 0;
    CONST SECTION = 'Diskquota';

    function __construct() {
    }   
    
    /**
     *  this is called when the first Action is called. Action manager 
     *  gets the listeners for each plugin.
     */
    public function initialize() {
        ActionManager::registerListeners($this, $this->getListeners());
    }

    /**
     *  Different types of plugins are possible -
     *  GFPlugin::SITE_PLUGIN = 'site'
     *  GFPlugin::PROJECT_PLUGIN = 'project'
     *  GFPlugin::USER_PLUGIN = 'user'
     */
    public function getType() {
        return GFPlugin::SITE_PLUGIN;
    }   
    
    /**
     *  get the name of this plugin.
     *  @return string
     */
    public function getName() {
        return 'diskquota';
    }

    /**
     *  get the array of action names that this plugin answers to.
     *  @return array
     */
    public function getListeners() {
        return array(
            'PROJECT_INFO_PANELS',
            'SITE_ADMIN_INDEX'
        );
    }
    
    public function search(GFSearch $search) {
        
    }

    /**
     *  handle action has a case statement to redirect the action to the appropriate function.
     *  @param string
     *  @param mixed
     *  @return PluginResult
     */
    public function handleAction($listenerName, $params) {
        switch ($listenerName) {
            case 'PROJECT_INFO_PANELS' :
                return $this->projectInfoPanels($params);
            case 'SITE_ADMIN_INDEX':
                return $this->getPluginSiteAdminElements($params);
        }
    }

    private function projectInfoPanels($params) {
        $project = $params['project'];
        $lang = Language::getSessLang();
        
        $panel = array();
        $panel['title'] = $lang->getText('Diskquota.Title');

        $contents = '<div align="center"><table border="0" cellspacing="0" cellpadding="0">';
        $diskUsage = SettingPeer::FindSetting('project', $project->getId(), 'disk-used');
        $diskQuota = SettingPeer::FindSetting('project', $project->getId(), 'project-quota');

        $quotaAvailable = is_object($diskQuota) ? (float)$diskQuota->getSettingValue() : GFConf::get('diskquota.projectMaxQuota'); 
        $quotaUsed = is_object($diskUsage) ? (float)$diskUsage->getSettingValue() / (1024*1024) : '0';

        $contents .= '<tr><td colspan="3" align="center">';
        if ($quotaAvailable != 0) {
          $contents .= $lang->getText('Diskquota.Used', array(round($quotaUsed, 2), round($quotaAvailable / (1024*1024), 2)));
          $quotaConsumed = is_object($diskUsage) ? (float)$diskUsage->getSettingValue()  : '0';
          $percent_used = number_format(($quotaConsumed / $quotaAvailable) * 100, 2);
          $contents .= $lang->getText('Diskquota.ProjectSummaryPercent', array($percent_used)) . '</td></tr>';

        } else {
          $contents .= $lang->getText('Diskquota.UsedUnlimited', array(round($quotaUsed, 2)));
        }

        $contents .= '</table></div>';
        $panel['contents'] = $contents;

        $pr = new PluginResult($this);
        $pr->setResult(array($panel)); // We only show one panel but we could show more
        return $pr;
    }

    private function getPluginSiteAdminElements($params) {
        $arr = array(
            'title' => Language::getSessText('Diskquota.Title'),
            'elements' => array('line' => '<a href="'.GFUrl::getUrl()->baseURL('ManageDiskQuota','diskquota','admin').'">'.Language::getSessText('Diskquota.ManageQuota').'</a>')
        );
        $pr = new PluginResult($this);
        $pr->setResult($arr);
        return $pr;
    }

    private function buildMainTabsListener($params) {
        $project = $params['project'];
        $result = new PluginResult($this);
        $value = array(
            'text'  => Language::getSessText('Diskquota.Title'),
            'action'=> GFUrl::getUrl()->projectURL($project,'','diskquota'),
            'id'    => 'diskquota'
        );
        $result->setResult($value);
        return $result;
    }

    /* 
    Check if a project is over its allowed disk quota
    The quota is stated in project_text6, in bytes, or if it is not set there, in the config file for this plugin 
    Returns true if the project is within its allowed quota
    */
    
    static public function checkDiskQuota($params,$listenerName) {
        $project=null;

        switch ($listenerName) {
            case 'TRACKER_ITEM_SAVE_BEFORE': {
                $project=$params->getTracker()->getProject();
                break;
            }
            case 'FRS_RELEASE_SAVE_BEFORE': {
                $project=$params->getFrsPackage()->getProject();
                break;
            }

            case 'FILESYSTEM_SAVE_BEFORE': {

                switch ($params->getSection()) {
                    case 'htmleditor': {
                        return true;
                        break;
                    }
                    case ForumMessage::SECTION: {
                        $forum = ForumMessagePeer::retrieveByPk($params->getRefId())->getForumThread()->getForum();
                        switch($forum->getSection()) {
                            case 'project':
                                $project = ProjectPeer::retrieveByPk($forum->getRefId());
                                break;
                            case 'mailman':
                                $project = MailmanPeer::retrieveByPk($forum->getRefId())->getProject();
                                break;
                            case 'news':
                                $project = NewsPeer::retrieveByPk($forum->getRefId())->getProject();
                                break;
                            default:
                                return true;
                                break;
                        }
                        break;
                    }
                    case TrackerItem::SECTION: {
                        $project = TrackerItemPeer::retrieveByPk($params->getRefId())->getTracker()->getProject();
                        break;
                    }
                    case UserwikiPlugin::SECTION:
                        $wikiversion = UserWikiVersionPeer::retrieveByPk($params->getRefId());
                        $wikipage = $wikiversion->getWikiPage();

                    case WikiPlugin::SECTION:
                        $wikiversion = WikiVersionPeer::retrieveByPk($params->getRefId());
                        $wikipage = $wikiversion->getWikiPage();
                    case Wiki::WIKI_FILES_SECTION:
                        if($params->getSection()==Wiki::WIKI_FILES_SECTION) {
                            $wikipage = WikiPagePeer::retrieveByPk($params->getRefId());
                        }
                    case WikiPage::SECTION: {
                        if($params->getSection()==WikiPage::SECTION) {
                            $wikipage = WikiPagePeer::retrieveByPk($params->getRefId());
                        }
                        switch($wikipage->getSection()) {
                            case 'project':
                                $project = ProjectPeer::retrieveByPk($wikipage->getRefId());
                                break;

                            case 'user':
                                $projects = UserPeer::retrieveByPk($wikipage->getRefId())->getAllProjects();
                                if(count($projects)==1) {
                                    $project = $project[0];
                                }
                                else {
                                    // Don't know really what to do, the user either does not belong to a project, or belongs to more than one.
                                    // Allow it to continue.
                                    return true;
                                }
                                break;
                            
                            default:
                                return true;
                        }
                        break;
                    }
                    case FrsRelease::SECTION: {
                        $project = FrsReleasePeer::retrieveByPK($params->getRefId())->getFrsPackage()->getProject();
                        break;
                    }
                    case DocmanFileVersion::SECTION: {
                        $project = DocmanFileVersionPeer::retrieveByPk($params->getRefId())->getDocmanFile()->getDocmanFolder()->getProject();
                        break;
                    }
                    case Project::SECTION: {
                        $project = ProjectPeer::retrieveByPK($params->getRefId());
                        break;
                    }
                }
                break;
            }
            
            case 'DOCMAN_FILE_SAVE_BEFORE': {
                $project = $params['item']->getDocmanFolder()->getProject();
                break;
            }
            case 'SCMCVS_PRECOMMIT': {
                $project=$params;
                break;
            }
            case 'SCMSVN_PRECOMMIT': {
                $project=$params;
                break;
            }
        }
        // Get the latest disk usage figure
        if(is_object($project)) {
            if ($project->getProjectText5()==self::TYPE_GFORGEGROUP) {
                return true;
            }
            $c = new Criteria();
            $c->add(DiskUsagePeer::PROJECT_ID, $project->getProjectId());
            $c->addDescendingOrderByColumn(DiskUsagePeer::CHECK_DATE);
            $du = DiskUsagePeer::doSelectOne($c);

            if (is_object($du)) {
                if ($project->getProjectText5()==self::TYPE_FREE || $project->getProjectText5()=='') {
                    if ($du->getBytesUsed() > GFConf::get('hosting.projectMaxQuota')) {
                        return false;
                    } else {
                        return true;
                    }
                } else {
                    $quota = intval($project->getProjectText6());
                    if ($quota == 0) {
                        $quota = GFConf::get('hosting.projectMaxQuota');
                    }

                    if ($quota > $du->getBytesUsed()) {
                        return true;
                    } else {
                        return false;
                    }
                }
            } else {
                return true;
            }
        } else {
            throw new Exception('Invalid Project Object.');
        }
    }
}
