<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version $Id$
 * @author Daniel A. P�rez	daniel@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once("exceptions/PermissionDeniedException.php");
require_once("GFPermission.interface.php");
require_once("plugin/ActionManager.php");
require_once('DocmanFileVersionPeer.php');
require_once('lib/core/FilesystemPeer.php');
require_once('lib/core/UserPeer.php');
require_once('plugins/docman/lib/om/BaseDocmanFile.php');
require_once('plugins/docman/lib/DocmanFileNotification.php');

class DocmanFile extends BaseDocmanFile implements GFPermission {
	private $file_name;
	const SECTION='docmanfile';
	const EDITED_INLINE = 'edited_inline';
	const EDITED_UPLOAD = 'edited_upload';
	
	private $monitoring_users;

	/**
	 * setFilename - sets the filename to be used by the notification
	 */
	function setFilename($file_name) {
		$this->file_name = $file_name;
	}

	/**
	 * isPublic - true/false
	 * @return boolean
	 */
	function isPublic() {
		return $this->getDocmanFolder()->isPublic();
	}

	/**
	 * Specifies if the current user can create an entry in the database for this object
	 *
	 * @return bool
	 * @todo Program the logic
	 */
	public function userCanCreate() {
		if (!Session::getSession()->isLoggedIn()) {
			return false;
		}
		if ($this->isPublic()) {
			return true;
		}

		if ($this->getDocmanFolderId() > 0) {
			$docmanFolder = DocmanFolderPeer::retrieveByPk($this->getDocmanFolderId());
				
			if (is_object($docmanFolder)) {
				if ($docmanFolder->userCanCreate()) {
					return true;
				}
			}
		}
		
		return false;
	}

	/**
	 * If this person is an admin for this object
	 * @return boolean
	 */
	public function userIsAdmin() {
		if (!Session::getSession()->isLoggedIn()) {
			return false;
		}

		$rootfolderid=$this->getDocmanFolder()->getTopParentFolderId();
		$perm = $this->getDocmanFolder()->getProject()->getPermission();
		if ($perm->check(DocmanPlugin::SECTION, 0, DocmanPlugin::ADMIN)) {
			return true;
		}

		$folder = $this->getDocmanFolder();
		if ($folder->userCanCreate()) {
			return true;
		}	

		//Get the folders that matters		
		$folders = $this->getTopParentFolders();
		if(!count($folders)) {
			return false;
		}

		//Take the first folder so we query for permissions only once
		$perm = $folders[0]->getProject()->getPermission();
		if ($perm->check(DocmanPlugin::SECTION, 0, DocmanPlugin::ADMIN)) {
			return true;
		}

		foreach($folders as $folder){
			if($perm->check(DocmanFolder::SECTION, $folder->getDocmanFolderId(), DocmanFolder::PERMLEVEL_WRITE)) {
				return true;
			}
		}
	
		return false;		
	}

	/**
	 * Specifies if the current user can update this object's entry in the database
	 *
	 * @return bool
	 * @todo Program the logic
	 */
	public function userCanUpdate() {
		return $this->userIsAdmin();
	}
	
	/**
	 * Specifies if the current user can delete this object's entry from the database
	 *
	 * @return bool
	 * @todo Program the logic
	 */
	public function userCanDelete() {
		return $this->userIsAdmin();
	}

	/**
	 * Specifies if the current user has permissions to access this object
	 *
	 * @return bool
	 * @todo Program the logic
	 */
	public function userCanRead($folder_id=null) {
	
		if ($this->userIsAdmin()) {
			return true;
		}
		if ($this->isPublic()) {
			return true;
		}
		
		/**
		if($folder_id && $perm->check(DocmanFolder::SECTION, $folder_id, DocmanFolder::PERMLEVEL_READ)){
			return true;
		}*/

		return $this->checkMaximumAccessType(DocmanFolder::PERMLEVEL_READ);
	}
	

	/**
	 * Removes this object from datastore and sets delete attribute. Informs the plugin manager.
	 * This method was added by propel_customizer.php.
	 * 
	 * @param Connection $con
	 * @return void
	 * @throws PropelException
	 * @see DocmanFile::setDeleted()
	 * @see DocmanFile::isDeleted()
	 */
	function delete($con = null) {
		if (!$this->userCanDelete()) {
			throw new PermissionDeniedException("Could not delete DocmanFile object: permission denied");
		}
		$this->getDocmanFolder()->setDocCount($this->getDocmanFolder()->getDocCount()-1);
		$this->getDocmanFolder()->save();

		// Audit trail
		$audit = new AuditTrail();
		$audit->setUser(Session::getSession()->getUser());
		$audit->setSection(Project::SECTION);
		$audit->setRefId($this->getDocmanFolder()->getProjectId());
		$audit->setFieldName(self::SECTION);
		$audit->setNewValue('_deleted_');
		$audit->setOldValue($this->getLatestVersion()->getFile()->getFileName());
		$audit->save();

		$params = array(
			'is_new' => $this->isNew(),
			'item' => $this
		);

		ActionManager::notify("DOCMAN_FILE_DELETE", $params);
		parent::delete($con);		
		
		$c = new Criteria();
		$c->add(MonitorPeer::SECTION, self::SECTION);
		$c->add(MonitorPeer::REF_ID, $this->getDocmanFileId());
		MonitorPeer::doDelete($c);
		
	}

	/**
	 * Stores the object in the database using the Propel methods. Informs the plugin manager.
	 * This method was added by propel_customizer.php.
	 *
	 * @param Connection $con
	 * @return void
	 * @throws PropelException
	 */
	function save($con = null, $do_notify=true) {
		$errors=array();
		if (strlen($this->description) < 1) {
			$errors['description'] = Language::getSessText('Docman.EmptyDescription');
		}

		if (count($errors) > 0) {
			throw new InvalidDataException($errors);
		}

		//Check if unprivileged upload is available
		$setting = DocmanProjectSettingPeer::retrieveByPk($this->getDocmanFolder()->getProjectId());
		if (is_object($setting)) {
			$allow_unprivileged_upload = $setting->getAllowUnprivilegedUpload();
		} else {
			$allow_unprivileged_upload = false;
		}

		if ($this->isNew()) {
			if (!($this->userCanCreate() || (Session::getSession()->isLoggedIn() && $allow_unprivileged_upload))) {
				throw new PermissionDeniedException("Could not create DocmanFile object: permission denied");
			}
		} else {
			if (!($this->userCanUpdate() || (Session::getSession()->isLoggedIn() && $allow_unprivileged_upload))) {
				throw new PermissionDeniedException("Could not update DocmanFile object: permission denied");
			}
		}

		$isNew = $this->isNew();

		$params = array(
			'is_new' => $this->isNew(),
			'item' => $this
		);

		ActionManager::notify("DOCMAN_FILE_SAVE_BEFORE", $params);
		parent::save($con);
		ActionManager::notify("DOCMAN_FILE_SAVE", $params);
		// Notify there was a change in the file
		if ($do_notify) {
			$notification = new DocmanFileNotification($this, $this->file_name);
			$notification->setNew($isNew);
			NotificationManager::getManager()->queue($notification);
		}
	}

	function getDocumentUrl() {
		return GFUrl::getURL()->projectURL($this->getDocmanFolder()->getProject(),'','docman','',array('file_id'=>$this->getDocmanFileId()));
	}

	function getDocumentLink($full=false) {
		$url = $this->getDocumentUrl();
		if ($full) {
			return '<a href="'.$url.'">'.$this->getDescription().'</a>';
		} else {
			return $url;
		}
	}
	
	/**
	 * Returns the latest version of the document
	 * @return	DocmanFileVersion
	 */
	public function getLatestVersion($no_pending=true) {
		$c = new Criteria();
		$c->add(DocmanFileVersionPeer::DOCMAN_FILE_ID, $this->getDocmanFileId());
		if ($no_pending) {
			$c->add(DocmanFileVersionPeer::IS_APPROVED, true);
		}
		$c->addDescendingOrderByColumn(DocmanFileVersionPeer::CREATE_DATE);
		
		return DocmanFileVersionPeer::doSelectOne($c);
	}

	/**
	 * Returns the versions count
	 * @return	Int versions count
	 */
	public function getVersionsCount() {
		$c = new Criteria();
		$c->add(DocmanFileVersionPeer::DOCMAN_FILE_ID, $this->getDocmanFileId());
		$c->add(DocmanFileVersionPeer::IS_APPROVED, true);

		return DocmanFileVersionPeer::doCount($c);
	}


	
	/**
   	 * Tells if the user is monitoring this item
	 * @return bool
	 */
	public function userIsMonitoring(User $user) {
		$c = new Criteria();
		$c->add(MonitorPeer::SECTION, self::SECTION);
		$c->add(MonitorPeer::REF_ID, $this->getPrimaryKey());
		$c->add(MonitorPeer::USER_ID, $user->getUserId());
		$mon = MonitorPeer::doSelectOne($c);
		return (!is_null($mon)) ? true : false;
	}
	
	public function getMonitoringUsers() {
		$project_id = $this->getDocmanFolder()->getProject()->getProjectId();

		if (!$this->monitoring_users) {
			$con = Propel::getConnection(DocmanFilePeer::DATABASE_NAME);
			$sql = 'SELECT * FROM "user" WHERE user_id IN (
						SELECT distinct(user_id) FROM "user" WHERE user_id IN (
						SELECT user_id FROM monitor WHERE section=\''.self::SECTION.'\' AND ref_id='.$this->getDocmanFileId().'
						UNION
						SELECT user_id FROM monitor WHERE section=\''.DocmanPlugin::SECTION.'\' AND ref_id='.$project_id.'
						UNION
						SELECT user_id FROM monitor WHERE section=\''.DocmanFolder::SECTION.'\' AND ref_id IN 
							(SELECT docman_folder_id FROM docman_file_folder WHERE docman_file_id='.$this->getDocmanFileId().')
					))';
			$stmt = $con->createStatement();
			$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_NUM);
			$this->monitoring_users = UserPeer::populateObjects($rs);
		}
		
		return $this->monitoring_users;
	}

	public function getTopParentFolders() {
		$con = Propel::getConnection(DocmanFilePeer::DATABASE_NAME);
		$sql = 'SELECT * FROM docman_folder WHERE docman_folder_id IN (SELECT DISTINCT top_parent_folder_id FROM docman_folder
			WHERE docman_folder_id IN (SELECT docman_folder_id FROM docman_file_folder WHERE docman_file_id = '.
			$this->getDocmanFileId() .'))';
		$stmt = $con->createStatement();
		$rs = $stmt->executeQuery($sql, ResultSet::FETCHMODE_NUM);
		$top_parent_folders = DocmanFolderPeer::populateObjects($rs);
		return $top_parent_folders;
	}	

	public function checkMaximumAccessType($access_level) {
		//Get the folders that matters
		$folders = $this->getTopParentFolders();

		if(!count($folders)){
			return false;
		}
		//Take the first folder so we query for permissions only once
		$perm = $folders[0]->getProject()->getPermission();
		foreach($folders as $folder){
			if($perm->check(DocmanFolder::SECTION, $folder->getDocmanFolderId(),$access_level)){
				return true;
			}
		}
		return false;
	}

	public function getDownloadUrl() {
		$download_url = GFUrl::getURL()->getFullURL(GFUrl::getURL()->downloadUrlFromFS($this->getLatestVersion()->getFile(),true));
		return $download_url;
	}
}

?>
