<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Daniel Perez	daniel@gforgegroup.com
 * @author Marcelo Mottalli	<marcelo@gforgegroup.com>
 * @date 2006-05-08
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once("session/Login.interface.php");
require_once("exceptions/LoginException.php");
require_once("core/User.php");
require_once("core/UserPeer.php");
require_once("core/UserSession.php");
require_once('core/CountryCode.php');
require_once('core/Language.php');
require_once('core/Theme.php');

class LdapLogin implements Login {
	/**
	 * @return UserSession
	 */
	public function doLogin($username, $password, $checkStatus=true) {
		
		// Read variables from configuration
		$server = GFConf::get('ldap.ldapServer');
		$port = GFConf::get('ldap.ldapPort');
		$base_dn = GFConf::get('ldap.baseDn');
		$type = (GFConf::varExists('ldap.version')) ? GFConf::get('ldap.version') : '';
		$dbpref = GFConf::get('ldap.dbLoginPrefered');
			
		if (function_exists("ldap_connect")) {	
			$connection = ldap_connect($server,$port);	
		} else {
			throw new LoginException("You must have php compiled with --with-ldap in order to use ldap authentication<br />Please do so, or disable ldap auth");
		}

		if (!$connection) {
			throw new LoginException('Couldn\'t connect to LDAP server');
		}
		
		if (GFConf::varExists('ldap.version')) {
			ldap_set_option($connection, LDAP_OPT_PROTOCOL_VERSION, GFConf::get('ldap.version'));
		}
		
		// Windows 2003 support
		ldap_set_option($connection, LDAP_OPT_REFERRALS, 0);
		
		if (GFConf::varExists('ldap.useTLS') && GFConf::get('ldap.useTLS')) {
			if (!@ldap_start_tls($connection)) {
				throw new LoginException('Error starting TLS: '.ldap_error($connection));
			}
		}
		
		$username = strtolower($username);
		$c = new Criteria(UserPeer::DATABASE_NAME);
		$c->add(UserPeer::UNIX_NAME, $username);
		$user = UserPeer::doSelectOne($c);

		if (!is_object($user) || !$dbpref) {

			//may have to use standard username/pass to bind to ldap server
			if (GFConf::get('ldap.authenticatedBind')) {
				$ldap_bind_user = GFConf::get('ldap.authenticatedBindUser');
				if ($ldap_bind_user) {
					$bind = ldap_bind($connection, $ldap_bind_user, GFConf::get('ldap.authenticatedBindPass')); 
				} else {
					$ldap_bind_user = $username;
					$bind = ldap_bind($connection, $username, $password); 
				}
				if (!$bind) {
					throw new LoginException('Invalid password for user "'.$ldap_bind_user.'"');
					exit(1);
				}
			}
			
			// We either have to import the user from LDAP or we have to update his information
			
			// Search for the user's DN
			$filter = GFConf::get('ldap.usernameAttribute').'='.$username;
			$result = @ldap_search($connection, $base_dn, $filter);
			if (!$result) {
				throw new LoginException('Couldn\'t search LDAP server using base DN "'.$base_dn.'"');
			}
			$info = ldap_get_entries($connection, $result);
			
			if ($info['count'] == 0) {
				throw new LoginException('User "'.$username.'" is not in LDAP database');
			}
			
			// Bind as the selected user using the specified password
			$user_dn = $info[0]['dn'];
			$bind = @ldap_bind($connection, $user_dn, $password);
			if (!$bind) {
				throw new LoginException('Invalid password for user "'.$username.'"');
			}
			
			if (is_object($user)) {
				// Update the password in the database
				if (md5($password) != $user->getPasswordMd5()) {
					// DB password is different from LDAP's, update the DB one
					$user->setPassword($password);
					$user->save();
				}
			} else {
				// We need to import all the info from the database
				// Retrieve the user information
				$fields = array(GFConf::get('ldap.usernameAttribute'),
								GFConf::get('ldap.emailAttribute'),
								GFConf::get('ldap.firstnameAttribute'),
								GFConf::get('ldap.lastnameAttribute'));
				$result = ldap_search($connection, $base_dn, $filter, $fields);
				$info = ldap_get_entries($connection, $result);
				
				$user_data = $this->prepareData($info[0]);
				try {
					$user = new User();
					$user->setEmail($user_data['email']);
					$user->setFirstname($user_data['firstname']);
					$user->setLastname($user_data['lastname']);
					$user->setPassword($password);
					$user->setUnixName($username);
					$user->setThemeId($user_data['theme']->getPrimaryKey());
					$user->setLanguageId($user_data['language']->getPrimaryKey());
					$user->setCcode($user_data['country']->getID());
					$user->setConfirmHash(substr(md5(Session::getSession()->getSessionHash() . $password . time()),0,16));
					
					// user should be directly added to the database -- no confirmation needed
					// (note that if the configuration says that the administrator must aprove
					// users first, the user will NOT be active immediatly)
					$user->bypassConfirmation();
					$user->save();
					
					if ($checkStatus && $user->getStatus() != User::STATUS_ACTIVE) {
						throw new LoginException(Language::getSessText('Ldap.ImportedDataInactive', array($username)));
					}
				} catch (InvalidDataException $e) {
					$errors = $e->getErrors();
					$feedback = '* '.implode('<br />* ', $errors);
					throw new LoginException($feedback);
				} catch (Exception $e) {
					throw new LoginException($e->getMessage());
				}
			}
			
			// At this point we have the correct User object
			if ($checkStatus && $user->getStatus() != User::STATUS_ACTIVE) {
				throw new LoginException("User not active");
			}
		} else {
			// This means we've found the user in the database and that the database
			// takes precedence over the LDAP server. Don't update the user information.
			if (md5($password) == $user->getPasswordMd5()) {
				if ($checkStatus && $user->getStatus() != User::STATUS_ACTIVE) {
					throw new LoginException("User not active");
				}
			} else {
				throw new LoginException("Invalid password");
			}
		}
		
		// all checks passed, allow the session
		// See if we already have an user sessio
		$c = new Criteria();
		$c->add(UserSessionPeer::SESSION_HASH);
		$session = UserSessionPeer::doSelectOne($c);
		if (is_null($session)) {
			$session = new UserSession();
		}
		$session->setUser($user);
		
		return $session;
	}
	
	/**
	 * Prepares the data from ldap and some defaults to insert the user into the db
	 *
	 * @param array $data
	 * @return array
	 */
	private function prepareData($data) {
		$result = array();
		
		$firstnameAttribute = GFConf::get('ldap.firstnameAttribute');
		if (!array_key_exists($firstnameAttribute, $data)) {
			throw new LoginException('Invalid first name field "'.$firstnameAttribute.'" when importing data from LDAP');
		} else {
			$result['firstname'] = $data[$firstnameAttribute][0] ;
		}
		
		$lastnameAttribute = GFConf::get('ldap.lastnameAttribute');
		if (!array_key_exists($lastnameAttribute, $data)) {
			throw new LoginException('Invalid last name field "'.$lastnameAttribute.'" when importing data from LDAP');
		} else {
			$result['lastname'] = $data[$lastnameAttribute][0] ;
		}

		$emailAttribute = GFConf::get('ldap.emailAttribute');
		if (array_key_exists($emailAttribute, $data)) {
			$result['email'] = $data[$emailAttribute][0];
		} else {
			$usernameAttribute = GFConf::get('ldap.usernameAttribute');
			$result['email'] = $data[$usernameAttribute][0] . '@' . GFConf::get('system.hostName') ;
		}
		
		// if any is set, use the id of country/language/theme, otherwise use the first one
		$ccode = (GFConf::varExists('ldap.defaultCountryCode')) ? GFConf::get('ldap.defaultCountryCode') : 'US';
		$themeid = (GFConf::varExists('ldap.defaultThemeId')) ? GFConf::get('ldap.defaultThemeId') : GFConf::get('system.defaultThemeId');
		$languageid = (GFConf::varExists('ldap.defaultLanguageId')) ? GFConf::get('ldap.defaultLanguageId') : GFConf::get('system.defaultLanguageId');
		
		$con = Propel::getConnection(UserPeer::DATABASE_NAME);
		
		// get the config country
		$c = new Criteria();
		$c->add(CountryCodePeer::CCODE,$ccode);
		$country = CountryCodePeer::doSelectOne($c);

		// get the config theme
		$theme = ThemePeer::retrieveByPk($themeid);
		
		// get the config theme
		$language = LanguagePeer::retrieveByPk($languageid);

		if (!($theme instanceof Theme)) {
			throw new LoginException('Invalid default theme in LDAP configuration');
		}
		if (!($language instanceof Language )) {
			throw new LoginException('Invalid default language in LDAP configuration');
		}
		if (!($country instanceof CountryCode )) {
			throw new LoginException('Invalid default country in LDAP configuration');
		}
		
		$result['country'] = $country;
		$result['language'] = $language;
		$result['theme'] = $theme;
		
		return $result ;		
	}
	
	public function doLogout() {
	}
}
?>
