<?php
/**
 * Copyright (c) 2005-2013 GForge, LLC
 *
 * @version   $Id$
 * @author Mario A. Papayannis	<mpapayannis@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('plugin/GFPlugin.php');
require_once('plugin/ActionManager.php');
require_once('plugin/PluginResult.php');
require_once('plugins/pastebin/lib/PastebinPeer.php');
require_once('plugins/pastebin/lib/Pastebin.php');
require_once('core/Language.php');


class PastebinPlugin implements GFPlugin {

	CONST SECTION = 'pastebin';

	function __construct() {
	}

	/**
	 *	this is called when the first Action is called. Action manager
	 *	gets the listeners for each plugin.
	 */
	public function initialize() {
		ActionManager::registerListeners($this, $this->getListeners());
	}


	/**
	 *	Different types of plugins are possible -
	 *	GFPlugin::SITE_PLUGIN = 'site'
	 *	GFPlugin::PROJECT_PLUGIN = 'project'
	 *	GFPlugin::USER_PLUGIN = 'user'
	 */
	public function getType() {
		return GFPlugin::SITE_PLUGIN;
	}

	/**
	 *	get the name of this plugin. example: 'tracker'
	 *	@return	string
	 */
	public function getName() {
		return 'pastebin';
	}

	/**
	 *	get the array of action names that this plugin answers to.
	 *	@return	array
	 */
	public function getListeners() {
		return array(
			/*'BUILD_SEARCH_COMBO',
			'GET_PROJECT_REPORTING_NAV_ELEMENTS',
			'TRACKER_ITEM_SAVE',
			'SCM_COMMIT_SAVE',*/
			'BUILD_MAIN_TABS',
			'RENDER_PLUGIN_LINK',
			'CHAT_MESSAGE_RENDER',
			'BUILD_SEARCH_COMBO',
			'DO_SEARCH',
			'DO_TAG_SEARCH'
		);
	}

	/**
	 *	handle action has a case statement to redirect the action to the appropriate function.
	 *	@param string
	 *	@param mixed
	 *	@return PluginResult
	 */
	public function handleAction($listenerName, $params) {
		switch ($listenerName) {
			case 'BUILD_MAIN_TABS': {
				return $this->_buildMainTabsListener($params);
				break;
			}
			case 'BUILD_SEARCH_COMBO': {
				return $this->buildSearchOptions($params);
				break;
			}
			case 'DO_SEARCH': {
				return $this->getSearchResults($params);
				break;
			}
			case 'RENDER_PLUGIN_LINK': {
				return $this->renderPluginLink($params);
				break;
			}
			case 'CHAT_MESSAGE_RENDER': {
				return $this->renderPasteFromText($params);
				break;
			}
			case 'DO_TAG_SEARCH': {
				return $this->getSearchByTagsResult($params);
				break;
			}
		}
	}


	private function getSearchByTagsResult($params) {
		if (array_search($this->getName(), $params['search']) !== false) {
			$result = new PluginResult($this);
			$searcher = new GFSearch($params);
			$result->setResult($this->searchByTags($searcher));
			return $result;
		}
	}


	/**
	 *
	 * return the
	 *
	 * @param array $words_arr  the list of tags
	 * @param integer $project_id if null, then the search is a global search
	 *
	 * @return ResultSet
	 */

	public function searchByTags(GFSearch $search){

		$words = $search->getWordsForQuery();
		$project_id = $search->getProjectId();
		$template = $search->getTemplate();

		//Clean key words from tsearch operators
		$words = preg_replace('/,+/',',', $words);
		$tsearch_operators = array('(', ')', '!', '|','&');
		$replacements = array('','','','',',');
		$words = str_replace($tsearch_operators, $replacements, $words);
		$words_arr = explode(',', $words);

		$wc = count($words_arr);
		if($wc == 0){
			return '';
		}
		for($i = 0; $i < $wc; $i++){
			$words_arr[$i] = "'" . $words_arr[$i] . "'";
		}
		$words = implode(',', $words_arr);

		if($project_id != 0){
			$pastebin_prj = " AND p.ref_id=$project_id";
		} else {
			$pastebin_prj = ' ';
		}

		$pastebin_sql = "SELECT DISTINCT p.*
		FROM pastebin p
		WHERE p.pastebin_id IN
		(SELECT tag_entry.ref_id FROM tag_entry WHERE section='pastebin'
		AND tag_name IN ($words)
		$pastebin_prj
		GROUP BY tag_entry.ref_id
		HAVING COUNT(ref_id) =  $wc)";

		$con = Propel::getConnection();

		$stmt = $con->createStatement();
		$rs = $stmt->executeQuery($pastebin_sql, ResultSet::FETCHMODE_ASSOC);


		$Lang = Session::getSession()->getLanguage();
		if(!$project_id){
			$content = $template->tableTop(array(
				'#',
				$Lang->getText('Project.ListNameTitle'),
				$Lang->getText('Pastebin.Title'),
				$Lang->getText('Pastebin.Language')
			));
		} else{
			$content = $template->tableTop(array(
				'#',
				$Lang->getText('Pastebin.Title'),
				$Lang->getText('Pastebin.Language')
			));
		}

		$count = 0;
		if($rs->getRecordCount() > 0){
			while($rs->next()){
				$trow = array();

				$paste = PastebinPeer::retrieveByPK($rs->getInt('pastebin_id'));

				if (!is_object($paste) || !$paste->userCanRead()) {
					continue;
				}

				$trow[] = $paste->getPastebinId();

				if(!$project_id && $paste->getExposureType() == Pastebin::EXPOSURE_PROJECT && $paste->getRefId() != 0) {
					$project = ProjectPeer::retrieveByPk($paste->getRefId());
					if (is_object($project)) {
						$trow[] = '<a href="' . GFUrl::getUrl()->projectUrl($project->getProjectId()) . '">'. $project->getProjectName() . '</a>';
					}
				}

				$trow[] = '<a href="' . $paste->getPasteUrl() . '">'. $paste->getTitle() . '</a>';
				$trow[] = $paste->getLanguage();
				$content .= $template->tableRow($trow);
				$count++;
			}
		}
		$content .= $template->tableBottom();
		$title = $Lang->getText('Pastebin.SearchResultsTabTitle', array($count));
		if ($count == 0) {
			$content = $Lang->getText('Search.NoResults');
		}

		return array(
			'title' => $title,
			'count' => $count,
			'content' => $content
		);

	}



	/**
	 *	take a search object and handle the search internally.
	 *	@param	GFSearch
	 *	@return string
	 */
	public function search(GFSearch $search) {

		$sql = $this->getSearchSql($search, GFConf::get('system.dbtype'));

		$con = Propel::getConnection();
		$rs = $con->executeQuery($sql, ResultSet::FETCHMODE_ASSOC);

		$ret = $this->getSearchContent($search, $rs);

		return $ret;
	}


	private function getSearchSql($search, $dbType = 'postgres') {

		$sql = '';

		if (GFConf::get('system.dbtype') == 'postgres') {
			$con = Propel::getConnection(UserPeer::DATABASE_NAME);
			$words = $search->getWordsForQuery();

			if ($search->getStartDate()) {
				$date_sql = " AND post_date BETWEEN '".$search->getStartDate()."' AND '".$search->getEndDate()."' ";
			} else {
				$date_sql='';
			}

			$sql = "SELECT p.*, p.title AS headline_message
					FROM pastebin AS p
					WHERE
						p.title LIKE '%$words%' OR
						p.content LIKE '%$words%' OR
						p.language = '$words'
					";

			$sql .= $date_sql;
		}

		return $sql;
	}


	private function getSearchContent($search, $rs) {
		$template = $search->getTemplate();

		$count = 0;
		$content = $template->tableTop(array(
			'#',
			Language::getSessText('Pastebin.Title'),
			Language::getSessText('Pastebin.Language')
		));
		while ($rs->next()) {
			$pastebin_id = $rs->getInt('pastebin_id');
			$paste = PastebinPeer::retrieveByPK($pastebin_id);
			//$paste->hydrate($rs, 6);

			if (!$paste->userCanRead()) {
				continue;
			}

			$tid = $paste->getPastebinId();
			$trow = array();
			$trow[] = $tid;
			$trow[] = '<a href="'.GFUrl::getUrl()->baseURL('', 'pastebin', '', array()) . '/' . $paste->getPastebinId() . '">' . $paste->getTitle().'</a>';
			$trow[] = $paste->getLanguage();
			$content .= $template->tableRow($trow);
			$count++;
		}
		if ($count == 0) {
			$content .= $template->tableFullRow(Language::getSessText('Search.NoResults'));
		}
		$content .= $template->tableBottom();


		$ret = array(
			'title' => Language::getSessText('Pastebin.SearchResultsTabTitle', array($count)),
			'count' => $count,
			'content' => $content
		);

		return $ret;
	}


	/**
	 * returns the options to print in the combobox
	 *
	 * @param unknown_type $params
	 * @return unknown
	 */
	private function buildSearchOptions($params) {
		$result = new PluginResult($this);
		$value = array();
		if (isset($params['project'])) {
			$value[] = array(
				'value'	=> $this->getName(),
				'display'	=> Language::getSessText('Pastebin.ProjectPastes')
			);
		} else {
			$value[] = array(
				'value'	=> $this->getName(),
				'display'	=> Language::getSessText('Pastebin.Name')
			);
		}

		$result->setResult($value);
		return $result;
	}


	/**
	 * returns the search results for the given input
	 *
	 * @param array $params
	 * @return resultset
	 */

	private function getSearchResults($params) {
		if (array_search($this->getName(), $params['search']) !== false) {
			$result = new PluginResult($this);
			$searcher = new GFSearch($params);
			$result->setResult($this->search($searcher));
			return $result;
		}
	}


	/**
	 * returns an array of the fields where we want the search to be done upon
	 *
	 */
	/*
	private function getSearchFields() {
		return null;
	}
*/
	private function _buildMainTabsListener($params) {
		$result = new PluginResult($this);
		$value = array(
			'text'	=> Language::getSessText('Pastebin.TabString'),
			'action'	=> GFUrl::getUrl()->baseUrl('',$this->getName()),
			'id'	=> 'pastebin'
		);
		$result->setResult($value);
		return $result;
	}

	/**
	 *      getProjectAdminNavElements - get the local nav elements for the tracker
	 *      @return PluginResult
	 */
	private function getProjectAdminNavElements($params) {
		return null;
	}


	/**
	 *      renderPluginLink - renders a pastebin link for the chat
	 *      @return PluginResult
	 */
	function renderPluginLink($params) {
		if ($params['gf_url']->getPlugin()->getName() != PastebinPlugin::SECTION) {
			return '';
		}

		$url = $params['url'];

		$parsed = parse_url($url);
		$path = $parsed['path'];
		$arr_path = explode("/", $path);
		if (sizeof($arr_path) == 4) {
			$pastebin_hash = $arr_path[3];
			$pastebin_id = PastebinPeer::convertPasteHash($pastebin_hash);
			$pastebin = PastebinPeer::retrieveByPK($pastebin_id);
			if (is_object($pastebin) && $pastebin->userCanRead()) {
				$res = $this->renderMiniPastebinView($pastebin);
			} else {
				return '';
			}
		} else {
			return '';
		}

		$pr = new PluginResult($this);
		$pr->setResult($res);		// Mark the request as handled
		return $pr;
	}


	private function renderMiniPastebinView($pastebin) {
		$session = Session::getSession();
		$Lang = $session->getLanguage();
		$post_user = UserPeer::retrieveByPk($pastebin->getPostedBy());

		ob_start();
		require_once("plugins/pastebin/wwwlib/forms/PastebinDisplayMini-tpl.php");
		$content = ob_get_contents();
		ob_end_clean();

		$res = '
					<div class="link-summary pastebin">
						' . $content . '
					</div>';

		return $res;
	}


	/**
	 * @param array $params
	 * @return array|bool|string
	 */
	private function renderPasteFromText($params) {
		$ret = '';
		preg_match('/\\[\\#code( |)(?:(language="[a-zA-Z]*")?)(.*)\\]/', $params['body'], $matches);
		if (isset($matches[0])) {
			$matches_language = strtolower(trim($matches[2]));
			preg_match('/language="(.*)"/', $params['body'], $matches_language);
			$strLanguageCode = '';
			if (isset($matches_language[1])) {
				$strLanguageCode = $matches_language[1];
				$pos = strpos($strLanguageCode, '"');
				if ($pos > 0) {
					$strLanguageCode = substr($strLanguageCode, 0, $pos);
				}
			}
			$strContent = trim($matches[3]);

			if ($strContent != '') {
				$session = Session::getSession();
				$post_user = $session->getUser();

				$pastebin = new Pastebin();
				$pastebin->setTitle('Autopaste from chat');
				$pastebin->setContent( $strContent );
				$pastebin->setPostedBy( $post_user->getUserId() );
				$pastebin->setExposureType( Pastebin::EXPOSURE_PUBLIC );
				$pastebin->setLanguage( $strLanguageCode );
				$pastebin->setExpiration( Pastebin::EXPIRATION_NEVER );
				if ($pastebin->userCanWrite()) {
					$pastebin->save();
					$ret = $this->renderMiniPastebinView($pastebin);
				}
			}
		}

		$pr = new PluginResult($this);
		$pr->setResult($ret);

		return $pr;
	}

}
