<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli	marcelo@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('SystemUtils.php');
require_once('plugins/scm/lib/SCMRepository.php');
require_once('session/Session.php');

/**
 * RepositoryBrowser - Class that encapsulates repository browsers (currently, only
 * CVS and SVN browsing using ViewVC is supported)
 */
class RepositoryBrowser {
	private $repos;
	private $type;
	private $environ = array();
	private $querystring;
	private $isHtml;
	private $contentType;
	private $content;
	private $headers = array();
	
	const TYPE_MARKUP = 'markup';
	const TYPE_ANNOTATE = 'annotate';
	
	public function __construct(SCMRepository $repos) {
		$type = strtolower($repos->getType());
		
		$supported_types = array('cvs','svn','perforce','clearcase','git');
		if (!in_array($type,$supported_types)) {
			throw new Exception('Unsupported browsing of "'.$type.'" repository');
		}
		
		$this->type = $type;
		$this->repos = $repos;
	}
	
	private function buildEnvironment($vars) {
		if (array_key_exists('path', $vars)) {
			// $path MUST end with a /
			if (!preg_match('/\\/$/', $_GET['path'])) {
				$path = $_GET['path'].'/';
			} else {
				$path = $_GET['path'];
			}
			$path = urldecode($path);
		} else {
			$path = '/';
		}
		
		$this->checkPathPermissions($path);		
		$this->environ['PATH_INFO'] = $path;
		
		// Build a new querystring to pass to ViewVC
		$querystring = 'root='.$this->repos->getRepositoryName();
		foreach ($vars as $name => $val) {
			// these are our variables
			if ($name == 'path') continue;
			if ($name == 'action') continue;
			if ($name == 'feedback') continue;
			
			// this means the parameter is viewvc-specific
			$querystring .= '&'.$name.'='.urlencode($val);
		}
		$this->environ['QUERY_STRING'] = $querystring;
		$this->environ['HOME'] = '/tmp';
		
		$this->environ['REPOSITORY_TYPE'] = $this->repos->getType();
		$this->environ['REPOSITORY_NAME'] = $this->repos->getRepositoryName();
		$this->environ['REPOSITORY_ROOT'] = $this->repos->getRoot();
		$this->environ['HOSTNAME'] = GFConf::get('system.hostName');
	}
	
	public function checkPathPermissions($path) {
		// Security check: don't allow ".." into the path, otherwise
		// ViewVC can show ANY file on the filesystem
		if (strpos($path, '../') !== false) {
			throw new PermissionDeniedException();
		}
		
		if (!$this->repos->userHasAccess($path)) {
			throw new PermissionDeniedException('You don\'t have access to this path');
		}
	}
	
	public function parseQuery($querystring) {
		// For Perforce we don't use ViewVC
		if ($this->type == 'perforce') {
			require_once('plugins/scmperforce/lib/PerforceBrowser.php');
			$browser = new PerforceBrowser($this->repos);
			return $browser->parseQuery($querystring);
		}
		// Same for Git
		if ($this->type == 'git') {
			require_once('plugins/scmgit/lib/GitwebBrowser.php');
			$browser = new GitwebBrowser($this->repos);
			return $browser->parseQuery($querystring);
		}
		
		
		// read the environment variables for ViewVC from the query string
		$vars = array();
		parse_str($querystring, $vars);
		
		// Should we browse on a remote machine? (remote ViewVC)
		if ($this->remoteBrowse()) {
			return $this->doRemoteBrowse($vars);
		}
		
		// Use local ViewVC
		
		$this->buildEnvironment($vars);

		// Locate viewvc.cgi
		$viewvc_install = SystemUtils::buildFilePath(GFConf::get('system.libDir'), '/lib/external/viewvc/');
		$viewvc = SystemUtils::buildFilePath($viewvc_install, '/bin/cgi/viewvc.cgi');
		if (!file_exists($viewvc)) {
			throw new Exception('ViewVC must be installed to '.$viewvc_install);
		}
		
		// Build the command to execute
		$cmd = '';
		foreach ($this->environ as $var => $val) {
			$cmd .= $var.'='.escapeshellarg($val).' ';
		}
		$cmd .= $viewvc;
		$cmd .= ' 2>&1';		// redirect stderr to stdout
		
		// Execute ViewVC
		$output = '';
		// We don't want the output to be exploded since it may be a binary result.
		// PHP has a nasty problem where it treats both \n and \r\n as new lines, and
		// it removes both of them, so we don't have a way to reconstruct binary files
		$result = SystemUtils::execute($cmd, $output, false);
		$output = explode("\n", $output);

		if ($result != 0) {		// error
			$this->isHtml = true;
			$this->content = implode("\n", $output);
			return;
		}
		
		// Parse the headers
		$eoh = false;
		$this->contentType = 'text/html';
		$this->headers = array();
		while (true) {		// delete the headers
			$line = trim(array_shift($output));
			
			// First empty line marks the end of the HTTP headers
			$eoh = empty($line);
			if ($eoh) {
				// End of headers have TWO empty lines, so we must remove the next one
//				array_shift($output);
				break;
			}

			if (preg_match('/Content-type: (.*)/i', $line, $matches)) {
				$this->contentType = $matches[1];
			}
			$this->headers[] = $line;

		};
		
		if (strpos($this->contentType, 'text/html') === false) {
			$this->isHtml = false;
			// No further parsing is needed
			$this->content = implode("\n", $output);
			return;
		}
		
		// Now we need to parse the HTML generated by ViewVC in order to change some URLs
		$this->isHtml = true;
		$this->content = $this->parseOutput($output);
	}
	
	private function parseOutput($output) {
		$theme = Session::getSession()->getTheme();
		$newOutput = array();

		for ($i=0; $i < count($output); $i++) {
			$line = $output[$i];
			
			// replace all the references to images

			if (preg_match_all('/\\/\\*docroot\\*\\/images\\/([^\'""]*)/',$line,$matches)) {
				foreach ($matches[1] as $image) {
					$line = str_replace('/*docroot*/images/'.$image, $theme->getImgRoot().'/viewvc/'.$image, $line);
				}
			}

			//fix logo link
			if (strpos($line, 'href=".."')) {
				$line = str_replace('".."', '"/"', $line);

			}
			$line = str_replace("images/logo.png", $theme->getImgRoot().'/viewvc/logo.png', $line);
			//links to static help pages
			if (preg_match_all('/help_.+html/', $line, $matches) && !strpos($line, 'docroot')) {
				foreach ($matches[0] as $link) {
					$line = str_replace($link, '/*docroot*/'.$link, $line);
				}
			}
			
			// remove reference to stylesheet (it should be included in the theme stylesheet)
			$line = preg_replace('/<link [^>]*rel=.stylesheet.[^>]*\\/>/i', '', $line);
			
			$keep_parsing_line = true;
			
			// if we are viewing a file, don't parse the contents
			if (isset($_GET['view']) && $_GET['view'] == 'markup' && strpos($line, '<div id="vc_markup"><pre>') !== false) {
				while (true) {
					$line = $output[$i];
					$newOutput[] = $line;
					if (strpos($line, '</pre></div') !== false) {
						break;
					} else {
						// Increase $i until we find the end
						$i++;
					}
				}
			}
			
			// viewing a diff or annotate... skip the contents
			if (strpos($line, '<td class="vc_diff_nochange">') !== false ||
					strpos($line, '<td class="vc_diff_empty">') !== false ||
					strpos($line, '<td class="vc_diff_change">') !== false ||
					strpos($line, '<td class="vc_diff_change_empty">') !== false ||
					strpos($line, '<td class="vc_diff_remove">') !== false ||
					strpos($line, '<td class="vc_diff_add">') !== false ||
					strpos($line, '<td class="vc_blame_text">') !== false) {
				
				// Don't parse until the </td> is found
				while (true) {
					$line = $output[$i];
					$newOutput[] = $line;
					if (strpos($line, '</td>') !== false) {
						$keep_parsing_line = false;
						break;
					} else {
						$i++;
					}
				}
			} elseif (strpos($line, '<pre class="vc_raw_diff">') !== false) {	// raw diff
				// Don't parse until the </pre> is found
				while (true) {
					$line = $output[$i];
					$newOutput[] = $line;
					if (strpos($line, '</pre>') !== false) {
						$keep_parsing_line = false;
						break;
					} else {
						$i++;
					}
				}
			}
			
			if (!$keep_parsing_line) {
				// Move to the next line, don't parse the links
				continue;
			}
			
			if (preg_match_all('/href=[\'"]([^\'"]*)[\'"]/i', $line, $matches)) {		// link found, we must rewrite it
				foreach ($matches[1] as $fullLink) {
					// external link -- skip it
					if (preg_match('/^http/', $fullLink)) continue;
					
					$link = preg_replace('/&amp;/', '&', $fullLink);	// this must be done or else parse_str will do weird stuff
					$parts = explode('?', $link, 2);
					$path = $parts[0];
					// little hack to remove viewvc's path from the path
					if ($this->remoteBrowse()) {
						$path = str_replace($this->getRemoteViewVCPath(), '', $path);
					}
					
					// parameters in the URL
					$params = array();
					$params['path'] = $path;
	
					if (isset($parts[1])) {
						// there are some GET parameters
						$params2 = array();
						parse_str($parts[1], $params2);	// read the parameters
						unset($params2["root"]);		// we can leave this but it is unnecesary
						$params = array_merge($params, $params2);	// the order in which this arrays are merged is important!
						// we must put our variables first ("action", "path", etc)
					}
					
					$newLink = GFUrl::getUrl()->currentBaseURL('browse', $params);
					// small hack to convert the "#" to the real character instead of %23
					$newLink = str_replace('%23', '#', $newLink);
					$line = str_replace($fullLink, $newLink, $line);
					
				}
				$newOutput[] = $line;
			} elseif (preg_match('/action=[\'"]([^\'"]*)[\'"]/i', $line, $matches)) {
				// Found a form... we must pass the variables in a hidden form
				$oldLink = $matches[1];
				$newLink = GFUrl::getUrl()->currentBaseURL('');		// Action will be inserted as a hidden variable
				
				$line = str_replace($oldLink, $newLink, $line);
				$newOutput[] = $line;
				$newOutput[] = '<input type="hidden" name="action" value="browse" />';
				
				if ($this->remoteBrowse()) {
					$path = str_replace($this->getRemoteViewVCPath(), '', $oldLink);
				} else {
					$path = $oldLink;
				}
				$newOutput[] = '<input type="hidden" name="path" value="'.$path.'" />';
			} else if (preg_match('/<h.>Index of (.*)<\\/h.>/', $line, $matches) && $this->repos->userIsAdmin()) {
				$url = GFUrl::getUrl();
				$newOutput[] = $line;

				// Small hack to add a link to the modules editor
				$path = $matches[1];
				if ($path != '/') {					
					$link = GFUrl::getUrl()->projectURL($this->repos->getProject(), 'ScmModuleUserEdit', $url->getPlugin()->getName(), 'admin', array('path' => $path, 'redirect_to' => $url->currentUrl()));
					$button = HTMLWidget::button("button",   
																				Language::getSessText('ScmModuleUser.EditThis'),
																				"",
																				"",
																				"window.location = '$link'");
					$newOutput[] = $button;
				}
			} else {
				preg_match_all('/\\[#(\d+)\\]/', $line, $matches);
				if (count($matches) > 1) {
					// Found a tracker item ID like [#nnn]
					$tracker_item_ids = $matches[1];
					foreach ($tracker_item_ids as $tracker_item_id) {
						$TrackerItem = TrackerItemPeer::retrieveByPk($tracker_item_id);
						if (is_object($TrackerItem)) {
							$link = GFUrl::getUrl()->projectUrl($this->repos->getProject(), 'TrackerItemEdit', 'tracker', '', array('tracker_item_id' => $tracker_item_id));
							$line = preg_replace('/\\[#'.$tracker_item_id.'\\]/', '<a href="'.$link.'">[#'.$tracker_item_id.']</a>', $line);
						}
					}
				}
				$newOutput[] = $line;
			}
		}
		
		// If we output html and we found the mbstring extension, we
		// should try to encode the output of ViewVC in UTF-8
		$content = implode("\n", $newOutput);
		if (extension_loaded('mbstring')) {
			$encoding = mb_detect_encoding($content);		// Note: this may return false if encoding cannot be detected
			if ($encoding && $encoding != 'UTF-8') {
				$content = mb_convert_encoding($content, 'UTF-8', $encoding);
			}
		}
		
		return $content;
	}
	
	public function getContentType() {
		return $this->contentType;
	}
	
	public function isHtml() {
		return $this->isHtml;
	}
	
	public function getContents() {
		return $this->content;
	}
	
	public function getHTTPHeaders() {
		return $this->headers;
	}
	
	/**
	 * Returns the URL of the page used for browsing a file or directory's details
	 * @param	string $path	Path to browse
	 */
	public function getBrowsePathUrl($path, $revision=0) {
		$params = array();
		if ($path instanceof ScmCommitFile) {
			$params['path'] = $path->getModifiedFile();
			//For subversion when a file we should create a link 
			if(strtolower($this->repos->getType()) == 'svn'){
				if($path->getNewVersion() == 'deleted' || 
					$path->getNewVersion() == '' || 
					$path->getNewVersion() == 'none'){
						$params['pathrev'] = $path->getPrevVersion();
				} else {
						$params['pathrev'] = $path->getNewVersion();
				}

				return $this->buildBrowseUrl('browse',$params);
			}
	
			// Note for git: we need to pass the hash number, otherwise we don't know
			// which branch the file is into
			if (strtolower($this->repos->getType()) == 'git') {
				if (!$revision) {
					$params['newversion'] = $path->getNewVersion();
				} else {
					$params['newversion'] = $revision;
				}
			} elseif ($revision) {
				$params['pathrev'] = $revision;
			}
		} else {
			$params['path'] = $path;
			if (strtolower($this->repos->getType()) == 'git' && $revision) {
				$params['newversion'] = $revision;
			} elseif ($revision) {
				$params['pathrev'] = $revision;
			}
		}
		
		return $this->buildBrowseUrl('browse', $params);
	}

	public function getCommitBrowseUrl() {
		return $this->buildBrowseUrl('ScmCommitBrowse', array());
	}
	
	public function getCommitDetailUrl($commit_id) {
		return $this->buildBrowseUrl('ScmCommitDetail', array('scm_commit_id' => $commit_id));
	}

	/**
	 * @param	array
	 */
	private function buildBrowseUrl($action, $params) {
		$pluginName = $this->getPluginName();
		
		return GFUrl::getUrl()->projectUrl($this->repos->getProject(), $action, $pluginName, '', $params);
	}
	
	private function getPluginName() {
		$type = strtolower($this->repos->getType());
		$pluginName = null;
		
		if ($type == 'cvs') $pluginName = 'scmcvs';
		elseif ($type == 'svn') $pluginName = 'scmsvn';
		elseif ($type == 'perforce') $pluginName = 'scmperforce';
		elseif ($type == 'clearcase') $pluginName = 'scmclearcase';
		elseif ($type == 'git') $pluginName = 'scmgit';
		
		return $pluginName;
	}
	
	/**
	 * Returns the URL that displays the diff between two revisions
	 * @param	string $file	File to show diff
	 * @param	string $revfrom	First revision
	 * @param	string $revto	Second revision
	 */
	public function getDiffUrl($file, $revfrom, $revto) {

		$params = array('path' => $file, 'r1' => $revfrom, 'r2' => $revto);

		//Override default params according to scm

		if($this->type == 'svn'){
			$params['pathrev'] = $revto;
		}

		return $this->buildBrowseUrl('browse', $params);
	}
	
	/**
	 * URL for viewing a file's source
	 * @param	string $file
	 */
	public function fileSourceUrl($file, $viewType=self::TYPE_MARKUP) {
		return $this->buildBrowseUrl('browse', array('path' => $file, 'view' => $viewType));
	}
	
	/**
	 * Whether we should browse a local or a remote repository (don't confuse with "external" repository)
	 * @return bool
	 */
	private function remoteBrowse() {
		if (GFConf::get('system.splitScmServers')) {
			// If we're using a split SCM server setup, check if the repository is stored in this server
			// (in that case no remote browsing is needed)
			
			$remoteServerId = $this->repos->getScmConfiguration()->getScmServerId();
			
			$pluginName = $this->getPluginName();
			if (GFConf::varExists($pluginName.'.scmServerId') && GFConf::get($pluginName.'.scmServerId') == $remoteServerId) {
				return false;			// This means the "remote" server is actually this server
			}
			
			return true;
		}
		
		return false;
	}
	
	private function doRemoteBrowse($params) {
		$remoteHost = $this->repos->getScmConfiguration()->getScmServer()->getHostname();
		$remoteViewVCUrl = 'http://'.$remoteHost.$this->getRemoteViewVCPath();
		
		$this->invokeExternalViewVC($remoteViewVCUrl, $params);
	}
	
	private function getRemoteViewVCPath() {
		return '/cgi-bin/viewvc.cgi';
	}
	
	private function invokeExternalViewVC($url, $params) {
		if (isset($params['path'])) {
			if ($params['path'][0] != '/') {		// The path must always start with /
				$params['path'] = '/'.$params['path'];
			}
		} else {
			$params['path'] = '/';
		}
		
		$this->checkPathPermissions($params['path']);
		$url .= $params['path'];
		
		$querystring = 'root='.$this->repos->getRepositoryName();
		foreach ($params as $name => $val) {
			// these are our variables
			if ($name == 'path') continue;
			if ($name == 'action') continue;
			if ($name == 'feedback') continue;

			// this means the parameter is viewvc-specific
			$querystring .= '&'.$name.'='.urlencode($val);
		}
		
		$url .= '?'.$querystring;
		
		// We have the full URL now. Retrieve the data...
		$fh = @fopen($url, 'r');
		if (!$fh) {
			// Fake an error
			$this->isHtml = true;
			$this->content = Language::getSessText('Scm.ErrorRemoteBrowsing', array($url));
			return;
		}
		
		// Read the output of the script in the remote server
		$content = '';
		while (!feof($fh)) {
			$content .= fread($fh, 8192);
		}
		
		// fopen doesn't return the headers, so we must check if it's HTML indirectly
		//if (isset($params['path']) && (strpos($params['path'], '*checkout*') !== false)) {
		if ( (isset($params['view']) && $params['view'] == 'co') ||
			(isset($params['path']) && strpos($params['path'], '*checkout*') !== false) ||
			(isset($params['action']) && $params['action'] == 'browse' && isset($params['revision']) && !isset($params['view']))
			) {
			$this->isHtml = false;
			$this->content = $content;
			$this->contentType = 'application/octet-stream';
			$this->headers[] = "Content-type: $this->contentType";
		} else {
			// Now we need to parse the HTML generated by ViewVC in order to change some URLs
			$this->isHtml = true;
			$lines = explode("\n", $content);
			$this->content = $this->parseOutput($lines);
		}
	}
}
?>
