<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli	marcelo@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
 

require_once("system/Cronjob.php");
require_once('core/CronHistory.php');
require_once('plugins/scmcvs/lib/CVSRepository.php');
require_once('plugins/scm/lib/ScmProjectStats.php');
require_once('plugins/scm/lib/ScmUserStats.php');

// this could be a lengthy process
set_time_limit(0);

/**
 * Cronjob that parses the CVSROOT/history files
 */
class ParseCVSHistoryCronjob extends Cronjob {
	
	public function __construct() {
		$this->jobname = basename(__FILE__);
		parent::__construct();
	}	
	
	public function run() {
		// lets get all projects that use the plugin
		$c = new Criteria();
		$c->add(PluginPeer::PLUGIN_NAME, 'scmcvs');
		$project_plugins = ProjectPluginPeer::doSelectJoinPlugin($c);
		
		foreach ($project_plugins as $project_plugin) {
			$project = $project_plugin->getProject(); // get the project
			$repos = new CVSRepository($project);
			
			// repository hasn't been created yet... there isn't a history file
			if (!$repos->isCreated()) continue;
			
			$root = $repos->getRoot();
			$historyFile = SystemUtils::buildFilePath($root, '/CVSROOT/history');
			if (!file_exists($historyFile)) {
				$this->logCron('History file '.$historyFile.' doesn\'t exist for repository');
				continue;
			}
			
			$this->parseHistory($historyFile, $project);
		}
		
		$this->commitLog();
	}
	
	public function parseHistory($historyFile, Project $project) {
		$fh = fopen($historyFile, 'r');
		if (!$fh) {
			$this->logCron('Couldn\'t open history file '.$historyFile.' for reading');
			return;
		}
		
		$days = array();
		
		// See which day(s) we want to parse
		$day_parameter = $this->getCronjobParameter(0);
		if (strtolower($day_parameter) == 'all') {
			// Parse all days?
			$parse_all = true;
		} else {
			$parse_all = false;
			
			$days_to_parse = array();			
			
			if ($day_parameter && ($time = strtotime($day_parameter)) !== -1) {
				// Parse a particular day
				$parse_day = gmmktime(0, 0, 0, gmstrftime("%m", $time ), gmstrftime("%d", $time ), gmstrftime("%Y", $time));
				$days_to_parse[] = $parse_day;
			} else {
				// Parse today and yesterday. This script is usually ran once per day. Parsing
				// the two days will ensure yesterday's information is correct, otherwise
				// the information between this script's last run and yesterday midnight
				// would get lost.
				$today = time();
				$yesterday = time()-(60*60*24);
				
				$parse_day_today = gmmktime(0, 0, 0, gmstrftime("%m", $today), gmstrftime("%d", $today), gmstrftime("%Y", $today));
				$days_to_parse[] = $parse_day_today;
				
				$parse_day_yesterday = gmmktime(0, 0, 0, gmstrftime("%m", $yesterday), gmstrftime("%d", $yesterday), gmstrftime("%Y", $yesterday));
				$days_to_parse[] = $parse_day_yesterday;
			}
		}
		
		while (!feof($fh)) {
			$line = fgets($fh, 4096);
			$line = trim($line);
			
			if (empty($line)) continue;
			
			$arr = explode('|', $line);
			if (count($arr) != 6) {
				// Shouldn't happen, but it has been reported
				continue;
			}
			
			list($cvstime, $user, $dir, $module, $rev, $file) = $arr;
			
			$time_parsed = hexdec(substr($cvstime, 1, 8));
			$day = gmmktime( 0, 0, 0, gmstrftime("%m", $time_parsed ), gmstrftime("%d", $time_parsed ), gmstrftime("%Y", $time_parsed ) );
			
			if (!$parse_all && !in_array($day, $days_to_parse)) {
				// we don't want to parse this day
				continue;
			}
			
			if (!array_key_exists($day, $days)) {
				$days[$day] = array();
			}
			$days[$day][] = array('action' => substr($cvstime, 0, 1), 'user' => $user, 'dir' => $dir, 'module' => $module, 'rev' => $rev,
				'file' => $file);
		}
		
		fclose($fh);
		// No data to parse
		if (count($days) == 0) return;
		
		foreach ($days as $day => $day_entry) {
			$day_data = $this->parseDay($day_entry, $project);
			
			$day_str = gmstrftime('%Y-%m-%d', $day);
			
			$c = new Criteria();
			$c->add(ScmProjectStatsPeer::STAT_DATE, $day_str);
			$c->add(ScmProjectStatsPeer::PROJECT_ID, $project->getProjectId());
			$stats = ScmProjectStatsPeer::doSelectOne($c);
			if (is_null($stats)) {
				$stats = new ScmProjectStats();
				$stats->setProjectId($project->getProjectId());
				$stats->setStatDate($day_str);
			}
			
			$stats->setCheckouts($day_data['checkouts']);
			$stats->setCommits($day_data['commits']);
			$stats->setAdds($day_data['adds']);
			$stats->save();
			
			// now store the user information
			foreach ($day_data['user_data'] as $user_name => $user_data) {
				$user = UserPeer::retrieveByUnixName($user_name);
				if (is_null($user)) {
					// Not a GForge user -- use the undefined user
					$user = UserPeer::retrieveByPk(User::UNDEFINED_USER_ID);
					if (is_null($user)) continue;
				}
				
				$c = new Criteria();
				$c->add(ScmUserStatsPeer::STAT_DATE, $day_str);
				$c->add(ScmUserStatsPeer::PROJECT_ID, $project->getProjectId());
				$c->add(ScmUserStatsPeer::USER_ID, $user->getUserId());
				
				$stats = ScmUserStatsPeer::doSelectOne($c);
				if (!$stats) {
					$stats = new ScmUserStats();
					$stats->setStatDate($day_str);
					$stats->setProjectId($project->getProjectId());
					$stats->setUserId($user->getUserId());
				}
				
				$stats->setCheckouts($user_data['checkouts']);
				$stats->setCommits($user_data['commits']);
				$stats->setAdds($user_data['adds']);
				$stats->save();
			}
			
		}
	}
	
	public function parseDay($dayData) {
		$commits = 0;
		$checkouts = 0;
		$adds = 0;
		
		$user_data = array();
		
		foreach ($dayData as $entry) {
			$user = $entry['user'];
			if ($user && !array_key_exists($user, $user_data)) {
				$user_data[$user] = array('checkouts' => 0, 'commits' => 0, 'adds' => 0);
			}
			if (strtolower($entry['action']) == 'm') {
				$commits++;
				if ($user) $user_data[$user]['commits']++;
			}
			if (strtolower($entry['action']) == 'a') {
				$adds++;
				if ($user) $user_data[$user]['adds']++;
			}
			if (strtolower($entry['action']) == 'o') {
				$checkouts++;
				if ($user) $user_data[$user]['checkouts']++;
			}
		}
		
		$return = array(
					'commits' => $commits,
					'adds' => $adds,
					'checkouts' => $checkouts,
					'user_data' => $user_data
				);
		return $return;
	}
}
$cronjob = new ParseCVSHistoryCronjob();
$cronjob->run();
?>