<?php
/**
 * Copyright (c) 2005-2008 GForge, LLC
 *
 * @version $Id$
 * @author Daniel A. P�rez	daniel@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('system/Cronjob.php');
require_once('core/CronHistory.php');
require_once('core/ProjectPeer.php');
require_once('core/PluginPeer.php');
require_once('core/ProjectPluginPeer.php');
require_once('plugins/scmgit/lib/ScmgitPlugin.php');
require_once('core/Role.php');
require_once('core/RoleSetting.php');
require_once('GFConf.php');
require_once('plugins/scmgit/lib/GITRepository.php');

class CreateDavGitAuth extends Cronjob {
	private $repoPath;
	private $accessFilePath;
	private $authFilePath;

	public function __construct() {
		$this->jobname = basename(__FILE__);
		$this->repoPath = GFConf::get('scmgit.localGitRoot');
		if ($this->repoPath[strlen($this->repoPath)-1] != '/') {
			$this->repoPath .= '/';
		}

		$this->accessFilePath = SystemUtils::buildFilePath(GFConf::get('system.varDir'), 'gitroot-access');
		$this->authFilePath = SystemUtils::buildFilePath(GFConf::get('system.varDir'), 'gitroot-authfile');

		parent::__construct();
	}


	/**
	 * Checks whether the project has the git repository created or not
	 *
	 * @param project object
	 * @return boolean
	 */
	private function repoExists($project) {
		if (is_dir($this->repoPath . $project->getUnixName())) {
			return true;
		} else {
			return false;
		}
	}

	public function run($useSSH = false) {
		if ($useSSH) {
			$this->deleteAccessFile();
			$this->deleteAuthFile();
		} else {
			$this->createAccessFile();
			$this->createAuthFile();
		}
		$this->commitLog();
	}
	
	private function deleteAccessFile() {
		if (file_exists($this->accessFilePath) && !is_writable($this->accessFilePath)) {
			$this->logCron("Can't open " . $this->accessFilePath . " for read/write, please check permissions");
			$this->commitLog();
			return false;
		}

		$contents = '[groups]'."\n";
		if (!file_put_contents($this->accessFilePath, $contents)) {
			$this->logCron('Couldn\'t write to ' . $this->accessFilePath);
		}

	}

	private function deleteAuthFile() {
		if (file_exists($this->authFilePath) && !is_writable($this->authFilePath)) {
			$this->logCron("Can't open " . $this->authFilePath . " for read/write, please check permissions");
			$this->commitLog();
			return false;
		}
		$contents = "\n";
		if (!file_put_contents($this->authFilePath,$contents)) {
			$this->logCron('Couldn\'t write to ' . $this->authFilePath);
		}
	}

	private function createAccessFile() {
		if (file_exists($this->accessFilePath) && !is_writable($this->accessFilePath)) {
			$this->logCron("Can't open " . $this->accessFilePath . " for read/write, please check permissions");
			$this->commitLog();
			return false;
		}

		// Retrieve all the projects that use Git
		$c = new Criteria();
		$c->add(PluginPeer::PLUGIN_NAME, 'scmgit'); // get the id
		$c->addJoin(ProjectPluginPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->addJoin(ProjectPeer::PROJECT_ID, ProjectPluginPeer::PROJECT_ID);
		
		// If splitting SCM servers, create only the entries for the repositories hosted on this server
		if (GFConf::get('system.splitScmServers') && GFConf::varExists('scmgit.scmServerId')) {
			$c->addJoin(ProjectScmPeer::PROJECT_ID, ProjectPeer::PROJECT_ID);
			$c->add(ProjectScmPeer::SCM_SERVER_ID, GFConf::get('scmgit.scmServerId'));
		}
		
		$projects = ProjectPeer::doSelect($c);
		$groups = array();
		$paths = array();

		foreach ($projects as $project) {
			$committers = array();
			$readers = array();

			// Generate the "readers" and "writers" for this project
			$users = $project->getMembers();
			foreach ($users as $user) {
				$permission = $project->getUserPermission($user);
				
				if ($permission->check('scm', 0, ScmPlugin::WRITE)) {
					$committers[] = $user->getUnixName();
				} else if ($permission->check('scm', 0, ScmPlugin::READ)) {
					$readers[] = $user->getUnixName();
				}
			}
			unset($users);
			
			// See if other users can read
			$project_scm = ProjectScmPeer::retrieveByPk($project->getProjectId());
			if ($project_scm->getEnableAnonread() && $project->getIsPublic()) {
				$readers[] = 'anonymous';
			}

			// Store this project's info
			$paths[$project->getUnixName()] = $this->getPathPermissions($project);
			$groups[] = $project->getUnixName().'-readers = '.implode(',', $readers);
			$groups[] = $project->getUnixName().'-committers = '.implode(',', $committers);
		}
		unset($projects);
		
		// We have all the info now... generate the file
		$contents = '[groups]'."\n";
		$contents .= implode("\n", $groups);
		$contents .= "\n\n";
		foreach ($paths as $project_name => $subpaths) {
			foreach ($subpaths as $path => $perms) {
				$contents .= '['.$project_name.':'.$path.']'."\n";
				$contents .= implode("\n", $perms);
				$contents .= "\n\n";
			}
		}

		//echo $contents."\n";
		if (!file_put_contents($this->accessFilePath, $contents)) {
			$this->logCron('Couldn\'t write to ' . $this->accessFilePath);
		}

		// we should be root to run this
		$output = array();
		$return_value = 0;
		exec('chown ' . GFConf::get('system.apacheUser') . ' ' . $this->accessFilePath, $output, $return_value);
		if ($return_value != 0) {
			$this->logCron('Couldn\'t change permissions for apache user ' . $this->accessFilePath . ' , please change manually as root');
			foreach ($output as $errline) {
				$this->logCron($errline);
			}
		}
	}

	private function getPathPermissions(Project $project) {
		$repos = new GitRepository($project);
		$perms = array();

		// First, check the repository-wide permissions
		$perms['/'] = array();
		$perms['/'][] = '@'.$project->getUnixName().'-committers = rw';
		$perms['/'][] = '@'.$project->getUnixName().'-readers = r';

		// See if other users have read access
		$project_scm = ProjectScmPeer::retrieveByPk($project->getProjectId());
		if ($project_scm->getEnableAnonread() && $project->getIsPublic()) {
			$perms['/'][] = '* = r';
		}

		$modules = $repos->getModules();
		foreach ($modules as $module) {
			$path = $module->getPath();
			if (!array_key_exists($path, $perms)) {
				$perms[$path] = array();
			}

			$module_perms = $repos->getModulePermissions($module->getScmModuleId());
			//var_dump($module_perms);
			foreach ($module_perms as $module_perm) {
				if (is_null($module_perm['module_access'])) {
					// this may happen if the user's permission for the module aren't set
					continue;
				}
				
				// If it's a group, we must give the correct permissions to the users inside the group
				$user = $module_perm['user'];
				
				$users = array();
				if ($user->getIsGroup()) {
					$ugus = $user->getUsergroupUsersRelatedByUsergroupId();
					foreach ($ugus as $ugu) {
						$users[] = UserPeer::retrieveByPk($ugu->getMemberUserId());
					}
				} else {
					$users[] = $user;
				}

				foreach ($users as $user) {
					$add = "";
					if ($module_perm['module_access'] == ScmPlugin::NOACCESS) {
						//set no access privileges only for users with no previous access permissions set  
						if(!isset($perms[$path][$user->getUnixName()])){
							$add = ' ';
						}
					} elseif ($module_perm['module_access'] == ScmPlugin::READ) {
						//set read permissions only for users with not read access set 
						if( !isset($perms[$path][$user->getUnixName()]) || /*have not permissions set*/
							strpos($perms[$path][$user->getUnixName()], 'r') === false /*OR not read permission already set*/){
							$add .= 'r';
						}
					} elseif ($module_perm['module_access'] == ScmPlugin::WRITE) {

						if( !isset($perms[$path][$user->getUnixName()]) || /*have not permissions set */
							strpos($perms[$path][$user->getUnixName()], 'r') === false /*OR not read permission already set*/){
							$add .= 'r';
						}
						
						if( !isset($perms[$path][$user->getUnixName()]) || /*have not permissions set*/
							strpos($perms[$path][$user->getUnixName()], 'w') === false /*OR not write permission already set*/){
							$add .= 'w';
						}
					}
					
					if( !isset($perms[$path][$user->getUnixName()]) ||
						strpos($perms[$path][$user->getUnixName()], '=') === false ){
						$perms[$path][$user->getUnixName()] = $user->getUnixName() . " = " . $add;
					}else{													
						$perms[$path][$user->getUnixName()]  .=  $add;
					}										
				}
			}
		}
		unset($modules);

		return $perms;
	}

	private function createAuthFile() {
		if (file_exists($this->authFilePath) && !is_writable($this->authFilePath)) {
			$this->logCron("Can't open " . $this->authFilePath . " for read/write, please check permissions");
			$this->commitLog();
			return false;
		}

		/* Get all the users that have access to the Git repositories */
		// Retrieve all the projects that use Git
		$c = new Criteria();
		$c->add(PluginPeer::PLUGIN_NAME, 'scmgit'); // get the id
		$c->addJoin(ProjectPluginPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->addJoin(ProjectPeer::PROJECT_ID, ProjectPluginPeer::PROJECT_ID);
		$projects = ProjectPeer::doSelect($c);
		
		$users_access = array();

		foreach ($projects as $project) {
			$users = $project->getMembers();
			foreach ($users as $user) {
				$unix_name = $user->getUnixName();
				if (array_key_exists($unix_name, $users_access)) continue;
				$permission = $project->getUserPermission($user);
				
				if ($permission->check('scm', 0, ScmPlugin::READ)) {
					$users_access[$unix_name] = $user;
				}
			}
		}
		
		$contents = '';
		foreach ($users_access as $unix_name => $user) {
			// We are using the crypt password because the htpasswd binary which creates (manually) the file
			// can use md5 encryption but it's a modified, propietary of apache one
			$contents .= $unix_name . ':' . $user->getPasswordCrypt() . "\n";
		}
		$contents .= "anonymous:\n";
		
		
		if (!file_put_contents($this->authFilePath,$contents)) {
			$this->logCron('Couldn\'t write to ' . $this->authFilePath);
		}
		$output = array();
		$return_value = '';
		exec('chown ' . GFConf::get('system.apacheUser') . ' ' . $this->authFilePath,$output,$return_value); // we should be root to run this
		if ($return_value!=0) {
			$this->logCron('Couldn\'t change permissions for apache user to ' . $this->authFilePath . ' , please change manually as root');
			foreach ($output as $errline) {
				$this->logCron($errline);
			}
		}
	}
}

$useSSH = GFConf::get('scmgit.useSSH');
$cronjob = new CreateDavGitAuth();
$cronjob->run($useSSH);
?>
