<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id: aclcheck.php 816 2006-04-18 17:36:20Z marcelo $
 * @author Marcelo Mottalli	marcelo@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
 
require_once('initialize.php');
require_once('GFConf.php');
require_once('SystemUtils.php');
require_once('core/Project.php');
require_once('plugins/scm/lib/ScmCommit.php');
require_once('plugins/scm/lib/ScmPreference.php');
require_once('plugins/scm/lib/ScmPlugin.php');
require_once('plugins/scmsvn/lib/SVNRepository.php');


// Everything must be written to stderr
function log_error($error) {
	file_put_contents("php://stderr", $error);
}

if ($argc != 3) {
	log_error('Usage: '.basename(__FILE__).' <repository path> <txn>');
	exit(1);
}

// Get the project UNIX name from the SVNROOT
$svnroot = preg_replace('/\\/$/', '', $argv[1]);	// remove trailing slash
$path_arr = explode('/', $argv[1]);
$project_name = $path_arr[count($path_arr)-1];
if (!$project_name) {
	log_error('Couldn\'t retrieve project name from SVNROOT: '.$argv[1]);
	exit(1);
}

$project = ProjectPeer::retrieveByUnixName($project_name);
if (is_null($project)) {
	log_error('Invalid project name: '.$project_name);
	exit(1);
}

// Get svnlook path
$svnlook = SystemUtils::buildFilePath(GFConf::get('scmsvn.svnpath'), '/svnlook');
	

$repository = new SVNRepository($project);

if ($repository->getAccessType() == SVNRepository::ACCESS_TYPE_SSH) {

	/**************************************************
	 ******************* ACL CHECK ********************
	 **************************************************/
	// Retrieve the user name from the environment
	$user_arr = posix_getpwuid(posix_geteuid());
	$user_name = $user_arr['name'];
	
	if (!$user_name) {
		log_error('Couldn\'t retrieve user name');
		exit(1);
	}
	
	if ($user_name == 'anonymous') {
		log_error('You cannot commit as the anonymous user.');
		exit(1);
	}
		
	// Start the session as the user that's committing
	$user = UserPeer::retrieveByUnixName($user_name);
	if (is_null($user)) {
		log_error('Invalid user name: '.$user_name."\n");
		exit(1);
	}
	Session::getSession()->setUser($user);

	// Get the directories where the user is committing
	$cmd = $svnlook.' dirs-changed -t '.escapeshellarg($argv[2]).' '.escapeshellarg($argv[1]);
	$dirs = array();
	SystemUtils::execute($cmd, $dirs);
	
	foreach ($dirs as $dir) {
		$dir = SystemUtils::buildFilePath('/', $dir);
		if (!$repository->userCanCommit($dir)) {
			log_error(Language::getSessText('Scm.NoWriteAccess', array($user_name, $dir))."\n");
			exit(1);
		}
	}
}

/**************************************************
 ******************* LOG CHECK ********************
 **************************************************/
ActionManager::notify("SCMSVN_PRECOMMIT", $project);

// See if we must check for a tracker item ID in the message log
$preferences = ScmPreferencePeer::retrieveByPk($project->getProjectId());
if (!is_null($preferences) && $preferences->getRequireTrackerItemId()) {

	$log = array();
	$cmd = 'LC_ALL="en_US.UTF-8" '.$svnlook.' log -t '.escapeshellarg($argv[2]).' '.escapeshellarg($argv[1]);
	SystemUtils::execute($cmd, $output);
	$log = implode("\n", $output);

	$invalid_ids = array();
	$messages = array();

    $arr_statuses = ScmCommit::getStatusFromMessages($log);
    $arr_time_entries = ScmCommit::getTimeEntriesFromMessages($log);
    $tids_valid = ScmCommit::checkTrackerItemIds($log, $project, $invalid_ids, $messages);
	if (!$tids_valid || count($arr_statuses['invalid']) > 0 || count($arr_time_entries['invalid']) > 0) {

        if (!$tids_valid) {
            if (count($invalid_ids) > 0) {
                log_error('The following Tracker Item Ids are not valid for this Project or you are violating commit rules:'."\n");
                log_error('Details: ' . "\n");
                foreach($messages as $id => $message) {
                    log_error('[#'.$id.'] = ' . $message . "\n");
                }
            } else {
                log_error('Commit not related to any Tracker Item.'."\n");
            }
            log_error("\n");
            log_error('Please specify one or more valid Tracker Item Ids in the commit '."\n");
            log_error('log as "[#nnnn]". For example:'."\n");
            log_error('$ svn commit -m "Fixed [#2708]"'."\n");
        }

        if (count($arr_statuses['invalid']) > 0) {
            log_error("\nThe commit message contains invalid status update references:\n");
            foreach($arr_statuses['invalid'] as $tid => $arr_data) {
	            $status = $arr_data['status'];
	            $err_code = $arr_data['errcode'];

	            $error_msg = "";
	            switch($err_code) {
		            case ScmCommit::STATUS_UPDATE_INVALID_ALIAS:
			            log_error(Language::getSessText('ScmCommitMessage.invalidAlias'));
			            break;

		            case ScmCommit::STATUS_UPDATE_CANT_TRANSITION:
			            log_error(Language::getSessText('ScmCommitMessage.cannotTransition'));
			            break;

		            case ScmCommit::STATUS_UPDATE_COMMIT_ACTION:
			            log_error(Language::getSessText('ScmCommitMessage.commitAction'));
			            break;

		            case ScmCommit::STATUS_UPDATE_NO_NEXT_STATUS:
			            log_error(Language::getSessText('ScmCommitMessage.noNextStatus'));
			            break;
	            }

	            log_error("\n");
	            if ($err_code == ScmCommit::STATUS_UPDATE_INVALID_ALIAS || $err_code == ScmCommit::STATUS_UPDATE_NO_NEXT_STATUS) {
		            log_error(" > Id " . $tid . ", status: " . $status . "\n");
		            log_error(" > Available statuses: " . ScmCommit::getStatusesByTid($tid));
	            }

            }
        }

        if (count($arr_time_entries['invalid']) > 0) {
            log_error("\nThe commit message contains invalid time entry references:\n");
            foreach($arr_statuses['invalid'] as $tid => $time_entry) {
                log_error(" > Id " . $tid . ", time entry: " . $time_entry . "\n");
                log_error(" > Time entries must be greater than : " . ScmCommit::MIN_TIME_ENTRY);
            }
        }

		exit(1);
	}
}

exit(0);
?>
