<?php
/**
 * Copyright (c) 2005-2011 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli	<marcelo@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('SystemUtils.php');
require_once('plugins/scmsvn/lib/exceptions/SVNException.php');
require_once('plugins/scm/lib/SCMRepository.php');
require_once('plugins/scm/lib/ScmCommitFilter.php');
require_once('plugins/scm/lib/ScmCommitFilterProject.php');


class SVNRepository extends SCMRepository {
	private $svnRoot;
	private $revisionCache = array();
	private $preCommitFile, $postCommitFile;
	
	public function __construct(Project $project) {
		parent::__construct($project);

		$this->svnRoot = GFConf::get('scmsvn.localSVNRoot');
		$this->svnpath = GFConf::get('scmsvn.svnpath');
		
		$this->preCommitFile = SystemUtils::buildFilePath($this->repositoryPath, '/hooks/pre-commit');
		$this->postCommitFile = SystemUtils::buildFilePath($this->repositoryPath, '/hooks/post-commit');

	}
	
	public function getType() {
		return 'svn';
	}
	
	/**
	 * Initialize this repository
	 */
	protected function initializeRepository() {
		$repos_type = GFConf::get('scmsvn.reposType');
		$error = SystemUtils::execute("{$this->svnpath}/svnadmin create --fs-type {$repos_type} {$this->repositoryPath}", $output);
		$this->createDefaultSkeleton();
	}
	
	/**
	 * Change the repository's permissions in order to enable anonymous access
	 */
	public function enableAnonAccess() {
		SystemUtils::execute('chmod -R o+r '.$this->repositoryPath, $output);
		// Anonymous can't write
		SystemUtils::execute('chmod -R o-w '.$this->repositoryPath, $output);
		SystemUtils::execute('chmod o+x '.$this->repositoryPath, $output);
		// Allow access to directories
		SystemUtils::execute('find '.$this->repositoryPath.' -type d -exec chmod o+x {} \;', $output);
	}
	
	/**
	 * Change the repository's permissions in order to disable anonymous access
	 */
	public function disableAnonAccess() {
		SystemUtils::execute('chmod -R o-r '.$this->repositoryPath, $output);
		SystemUtils::execute('chmod -R o-w '.$this->repositoryPath, $output);
		SystemUtils::execute('chmod o-x '.$this->repositoryPath, $output);
		// Disable access to directories
		SystemUtils::execute('find '.$this->repositoryPath.' -type d -exec chmod o-x {} \;', $output);
	}

	public function updateHooks() {
		if (!file_exists($this->preCommitFile)) {
			$preCommit = $this->getDefaultPreCommitFile();
		} else {
			$preCommit = file($this->preCommitFile);
			$preCommit = array_map("trim", $preCommit);
		}
		
		if (!file_exists($this->postCommitFile)) {
			$postCommit = $this->getDefaultPostCommitFile();
		} else {
			$postCommit = file($this->postCommitFile);
			$postCommit = array_map("trim", $postCommit);
		}
		
		// Add misc configuration directives
		$this->updatePreCommitHooks($preCommit);
		$this->updatePostCommitHooks($postCommit);
		$this->updateSVNMail($postCommit);
		
		SystemUtils::writeFile($this->preCommitFile, $preCommit);
		SystemUtils::writeFile($this->postCommitFile, $postCommit);
	}
	
	public function updateExtraPermissions() {
		$output = array();
		
		SystemUtils::execute("chmod 770 {$this->preCommitFile}", $output);
		SystemUtils::execute("chmod 770 {$this->postCommitFile}", $output);
	}
	
	public function updateSVNMail(&$lines) {
		// Read the preferences
		$preferences = $this->getScmPreferences();

 		if (!is_null($preferences) && $preferences->getMailCommits()) {
			$this->addSVNMail($lines);
		} else {
			$this->removeSVNMail($lines);
		}
	}

	
	/**
	* This function creates the post-commit file in svn hooks (if it doesn't exist)
	*/
	function getDefaultPostCommitFile() {
		return array(
			'#!/bin/sh',
			'REPOS="$1"',
			'REV="$2"',
			''	// empty line
		);
	}
	
	/**
	* This function creates the pre-commit file in svn hooks (if it doesn't exist)
	*/
	function getDefaultPreCommitFile() {
		return array(
			'#!/bin/sh',
			'REPOS="$1"',
			'TXN="$2"',
			''	// empty line
		);
	}

	public function addSVNMail(&$lines) {
		// First check if we have mailman configured
		if (!$this->project->usesPlugin('mailman')) {
			return;
		}

		$found = false;
		$i=0;
		foreach ($lines as $line) {
			if (strpos($line, 'commit-email.pl') !== false) {
				$found = true;
				break;
			}
			$i++;
		}

		$commitScript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmsvn/bin/commit-email.pl');
		if ($found) {
			// File already has SVN mail, overwrite the line since the
			// host might have changed
			$lines[$i] = $commitScript.' "$REPOS" "$REV" '.$this->repositoryName.'-commits@'.GFConf::get('mailman.listsHost');
		} else {
			$lines[] = $commitScript.' "$REPOS" "$REV" '.$this->repositoryName.'-commits@'.GFConf::get('mailman.listsHost');
		}
	}
	
	public function removeSVNMail(&$lines) {
		$new_content = array();
		for ($i=0; $i < count($lines); $i++) {
			$line = trim($lines[$i]);
			
			if (strpos($line, '/scmsvn/bin/commit-email.pl') !== false) {
				// Found commit-email line... skip current line
			} else {
				$new_content[] = $line;
			}
		}
		$lines = $new_content;
	}
	
	/**
	 * updatePostCommitHooks - Add the commit hooks for ActionManager
	 */
	public function updatePostCommitHooks(&$lines) {
		$this->insertCommitFilters($lines, ScmCommitFilter::POST_COMMIT);

		$found = false;
		foreach ($lines as $line) {
			if (strpos($line, '#BEGIN GForge commit hooks') !== false) {
				$found = true;
				break;
			}
		}
		// File already has commit hooks
		if ($found) {
			return;
		}

		$postscript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmsvn/bin/post_commit.php');
        
        $lines[] = '#BEGIN GForge commit hooks';
        $lines[] = 'php -q -d include_path='.ini_get('include_path').' '.$postscript.' "$REPOS" "$REV"';
        $lines[] = '#END GForge commit hooks';
	}
	
	/**
	 * updatePreCommitHooks - Add pre-commit hooks
	 */
	public function updatePreCommitHooks(&$lines) {
		$this->insertCommitFilters($lines, ScmCommitFilter::PRE_COMMIT);

		$found = false;
		foreach ($lines as $line) {
			if (strpos($line, '#BEGIN GForge pre-commit checks') !== false) {
				$found = true;
				break;
			}
		}
		// File already has commit hooks
		if ($found) {
			return;
		}

		$precommitcheckscript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmsvn/bin/pre_commit_checks.php');
        
        $lines[] = '#BEGIN GForge pre-commit checks';
        $lines[] = 'php -q -d include_path='.ini_get('include_path').' '.$precommitcheckscript.' "$REPOS" "$TXN"';
        $lines[] = '#END GForge pre-commit checks';
	}
	
	/**
	 * Inserts the commit filter commands
	 */
	private function insertCommitFilters(&$lines_src, $type=ScmCommitFilter::PRE_COMMIT) {
		$lines_dest = array();

		for ($i = 0; $i < count($lines_src); $i++) {
			$line = $lines_src[$i];
			if (strpos($line, '#BEGIN GForge filters') !== false) {
				// Seek to the end of the marker
				for (; $i < count($lines_src); $i++) {
					$line = $lines_src[$i];
					if (strpos($line, '#END GForge filters') !== false) {
						break;
					}
				}
			} else {
				$lines_dest[] = $line;
			}
		}
		
		// Now append the filters

		$c = new Criteria();
		$c->add(ProjectPluginPeer::PROJECT_ID, $this->project->getProjectId());
		$c->addJoin(ProjectPluginPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->add(PluginPeer::PLUGIN_NAME, ScmsvnPlugin::getName());		// Make sure we're not including filters we shouldn't
		$c->addJoin(ScmCommitFilterProjectPeer::PROJECT_ID, ProjectPluginPeer::PROJECT_ID);
		$c->addJoin(ScmCommitFilterPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->addJoin(ScmCommitFilterProjectPeer::SCM_COMMIT_FILTER_ID, ScmCommitFilterPeer::SCM_COMMIT_FILTER_ID);
		$c->add(ScmCommitFilterPeer::FILTER_TYPE, $type);
		$filters = ScmCommitFilterPeer::doSelect($c);
		
		$lines_dest[] = '#BEGIN GForge filters';
		foreach ($filters as $filter) {
			$command = $filter->getFilterString();
			$command = StringUtils::unHtmlSpecialChars($command);
			$command = preg_replace('/\\$PROJECT/i', $this->project->getUnixName(), $command);
			$command = preg_replace('/\\$PATH/i', '\\$1', $command);
			$command = preg_replace('/\\$FILE/i', '\\$2', $command);
			$lines_dest[] = $command . ' || exit $? ';
		}
		$lines_dest[] = '#END GForge filters';
		
		$lines_src = $lines_dest;
	}
	/**
	 * checkoutFile - Retrieve a file into a temporary directory from the repository
	 *
	 * @param	string $filename	File to checkout
	 * return	string	The name of the copy where the file has been saved
	 */

	public function checkoutFile($filename) {
	}
	
	/**
	 * commitFile - Commit a file to the repository
	 *
	 * @param	string $file	File to commit
	 * @param	string $message	Commit message
	 */
	function commitFile($file,$message="Automatic updated by scmsvn") {
	}
	
	/**
	 * releaseFile - Remove the file that was just checked out from the repository
	 * @see checkoutFile
	 */
	function releaseFile($file) {
		// $file is something like /tmp/(tmp_dir)/path/to/file
		// we must delete /tmp/tmp_dir
		if (!preg_match("/^(\/tmp\/[^\/]+)\/.*/", $file, $result)) {		// Make sure the dir is under /tmp
			$this->err = "Trying to release a directory not in /tmp. Skipping...";
			$this->logCron();
			return;
		}
		$dir = $result[1];
		
		// this shouldn't happen... but add it as a security check
		if (!preg_match('/[^\\/]/',$dir)) {
			$this->err = "Trying to delete root dir. Skipping...";
			$this->logCron();
			return;
		}
		$dir = escapeshellarg($dir);
		system("rm -rf ".$dir);
	}

	/**
	 * Returns the SVNROOT for this repository
	 * @return	string
	 */
	public function getRoot() {
		return $this->repositoryPath;
	}
	
	/**
	 * Returns the for ALL repositories
	 * @return	string
	 */
	public function getScmRoot() {
		return GFConf::get('scmsvn.localSVNRoot');
	}
	
	/**
	 * Create the SVN repo default skeleton
	 * /trunk, /branches and /tags
	 *
	 */
	public function createDefaultSkeleton() {
		$path = "file://{$this->repositoryPath}/trunk file://{$this->repositoryPath}/branches file://{$this->repositoryPath}/tags";
		$cmd = "{$this->svnpath}/svn mkdir $path -m 'Initial Repository Skeleton'";
		SystemUtils::execute($cmd, $output);
	}
	
	public function doLocalCheckout($destdir, $module=null) {
		$path = $this->repositoryPath;
		if ($module) {
			$path = SystemUtils::buildFilePath($path, $module);			// Will be something like /svnroot/project/trunk/module
		}
		
		$cmd = SystemUtils::buildFilePath($this->svnpath, 'svn')." co file://$path $destdir";
		$output = array();
		$error = SystemUtils::execute($cmd, $output);
		
		if ($error) {
			throw new Exception("Unable to checkout the code - Error code $error - Command was '$cmd'");
		}
	}
	
	public function doLocalExport($destdir, $module=null) {
		$path = $this->repositoryPath;
		if ($module) {
			$path = SystemUtils::buildFilePath($path, $module);			// Will be something like /svnroot/project/trunk/module
		}
		
		$cmd = SystemUtils::buildFilePath($this->svnpath, 'svn')." export --force file://$path $destdir";
		$output = array();
		$error = SystemUtils::execute($cmd, $output);
		
		if ($error) {
			throw new Exception("Unable to export the code - Error code $error - Command was '$cmd'");
		}
	}	

	public function getCheckedoutFileRevision($checkout_dir, $file) {
		if (!isset($this->revisionCache[$checkout_dir])) {
			$this->revisionCache[$checkout_dir] = array();
		} elseif (isset($this->revisionCache[$checkout_dir][$file])) {
			return $this->revisionCache[$checkout_dir][$file];
		}

		$fullpath = SystemUtils::buildFilePath($checkout_dir, $file);
		
		$output = array();
		$cmd = $this->svnpath."/svn info --xml $fullpath";
		$error = SystemUtils::execute($cmd, $output);		// Note: DON'T use $explode_output=false on the call to SystemUtils because we need the return value
		
		if ($error) {
			$rev = -1;
		} else {
			$output = implode("\n", $output);
			$xml = new SimpleXMLElement($output);
			if (!$xml || !isset($xml->entry)) {
				$rev = -1;
			} else {
				$rev = intval($xml->entry['revision']);
			}
		}
		
		// Cache for speeded-up access (can be called several times from the scm_index cronjob)
		$this->revisionCache[$checkout_dir][$file] = $rev;
		return $rev;
	}

	protected function getGlobalAccessType() {
		return (GFConf::get('scmsvn.useSSH') ? SCMRepository::ACCESS_TYPE_SSH : SCMRepository::ACCESS_TYPE_DAV);
	}
}
?>
