<?php

require_once('plugins/tracker/lib/TrackerPlugin.php');
require_once('plugins/tracker/lib/TrackerItemSupportNotification.php');

class TrackerItemNagEmailCronjob extends Cronjob {
	const SUPPORT_TEAM = 'tracker_nag';
	const CUSTOMER =  'tracker_cus_nag';

	public function __construct() {
		$this->jobname = basename(__FILE__);
		parent::__construct();
	}

	public function showDate($timestamp) {
		return date("Y-m-d H:i:s", $timestamp);
	}


	public function run() {
		global $argv;
		$current_time = gmmktime();
		$this->logCron('Real current time: ' . $current_time . ' ('.$this->showDate($current_time).')');

		if (isset($argv[3])) {
			$current_time = intval($argv[3]);
			$this->logCron('Test mode, current time: ' . $current_time);
		}

		// Run this cronjob only on weekdays
		$week_day = date("w");
		if ($week_day == 6 || $week_day == 0) {
			// Saturday or sunday
			//return;
		}

		//Set the nobody user as the session user
		$user = UserPeer::retrieveByPk(User::UNDEFINED_USER_ID);
		$session = Session::getSession();
		$session->setUser($user, true);
		$this->logCron('Session user: ' . $user->getUserId());

		//Process status alerts
		$this->processTrackers($current_time);

		$this->commitLog();
	}

	public function formatHoursSeconds($seconds) {
		$hours = floor($seconds / 3600);
		$minutes = floor(($seconds / 60) % 60);

		return sprintf("%02d hours %02d minutes", $hours, $minutes);
	}


	public function processTrackers($current_time) {
		//Get trackers that have this feature enabled
		$c = new Criteria();
		$c1 = $c->getNewCriterion(SettingPeer::SETTING_NAME, 'cus_nagalerts');
		$c2 = $c->getNewCriterion(SettingPeer::SETTING_NAME, 'sup_nagalerts');
		$c1->addOr($c2);
		$c->add(SettingPeer::SETTING_VALUE, '1');
		$c->addJoin(TrackerPeer::TRACKER_ID, SettingPeer::REF_ID);
		$c->setDistinct();
		$Trackers = TrackerPeer::doSelect($c);

		$this->logCron('Retrieved trackers that have nagging enabled: ' . sizeof($Trackers));

		//Process each Tracker
		foreach($Trackers as $Tracker) {
			$this->logCron('Processing Tracker: ' . $Tracker->getTrackerName() . ', Project: ' . $Tracker->getProject()->getProjectName());

			//Get auto-close settings for the tracker
			$auto_close_setting = $Tracker->getAutoCloseSeting();
			$auto_close_status = $Tracker->getAutoCloseStatus();
			$close_status = TrackerExtraFieldElementPeer::retrieveByPk($auto_close_status);
			$auto_close_status_alias = '';
			if (is_object($close_status)) {
				$auto_close_status_alias = $close_status->getAlias();
			}

			//Get nag settings for submitter
			$nag_submitter_enabled = $Tracker->getStatusNag(true);

			//Get nag settings for assignee
			$nag_assignee_enabled = $Tracker->getStatusNag(false);

			//Get the tracker statuses
			$tracker_statuses = $Tracker->getTrackerStatuses(Tracker::STATUS_OPEN);
			$this->logCron('Tracker statuses: ' . sizeof($tracker_statuses));

			//Loop on tracker statuses
			foreach($tracker_statuses as $tracker_status) {
				//Get nag interval for submitter
				$nag_submitter_interval = $Tracker->getStatusNagtime($tracker_status->getElementId(), true);
				//Get nag interval for assignee
				$nag_assignee_interval = $Tracker->getStatusNagtime($tracker_status->getElementId(), false);

				//Get tracker items that currently have this status
				$this->logCron('Element ID to pull tracker items: ' . $tracker_status->getElementId());
				$tracker_items = $Tracker->getTrackerItemsByStatus($tracker_status->getElementId());
				$this->logCron('Tracker items: ' . sizeof($tracker_items));

				//Loop through tracker items and check for each one if we need to send the nag email
				foreach($tracker_items as $tracker_item) {
					//Get submitter
					$submitter_id = $tracker_item->getSubmittedBy();

					//Get assignees
					$assignees = $tracker_item->getAssignedUsers();

					//Check who is the TI waiting for
					$send_to_submitter = false;
					if ($tracker_item->getWaitingFor() == TrackerItem::WAITINGFOR_SUBMITTER) {
						$send_to_submitter = true;
					}

					if ($send_to_submitter) {
						//if autoclose is disabled, skip this TI
						if (!$tracker_item->getAutocloseEnabled()) {
							$this->logCron('Autoclose disabled for Tracker item: #' . $tracker_item->getTrackerItemId() . ", skipping");
							continue;
						}

						//Gets the saved auto-close timestamp for this TI
						$auto_close_timestamp = $tracker_item->getAutoCloseTimestamp();
						$this->logCron('Auto close timestamp for tracker item: [#' . $tracker_item->getTrackerItemId() . '], is: ' . $auto_close_timestamp . ' ('.$this->showDate($auto_close_timestamp).')');
						if ($auto_close_timestamp == 0) {
							//There's no timestamp set, create one
							$auto_close_timestamp = gmmktime();
							$tracker_item->setAutoCloseTimestamp($auto_close_timestamp);
						}
					} else {
						$tracker_item->setAutoCloseTimestamp(0);
					}

					$last_alert_date = $tracker_item->getLastAlertDate($send_to_submitter);

					if ($last_alert_date == 0) {
						$this->logCron('No timestamp set for TI [#'.$tracker_item->getTrackerItemId().'], setting timestamp and skipping');
						$tracker_item->setAlertEntry($send_to_submitter, $current_time);
						continue;
					}

					$this->logCron('Tracker item: [#' . $tracker_item->getTrackerItemId() . ']');
					$this->logCron('Tracker item last alert date: ' . $last_alert_date);

					//Check is nagging is enabled for support team or customer (submitter), and get nag interval
					if ($send_to_submitter) {
						$nag_enabled = $nag_submitter_enabled;
						$nag_interval = $nag_submitter_interval;
						$this->logCron('TI waiting on submitter, nag enabled: ' . $nag_enabled);
					} else {
						$nag_enabled = $nag_assignee_enabled;
						$nag_interval = $nag_assignee_interval;
						$this->logCron('TI waiting on assignee, nag enabled: ' . $nag_enabled);
					}

					//If it's assigned to submitter, check if in auto-close condition and close the ti
					$pending_hours_to_close = 0;
					if ($send_to_submitter) {
						if ($auto_close_timestamp > 0 && (($auto_close_timestamp + $auto_close_setting*3600) <= $current_time)) {
							$this->logCron('This TI time expired, auto-closing');
							$tracker_item->bypassConfirmation();
							$tracker_item->closeItem($auto_close_status_alias);
							continue;
						} else {
							if ($auto_close_timestamp > 0) {
								$remaining_seconds = ($auto_close_timestamp + $auto_close_setting*3600 - $current_time);
							} else {
								$remaining_seconds = $auto_close_setting*3600;
							}
							$pending_hours_to_close = $this->formatHoursSeconds($remaining_seconds);
							$this->logCron('Pending hours to close: ' . $pending_hours_to_close);
						}
					}

					$this->logCron('Validating action, last alert: ' . $last_alert_date . ' ('.$this->showDate($last_alert_date).')' . ', last alert + interval: ' . ($last_alert_date + $nag_interval*3600) . ' ('.$this->showDate(($last_alert_date + $nag_interval*3600) ).')');
					if ($nag_enabled && ( ($last_alert_date + $nag_interval*3600) <= $current_time) ) {
						$this->logCron('Cleared to send nag email for ti: [#' . $tracker_item->getTrackerItemId() . ']');

						$send_to_users = array();
						//Build list of users to email
						if ($send_to_submitter) {
							$user = UserPeer::retrieveByPk($submitter_id);
							$send_to_users[] = $user;
						} else {
							foreach($assignees as $assignee) {
								if ($assignee->getUserId() != User::UNDEFINED_USER_ID && $assignee->isActive()) {
									$send_to_users[] = $assignee;
								}
							}
						}

						//Create the notification object
						$notification = new TrackerItemSupportNotification($tracker_item, $send_to_users, $pending_hours_to_close, $send_to_submitter);
						//Queue the notification
						NotificationManager::getManager()->queue($notification);

						//Create an entry in audit trail for this tracker item
						$this->logCron('Saving last alert date for ti: [#' . $tracker_item->getTrackerItemId() . '], alert time: ' . $current_time);
						$tracker_item->setAlertEntry($send_to_submitter, $current_time);
					} else {
						$this->logCron('NOT cleared to send nag email for ti: [#' . $tracker_item->getTrackerItemId() . '] yet');
					}
				}
			}
		}
	}

}

$cronjob = new TrackerItemNagEmailCronjob();
$cronjob->run();

?>
